/*
 * Copyright 2000, 2001 The Regents of the University of California 
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

/* High-level functions for Waikato DAG interfaces (devices or tracefiles).
 *
 *  $Id: coral_iface_dag.c,v 1.5.2.3 2001/11/21 01:33:47 kkeys Exp $
 *
 */

#include "config.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>

#include "libcoral.h"
#include "libcoral_priv.h"
#include "coral_clock.h"

/* These aren't used by code, but they're useful in a debugger. */
typedef struct {
    coral_timestamp_t ts;
    int wand_crc;
    union atm_hdr atm_hdr;
    char payload[ATM_PAYLOAD_SIZE];
} coral_dag_atm_cell_t;
typedef struct {
    coral_timestamp_t ts;
    int caplen;
    int totlen;
    char payload[ATM_PAYLOAD_SIZE];
} coral_dag_pos_cell_t;

static const char RCSid[] = "$Id: coral_iface_dag.c,v 1.5.2.3 2001/11/21 01:33:47 kkeys Exp $";

/* always little endian */	/* XXX */
static inline void dag_clock_order(const coral_iface_t *iface,
    const coral_timestamp_t *src, coral_timestamp_t *dst)
{
    dst->i[0] = crl_letohl(src->i[1]);
    dst->i[1] = crl_letohl(src->i[0]);
    /* not the same as dst->i64 = crl_letoh64(src->i64); */
}

static inline void dag_clock_sec_nsec(const coral_iface_t *iface,
    const coral_timestamp_t *t, time_t *sec, long *nsec, int normalize)
{
    coral_timestamp_t to;

    dag_clock_order(iface, t, &to);
    *sec = to.i[0];
    *nsec = (double)to.i[1] / 4.294967296 + .5;
}

/* returns -1 for error, 1 for complete packet, 0 for no packet */
static int coral_dag_pos_prep_pkt(coral_iface_t *iface)
{
    void *cell;
    coral_pkt_result_t *pkt_result = &iface->pkt_result;

    cell = coral_iccell(iface);

    iface->packetbuf.caplen = ntohl(*(uint32_t*)((char*)cell + 8));
    iface->packetbuf.totlen = ntohs(*(uint16_t*)((char*)cell + 14));

    iface->packetbuf.caplen -= 16;	/* DAG caplen includes DAG header */
    iface->packetbuf.totlen -= 2;	/* DAG totlen includes FCS */

    if (iface->packetbuf.caplen < iface->packetbuf.totlen &&
	!(coral_config.flags & CORAL_OPT_PARTIAL_PKT))
    {
	/* discard partial packet */
	iface->pkt_stats.truncated++;
	return 0;
    }

    iface->pkt_stats.pkts_recv++;
    iface->pkt_stats.ok_packet++;
    iface->packetbuf.buf = coral_cell_payload_req(iface, cell);
    iface->packetbuf.protocol = iface->iface_info.datalink;
    pkt_result->timestamp = coral_cell_time_req(iface, cell);
    pkt_result->packet = &iface->packetbuf;
    pkt_result->header = NULL;
    pkt_result->trailer = NULL;

    coral_clock_timespec_inline2(iface, iface->pkt_result.timestamp,
	&iface->ts, dag_clock_sec_nsec);

    return 1;
}

static void dag_iface_init(coral_iface_t *iface)
{
    iface->time_is_normal = 1;
}


#define CLOCKFUNC dag_clock_sec_nsec
#define coral_atm_prep_pkt coral_dag_prep_pkt
#include "coral_atm.h"


coral_iface_type_t coral_iface_type_dag_atm = {
    {
	64,	/* size */
	0,	/* offset of timestamp */
	12,	/* offset of ATM header */
	-1,	/* offset of ATM HEC */
	16,	/* offset of payload */
	-1,	/* offset of caplen */
	-1,	/* offset of totlen */
	-1	/* unused3 */
    },
    "dag (atm)",
    CORAL_API_BLOCK | CORAL_API_CELL | CORAL_API_PKT,
    dag_iface_init,
    dag_clock_order,
    coral_generic_clock_double,
    dag_clock_sec_nsec,
    NULL /* correct_clock */,
    coral_cell_consume,
    coral_atm_consume_pkt,
    coral_atm_prep_pkt,
    NULL /* update_pkt_stats */,
    coral_atm_iface_close,
    { { CORAL_RX, 48 }, KBPS_OC3c, CORAL_PHY_ATM, CORAL_DLT_ATM_RFC1483, NULL }
};

/* For internal use only.  User never sees POS "cells". */
coral_iface_type_t coral_iface_type_dag_pos = {
    {
	64,	/* size */
	0,	/* offset of timestamp */
	-1,	/* offset of ATM header */
	-1,	/* offset of ATM HEC */
	16,	/* offset of payload */
	-1,	/* offset of caplen */
	-1,	/* offset of totlen */
	-1	/* unused3 */
    },
    "dag (pos)",
    CORAL_API_PKT,
    dag_iface_init,
    dag_clock_order,
    coral_generic_clock_double,
    dag_clock_sec_nsec,
    NULL /* correct_clock */,
    coral_cell_consume,
    NULL /* consume_pkt */,
    coral_dag_pos_prep_pkt,
    NULL /* update_pkt_stats */,
    NULL /* close */,
    { { CORAL_RX, 48 }, KBPS_OC3c, CORAL_PHY_POS, CORAL_DLT_CHDLC, NULL }
};

