/* 
 * Copyright 1999, 2000, 2001 The Regents of the University of California 
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

/* High-level functions for FORE ATM interfaces (devices or tracefiles).
 *
 *  $Id: coral_iface_fatm.c,v 1.3.2.4 2001/11/21 01:33:47 kkeys Exp $
 *
 */

#include "config.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>

#include "libcoral.h"
#include "libcoral_priv.h"
#include "coral_clock.h"

/* Seconds in each tick of the hardware clock */
#define FORE_OC3_SEC_PER_TICK           40e-9

static const char RCSid[] = "$Id: coral_iface_fatm.c,v 1.3.2.4 2001/11/21 01:33:47 kkeys Exp $";

/*
 * The "hard" clock is a 16-bit counter kept on the card.  Almost (!) every
 * time it wraps around, the "soft" clock, a 32-bit counter, is incremented by
 * an interrupt handler.  It is possible for the soft clock to miss a hard
 * clock wrap, so the code handling these clocks needs to check for a hard
 * clock wrap where the soft clock didn't get changed.  We assume that each
 * timestamp passed into this function will always be >= the timestamp
 * passed into the previous call.
 *
 * To make life interesting, there are two different OC3 clock formats.
 * Say that S3S2S1S0 is the 32-bit firmware clock, where S3 is the MSB and S0 
 * the LSB.  Similarly, H1H0 is the hardware clock and F1F0 the FIFO depth.
 *
 * If the data is in network order, the format is:
 *       +-------------------+
 *       | S3 | S2 | S1 | S0 |
 *       +-------------------+
 *       | H1 | H0 | F1 | F0 |
 *       +-------------------+
 *
 * Otherwise, the data has had the byte order reversed completely:
 *       +-------------------+
 *       | F0 | F1 | H0 | H1 |
 *       +-------------------+
 *       | S0 | S1 | S2 | S3 |
 *       +-------------------+
 */

static inline void fatm_oc3_clock_order(const coral_iface_t *iface,
    const coral_timestamp_t *src, coral_timestamp_t *dst)
{
    if (iface->iface_info.time_is_le) {
	dst->i[0] = crl_letohl(src->i[1]);
	dst->i[1] = ((src->c[3] << 8) | src->c[2]);
    } else {
	dst->i[0] = crl_betohl(src->i[0]);
	dst->i[1] = ((src->c[4] << 8) | src->c[5]);
    }
}

static inline void fatm_oc3_clock_sec_nsec(const coral_iface_t *iface,
    const coral_timestamp_t *t, time_t *sec, long *nsec, int normalize)
{
    coral_timestamp_t to;
    double result;

    fatm_oc3_clock_order(iface, t, &to);
    result = ((double)0x10000 * to.i[0] + to.i[1]) * FORE_OC3_SEC_PER_TICK;
    *sec = result;
    *nsec = (result - *sec) * 1000000000 + 0.5;

    /*assert(iface->time_is_normal == 0);*/
    if (normalize) {
	*sec += iface->iface_info.capture_tv.tv_sec;
	*nsec += iface->iface_info.capture_tv.tv_usec * 1000;
	if (*nsec >= 1000000000) {
	    (*sec)++;
	    *nsec -= 1000000000;
	}
    }
}

/* When the FATM hardware clock wraps, it interrupts the firmware, which
 * increments the firmware clock.  But several cells may arrive before the
 * firmware handles the interrupt, so they will have incorrect timestamps.
 * We try to correct that here.
 */
static int fatm_correct_clock(coral_iface_t *iface,
    coral_blk_info_t *binfo, coral_fatm_atm_cell_t *cell)
{
    int i, count, indx;
    coral_timestamp_t to; /* ordered timestamp: 0 firmware, 1 hardware */

    if (coral_config.flags & CORAL_OPT_FATM_RAW_TIME)
	return 0;

    count = ntohl(binfo->cell_count);

    for (i = 0, indx = 0; i < count; i++) {
	fatm_oc3_clock_order(iface, &cell[i].t, &to);

	if (to.i[0] < iface->last_clock.to.i[0]) {
	    /* Fw decreased; reset our correction. */
	    iface->last_clock.correction = 0;
	    if (to.i[0] == 0) /* Clock was reset. */
		if (!iface->has_read_blocks) /* Only do this on the 1st block */
		    indx = i;

	} else {
	    /* If hw wrapped, increment correction. */
	    if (to.i[1] < iface->last_clock.to.i[1]) {
		iface->last_clock.correction++;
	    }
	    /* If fw increased, subtract the delta from the correction. */
	    if (to.i[0] > iface->last_clock.to.i[0]) {
		iface->last_clock.correction -=
		    (to.i[0] - iface->last_clock.to.i[0]);
		if (iface->last_clock.correction < 0)
		    iface->last_clock.correction = 0;
	    }
	}
	/* remember values */
	iface->last_clock.to = to;
	/* add correction */
	if (iface->last_clock.correction) {
	    to.i[0] += iface->last_clock.correction;
	    if (iface->iface_info.time_is_le) {
		cell[i].t.i[1] = crl_htolel(to.i[0]);
	    } else {
		cell[i].t.i[0] = crl_htobel(to.i[0]);
	    }
	}
    }
    return indx;
}

static void fatm_iface_init(coral_iface_t *iface)
{
    iface->time_is_normal = 0;
}


#define CLOCKFUNC fatm_oc3_clock_sec_nsec
#define coral_atm_prep_pkt coral_fatm_prep_pkt
#include "coral_atm.h"


coral_iface_type_t coral_iface_type_fatm = {
    OLD_CELL_LAYOUT,
    "fatm",
    CORAL_API_BLOCK | CORAL_API_CELL | CORAL_API_PKT,
    fatm_iface_init,
    fatm_oc3_clock_order,
    coral_generic_clock_double,
    fatm_oc3_clock_sec_nsec,
    fatm_correct_clock,
    coral_cell_consume,
    coral_atm_consume_pkt,
    coral_atm_prep_pkt,
    NULL /* update_pkt_stats */,
    coral_atm_iface_close,
    { { CORAL_RX, 48 }, KBPS_OC3c, CORAL_PHY_ATM, CORAL_DLT_ATM_RFC1483,
	NULL /* determined dynamically from iomode */ }
};

