/* 
 * Copyright 1999, 2000, 2001 The Regents of the University of California 
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

/* High-level functions for Apptel POINT interfaces (devices or tracefiles).
 *
 *  $Id: coral_iface_point.c,v 1.3.2.3 2001/11/21 01:33:48 kkeys Exp $
 *
 */

#include "config.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>

#include "libcoral.h"
#include "libcoral_priv.h"
#include "coral_clock.h"

/* Seconds in each tick of the hardware clock */
#define APPTEL_OC12_SEC_PER_TICK        25e-9

static const char RCSid[] = "$Id: coral_iface_point.c,v 1.3.2.3 2001/11/21 01:33:48 kkeys Exp $";

/* point clock is always little endian */
static inline void point_clock_order(const coral_iface_t *iface,
    const coral_timestamp_t *src, coral_timestamp_t *dst)
{
    dst->i[0] = crl_letohl(src->i[1]);

#ifdef WORDS_BIGENDIAN
    dst->i[1] = ((src->c[1] >> 4) | (src->c[2] << 4) | (src->c[3] << 12));
#else
    dst->i[1] = (src->i[0] >> 12);
#endif
}

static inline void point_clock_sec_nsec(const coral_iface_t *iface,
    const coral_timestamp_t *t, time_t *sec, long *nsec, int normalize)
{
    coral_timestamp_t to;
    double result;

    point_clock_order(iface, t, &to);
    result = ((double)0x100000 * to.i[0] + to.i[1]) * APPTEL_OC12_SEC_PER_TICK;
    *sec = result;
    *nsec = (result - *sec) * 1000000000 + 0.5;

    /*assert(iface->time_is_normal == 0);*/
    if (normalize) {
	*sec += iface->iface_info.capture_tv.tv_sec;
	*nsec += iface->iface_info.capture_tv.tv_usec * 1000;
	if (*nsec >= 1000000000) {
	    (*sec)++;
	    *nsec -= 1000000000;
	}
    }
}

/* Traces may contain a clock reset. */
static int point_correct_clock(coral_iface_t *iface,
    coral_blk_info_t *binfo, coral_point_atm_cell_t *cell)
{
    int i, count, indx;
    coral_timestamp_t to; /* host-ordered timestamp */

    if (coral_config.flags & CORAL_OPT_POINT_RAW_TIME)
	return 0;

    if (iface->has_read_blocks) return 0; /* Only do this on the 1st block */
    count = ntohl(binfo->cell_count);
    if (count < 2) return 0;

    point_clock_order(iface, &cell[0].t, &to);
    iface->last_clock.to = to;

    for (i = 1, indx = 0; i < count; i++) {
	point_clock_order(iface, &cell[i].t, &to);
	if (to.i[0] == 0 && to.i[0] < iface->last_clock.to.i[0]) {
	    /* Clock was reset. */
	    indx = i;
	}
	iface->last_clock.to = to;
    }
    return indx;
}

static void point_iface_init(coral_iface_t *iface)
{
    iface->time_is_normal = 0;
}


#define CLOCKFUNC point_clock_sec_nsec
#define coral_atm_prep_pkt coral_point_prep_pkt
#include "coral_atm.h"


coral_iface_type_t coral_iface_type_point = {
    OLD_CELL_LAYOUT,
    "point",
    CORAL_API_BLOCK | CORAL_API_CELL | CORAL_API_PKT,
    point_iface_init,
    point_clock_order,
    coral_generic_clock_double,
    point_clock_sec_nsec,
    point_correct_clock,
    coral_cell_consume,
    coral_atm_consume_pkt,
    coral_atm_prep_pkt,
    NULL /* update_pkt_stats */,
    coral_atm_iface_close,
    { { CORAL_RX, 48 }, 0, CORAL_PHY_ATM, CORAL_DLT_ATM_RFC1483, POINT_FW }
};

