/*
 * Copyright 2000, 2001 The Regents of the University of California 
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

/* High-level functions for TSH interfaces.
 *
 *  $Id: coral_iface_tsh.c,v 1.23.2.1 2001/11/21 01:33:48 kkeys Exp $
 *
 */

#include "config.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

#include "libcoral.h"
#include "libcoral_priv.h"
#include "coral_clock.h"

static const char RCSid[] = "$Id: coral_iface_tsh.c,v 1.23.2.1 2001/11/21 01:33:48 kkeys Exp $";

/* always big endian */
static inline void tsh_clock_order(const coral_iface_t *iface,
    const coral_timestamp_t *src, coral_timestamp_t *dst)
{
    dst->i[0] = ntohl(src->i[0]);
    dst->i[1] = ntohl(src->i[1]);
}

static inline void tsh_clock_sec_nsec(const coral_iface_t *iface,
    const coral_timestamp_t *t, time_t *sec, long *nsec, int normalize)
{
    coral_timestamp_t to;

    tsh_clock_order(iface, t, &to);
    *sec = to.i[0];
    *nsec = (to.i[1] & 0xFFFFFF) * 1000;
    if (normalize && !iface->time_is_normal) {
	*sec += iface->iface_info.capture_tv.tv_sec;
	*nsec += iface->iface_info.capture_tv.tv_usec * 1000;
	if (*nsec >= 1000000000) {
	    (*sec)++;
	    *nsec -= 1000000000;
	}
    }
}

/* returns -1 for error, 1 for complete packet, 0 for no packet */
static int coral_tsh_prep_pkt(coral_iface_t *target_iface)
{
    coral_iface_t *iface[2];
    void *cell;
    unsigned ifnum;
    const char *payload;
    const struct ip *ip;
    int hl;

    iface[0] = cinst[target_iface->src->id[0]];
    iface[1] = cinst[target_iface->src->id[1]];

    cell = coral_iccell(iface[0]);
    if ((ifnum = ((char*)cell)[4] - 1) > 1) {
	coral_diag(0, ("%s: bad interface number %u\n",
	    target_iface->src->filename, ifnum));
	if (coral_verbosity >= 2) {
	    coral_puts("record:\n");
	    coral_fprint_data(coral_get_errfile(), 2, cell,
		coral_cell_size(target_iface));
	}
	errno = -1;
	return -1;
    }

    if (iface[ifnum]->eof)
	return 0; /* discard data on this iface XXX? */

#if 0 /* XXX tsh packets are always truncated. */
    if (!(coral_config.flags & CORAL_OPT_PARTIAL_PKT)) {
	/* discard partial packet */
	target_iface->pkt_stats.truncated++;
	return 0;
    }
#endif

    if (iface[0]->packetbuf.is_dynamic) {
	free((char*)iface[0]->packetbuf.buf);
	iface[0]->packetbuf.is_dynamic = 0;
    }
    payload = coral_cell_payload_req(iface[0], cell);
    ip = (const struct ip*)payload;
    hl = ip->ip_hl * 4;
    if (ip->ip_v == 4 && hl >= 20) {
	if (hl > 20) {				/* IP options were lost */
	    char *buf = malloc(hl + 16);
	    if (!buf) return -1;
	    memcpy(buf, payload, 20);		/* copy IP header */
	    memset(buf + 20, 0, hl - 20);	/* pad IP options with 0 */
	    memcpy(buf + hl, payload + 20, 16);	/* copy transport header */
	    iface[0]->packetbuf.is_dynamic = 1;
	    iface[0]->packetbuf.buf = buf;
	} else {
	    iface[0]->packetbuf.buf = payload;	/* use payload without copy */
	}
	iface[0]->packetbuf.caplen = hl + 16;
	iface[0]->packetbuf.totlen = -1;
	iface[0]->packetbuf.protocol = CORAL_NETPROTO_IPv4;
    } else {
	iface[0]->packetbuf.buf = payload;
	iface[0]->packetbuf.caplen = 36;
	iface[0]->packetbuf.totlen = -1;
	iface[0]->packetbuf.protocol = CORAL_PROTO_UNKNOWN;
    }

    target_iface->pkt_stats.pkts_recv++;
    target_iface->pkt_stats.ok_packet++;
    iface[0]->pkt_result.timestamp = coral_cell_time_req(iface[0], cell);
    iface[0]->pkt_result.packet = &iface[0]->packetbuf;
    iface[0]->pkt_result.header = NULL;
    iface[0]->pkt_result.trailer = NULL;
    iface[1]->pkt_result = iface[0]->pkt_result;

#if 0
    coral_clock_timespec_inline2(iface[0], pkt_result->timestamp,
	&target_iface->ts, tsh_clock_sec_nsec);
#endif

    return 1;
}

static int coral_tsh_consume(coral_iface_t *target_iface, int want)
{
    coral_iface_t *iface[2];
    const coral_timestamp_t *ts;
    int ifnum;
    int result = 0;

    iface[0] = cinst[target_iface->src->id[0]];
    iface[1] = cinst[target_iface->src->id[1]];

    if (iface[0]->u.blk.node.data) {
        /* Move to the next cell. */
        iface[0]->u.blk.node.index++;
        if (iface[0]->u.blk.node.index < iface[0]->u.blk.node.cell_count) {
	    ts = coral_cell_time_req(iface[0], coral_iccell(iface[0]));

	    /* mod 2 to guarantee a legal ifnum.  An illegal ifnum will be
	     * caught in prep_pkt. */
	    ifnum = (ts->c[4] - 1) % 2;

	    /* Use next cell to update timestamp */
	    coral_clock_timespec_inline2(iface[0], ts,
		&iface[ifnum]->ts, tsh_clock_sec_nsec);
	    iface[!ifnum]->ts.tv_sec = iface[ifnum]->ts.tv_sec;
	    iface[!ifnum]->ts.tv_nsec = 1000000000; /* purposely illegal */
	    result = 1;
        } else {
            /* We've reached the end of the current block. */
            coral_diag(10, ("end of block on ifaces %d, %d\n",
                iface[0]->id, iface[1]->id));
            /* Get next blk from queue. (Assume empty blks are not queued.) */
	    iface[0]->u.blk.tail = NULL;
	    iface[0]->u.blk.node.data = NULL;
	    iface[0]->u.blk.node.cell_count = 0;
        }
    }

    /* coral_consume doesn't know ifaces are combined, so we must handle here */
    iface[0]->have_data = result;
    iface[1]->have_data = result;

    return result;
}

/* For internal use only.  User never sees TSH "cells". */
coral_iface_type_t coral_iface_type_tsh = {
    {
	44,	/* size */
	0,	/* offset of timestamp */
	-1,	/* offset of ATM header */
	-1,	/* offset of ATM HEC */
	8,	/* offset of payload */
	-1,	/* offset of caplen */
	-1,	/* offset of totlen */
	-1	/* unused3 */
    },
    "tsh",
    CORAL_API_PKT,
    NULL, /* time_is_normal was set by coral_tsh_init */
    tsh_clock_order,
    coral_generic_clock_double,
    tsh_clock_sec_nsec,
    NULL /* correct_clock */,
    coral_tsh_consume,
    NULL /* consume_pkt */,
    coral_tsh_prep_pkt,
    NULL /* update_pkt_stats */,
    NULL /* close */,
    { { CORAL_RX, 36 }, 0, CORAL_PROTO_UNKNOWN, CORAL_PROTO_UNKNOWN, NULL }
};

