/* 
 * Copyright 1999, 2000, 2001 The Regents of the University of California 
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

/*
 * coral_init.c - main entry point for libcoral.  
 *		sets up stuff to make it work.
 *		routines to initialize & open devices, check setup, etc...
 */

static const char RCSid[]="$Id: coral_init.c,v 1.122.2.3 2001/11/21 01:33:48 kkeys Exp $";

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <time.h>

#ifndef HAVE_PWD_H
# undef HAVE_GETPWNAM
#else
# ifdef HAVE_GETPWNAM
#  include <pwd.h>      /* getpwnam() */
# endif
#endif

#ifdef HAVE_LIBPCAP
# include <pcap.h>
#endif

#ifdef HAVE_CORAL_DEVICES
# include "drivers/coral_ioctl.h"
#endif
#include "libcoral.h"
#include "libcoral_priv.h"

#if 0
coral_internal_stats_t coral_internal_stats;
#endif

coral_source_t *csource[CORAL_MAXOPEN];
coral_iface_t *cinst[CORAL_MAXOPEN];
fd_set coral_select_fds;
#ifndef FD_SETSIZE
# define FD_SETSIZE 1024
#endif
coral_source_t *coral_fd_table[FD_SETSIZE];
int coral_select_fd_max = -1;
int coral_select_count = 0;	/* number of sources that need select() */
int coral_noselect_count = 0;	/* number of sources that don't need select() */
int coral_offline_ifaces = 0;	/* number of valid non-live interfaces */

coral_dev_config_t null_dev_config = {
    { CORAL_RX_UNKNOWN, -1 },	/* iomode */
    0,				/* bandwidth */
    CORAL_PROTO_UNKNOWN		/* protocol */
    /* remaining fields are 0 */
};

/* if you write a new coral source type, you need to add it here */
static const coral_src_type_t *coral_types[] = {
    &coral_src_type_file,
    &coral_src_type_dagfile,
#ifdef HAVE_CORAL_POINT
    &coral_src_type_point,
#endif
#ifdef HAVE_CORAL_FATM
    &coral_src_type_fatm,
#endif
#ifdef HAVE_CORAL_DAG
    &coral_src_type_dag,
#endif
#ifdef HAVE_LIBPCAP
    &coral_src_type_pcap,
    &coral_src_type_pcaplive,
#endif
    &coral_src_type_tsh,
    NULL
};

coral_iface_type_t *coral_iface_types[] = {
#ifdef HAVE_CORAL_POINT
    &coral_iface_type_point,
#endif
#ifdef HAVE_CORAL_FATM
    &coral_iface_type_fatm,
#endif
#ifdef HAVE_CORAL_DAG
    &coral_iface_type_dag_atm,
    &coral_iface_type_dag_pos,
#endif
#ifdef HAVE_LIBPCAP
    &coral_iface_type_pcap,
#endif
    &coral_iface_type_tsh,
    NULL
};

int coral_new_instance(coral_source_t *src, int sid)
{
    int id;

    for (id = 0; id < CORAL_MAXOPEN && cinst[id]; id++)
	;

    if (id == CORAL_MAXOPEN) {
	coral_diag(0, ("coral_open: too many open coral devices\n"));
	return -1;
    }

    cinst[id] = (coral_iface_t *)calloc(1, sizeof(coral_iface_t));
    if (!cinst[id]) {
	coral_diag(0, ("coral_new_instance: calloc: %s\n", strerror(errno)));
	return -1;
    }

    cinst[id]->src = src;
    src->id[sid] = cinst[id]->id = id;
    cinst[id]->sid = sid;
    cinst[id]->eof = 1; /* not started */
    cinst[id]->has_read_blocks = 0;
    if (coral_max_iface <= id)
	coral_max_iface = id + 1;

    return id;
}

static coral_source_t *coral_init_source(void)
{
    int sd, i;
    static int first_time = 1;

    if (first_time) {
	first_time = 0;
	memset(coral_fd_table, sizeof(coral_fd_table), 0);
    }

    for (sd = 0; sd < coral_config.maxsrc && csource[sd]; sd++)
	;

    if (sd == coral_config.maxsrc) {
	coral_diag(0, ("coral_open: too many coral sources\n"));
	errno = EMFILE;
	return NULL;
    }

    csource[sd] = (coral_source_t *)calloc(1, sizeof(coral_source_t));
    if (!csource[sd]) {
	coral_diag(0, ("coral_new_source: malloc: %s\n", strerror(errno)));
	return NULL;
    }
    csource[sd]->sd = sd;

    csource[sd]->eof = 1; /* not started */
    csource[sd]->fd = -1;  /* initialize to invalid fd */
    csource[sd]->version = 0;
    for (i = 0; i < CORAL_MAXOPEN; i++) {
	csource[sd]->id[i] = -1;
    }
    csource[sd]->dev_config = null_dev_config;
    csource[sd]->dev_config.vlan_hack = coral_config.dev_config.vlan_hack;
    csource[sd]->proto_rules.head = csource[sd]->proto_rules.tail = NULL;

    coral_config.source_count++;
    if (coral_max_src <= sd)
	coral_max_src = sd + 1;
    return csource[sd];
}

coral_source_t *coral_new_source(const char *name)
{
    const char *filename;
    coral_source_t *src;

    if (!(src = coral_init_source()))
	return NULL;

    if ((filename = strchr(name, ':'))) {
	filename++;
    } else {
	filename = name;
    }

    src->name = strdup(name);
    src->filename = coral_filename(filename, NULL, PATH_MAX);
    if (!src->name || !src->filename) {
	coral_diag(0, ("coral_new_source: malloc: %s\n", strerror(errno)));
	coral_close(src);
	errno = ENOMEM;
	return NULL;
    }

    return src;
}

coral_source_t *coral_new_fdsource(int fd, const char *name)
{
    coral_source_t *src;

    if (!(src = coral_init_source()))
	return NULL;

    if (name) {
	src->name = strdup(name);
	src->filename = strdup(name);
    } else if ((src->name = malloc(16))) {
	sprintf(src->name, "fd.%d", fd);
	src->filename = strdup(src->name);
    }
    if (!src->name || !src->filename) {
	coral_diag(0, ("coral_new_fdsource: malloc: %s\n", strerror(errno)));
	coral_close(src);
	errno = ENOMEM;
	return NULL;
    }
    src->fd = fd;

    return src;
}

/* If any dev_config fields are unset, set them from default_config. */
void coral_set_dev_config_defaults(coral_dev_config_t *dev_config,
    coral_dev_config_t *default_config)
{
    if (dev_config->iomode.flags == CORAL_RX_UNKNOWN)
	dev_config->iomode = default_config->iomode;
    if (dev_config->iomode.first_n < 0)
	dev_config->iomode.first_n = default_config->iomode.first_n;
    if (dev_config->bandwidth <= 0)
	dev_config->bandwidth = default_config->bandwidth;
    if (dev_config->physical == CORAL_PROTO_UNKNOWN)
	dev_config->physical = default_config->physical;
    if (dev_config->datalink == CORAL_PROTO_UNKNOWN)
	dev_config->datalink = default_config->datalink;
    if (!dev_config->firmware && default_config->firmware)
	dev_config->firmware = strdup(default_config->firmware);
}

/*
 * coral_open_all - 
 * first function called.  opens all initialized members of csource[].
 * this should do everything necessary to get the card ready to go,
 * including loading firmware, etc...  the next call should be coral_start
 *
 * returns - number of opened sources if successful, or -1 for failure.
 */

int coral_open_all(void)
{
    int total = 0;	/* number of opened interfaces */
    int n, sd;

    for (sd = 0; sd < CORAL_MAXOPEN; sd++) {
	if (!csource[sd]) continue;
	if ((n = coral_open(csource[sd])) < 0)
	    return -1;
	total += n;
    }
    return total;
}

int coral_open(coral_source_t *src)
{
    int i, id, error, len;
    struct stat statbuf;
    int preopened;

#if 0 /* why was this here? */
    static int sigpipe_handler_installed = 0;

    if (!sigpipe_handler_installed) {
	if (signal(SIGPIPE, SIG_IGN) == SIG_ERR)
	    coral_diag(0, ("signal: %s\n", strerror(errno)));
	else
	    sigpipe_handler_installed = 1;
    }
#endif

    validate_src(src, "coral_open", -1);
    preopened = (src->fd >= 0);

    /* If any src->dev_config fields are unset, set them from coral_config. */
    coral_set_dev_config_defaults(&src->dev_config, &coral_config.dev_config);
    /* If they're still unset, they'll be set to device defaults later. */

    /* Open first instance here; type.init() will open more as needed. */
    if ((id = coral_new_instance(src, 0)) < 0)
	goto coral_open_error;
    src->iface_count = 1;
    cinst[id]->devinfo.type = CORAL_TYPE_NONE; /* unknown */

    len = strcspn(src->name, ":/");
    if (src->name[len] == ':') {
	/* Detect prefix and set type explicitly */
	if (strncmp(src->name, "crl:", len) == 0) {
	    cinst[id]->devinfo.type = CORAL_TYPE_NONE;
	} else if (strncmp(src->name, "if:", len) == 0) {
	    cinst[id]->devinfo.type = CORAL_TYPE_PCAPLIVE;
	} else if (strncmp(src->name, "pcap:", len) == 0) {
	    cinst[id]->devinfo.type = CORAL_TYPE_PCAP;
	} else if (strncmp(src->name, "dag:", len) == 0) {
	    cinst[id]->devinfo.type = CORAL_TYPE_DAGFILE;
	} else if (strncmp(src->name, "tsh:", len) == 0) {
	    cinst[id]->devinfo.type = CORAL_TYPE_TSH;
	} else {
	    coral_diag(0, ("coral_open: %s: unknown prefix '%.*s'\n",
		src->name, len, src->name));
	    errno = ENOENT;
	    return -1;
	}

    } else {
	/* Look for suffix and set type explicitly */
	char *start, *end;
	start = end = strchr(src->filename, '\0');
retrysuffix:
	do {
	    if (--start <= src->filename || *start == '/') goto endsuffix;
	} while (*start != '.');
	if (strncmp(start, ".crl", end - start) == 0) {
	    cinst[id]->devinfo.type = CORAL_TYPE_NONE;
	} else if (strncmp(start, ".pcap", end - start) == 0) {
	    cinst[id]->devinfo.type = CORAL_TYPE_PCAP;
	} else if (strncmp(start, ".dag", end - start) == 0) {
	    cinst[id]->devinfo.type = CORAL_TYPE_DAGFILE;
	} else if (strncmp(start, ".tsh", end - start) == 0) {
	    cinst[id]->devinfo.type = CORAL_TYPE_TSH;
	} else if (strncmp(start, ".enc", end - start) == 0) {
	    end = start;
	    goto retrysuffix;
	} else if (strcmp(start, ".gz") == 0) {
	    end = start;
	    goto retrysuffix;
	}
    }
endsuffix:

    if (cinst[id]->devinfo.type == CORAL_TYPE_NONE) {
	/* Coral device or Coral tracefile */
	if (strcmp(src->filename, "-") == 0) {
	    src->fd = STDIN_FILENO;
	    cinst[id]->devinfo.type = CORAL_TYPE_FILE;

	} else {
	    if (!preopened) {
		if ((src->fd = open(src->filename, O_RDONLY, 0)) < 0) {
		    coral_diag(0, ("coral_open: %s: open: %s\n",
			src->name, strerror(errno)));
		    goto coral_open_error;
		}
		coral_diag(4, ("coral_open: %s: fd %d\n", src->name, src->fd));
	    }

	    if (fstat(src->fd, &statbuf) < 0) {
		coral_diag(0, ("coral_open: %s: stat: %s\n",
		    src->name, strerror(errno)));
		goto coral_open_error;
	    }
	    if ((statbuf.st_mode & S_IFMT) == S_IFDIR) {
		errno = EISDIR;
		coral_diag(0, ("coral_open: %s: %s\n",
		    src->name, strerror(errno)));
		goto coral_open_error;
	    }
	    if ((statbuf.st_mode & S_IFMT) != S_IFCHR)
		cinst[id]->devinfo.type = CORAL_TYPE_FILE;
	}
    }
    /* If type is still NONE, it must be a device (or invalid). */

    if (cinst[id]->devinfo.type != CORAL_TYPE_NONE) {
	/* not a Coral Device */
	if (src->dev_config.iomode.flags & CORAL_RX_ECHO)
	    coral_diag(1, ("warning: coral_open: %s: 'echo' iomode ignored"));
	if (src->dev_config.iomode.flags & CORAL_TX)
	    coral_diag(1, ("warning: coral_open: %s: 'tx' iomode ignored"));

    } else if (preopened) {
	coral_diag(0, ("coral_open: %s: unknown type", src->name));
	goto coral_open_error;

    } else {
#ifdef HAVE_CORAL_DEVICES
	/* device (or other non-regular, in which case ioctl will fail) */

	if (ioctl(src->fd, CORAL_FILLINFO, &cinst[id]->devinfo) < 0) {
	    coral_diag(0, ("coral_open: %s: ioctl FILLINFO: %s.  "
		"Retrying as regular file.\n", src->name, strerror(errno)));
	    /*goto coral_open_error;*/
	    goto end_device;
	}

	/* The drivers need O_RDWR because they do writes
	 * during initialization. (for now) */
	close(src->fd);
	if ((src->fd = open(src->filename, O_RDWR, 0)) < 0) {
	    coral_diag(0, ("coral_open: %s: open: %s\n",
		src->name, strerror(errno)));
	    goto coral_open_error;
	}
	/*coral_diag(4, ("coral_open: %s: fd %d\n", src->name, src->fd));*/

	if (cinst[id]->devinfo.version != CORAL_DRIVER_VERSION) {
	    coral_diag(0, ("coral_open: %s: "
		"incorrect driver version %d (expected %d)\n", 
		src->name, cinst[id]->devinfo.version,
		CORAL_DRIVER_VERSION));
	    errno = ENXIO;
	    goto coral_open_error;
	}

	/* Mmap in the driver's buffer for this card. */
	coral_diag(6, ("coral (%s): mmap %d blocks of %d bytes = %d bytes\n",
	    src->name,
	    cinst[id]->devinfo.num_blks, cinst[id]->devinfo.blk_size,
	    cinst[id]->devinfo.num_blks * cinst[id]->devinfo.blk_size));
	src->ubase = mmap(0,
	    cinst[id]->devinfo.num_blks * cinst[id]->devinfo.blk_size,
	    PROT_READ|PROT_WRITE, MAP_SHARED, src->fd, 0);
	if(src->ubase == MAP_FAILED) {
	    coral_diag(0, ("coral_open: %s: Could not mmap card's buffer: %s\n",
		src->name, strerror(errno)));
	    goto coral_open_error;
	}
# ifdef DEBUG
	/* coral_stats(src); */
# endif

	coral_diag(5, ("coral offset: %s: %p mmap: 0x%x\n",
	    src->name, cinst[id]->devinfo.kmem_buffer, (u_int)src->ubase));

#else	/* ! HAVE_CORAL_DEVICES */
	coral_diag(0, ("coral_open: %s is a device file, "
	    "and CoralReef was not built with device support.  "
	    "Retrying as regular file.\n", src->name));
	goto end_device;
	/*errno = ENODEV;*/
	/*goto coral_open_error;*/
#endif	/* HAVE_CORAL_DEVICES */
    }
end_device:

    /* find the fxn's for the type of device */
    for (i = 0; coral_types[i]; i++)
	if (coral_types[i]->coral_type == cinst[id]->devinfo.type) break;

    if (!coral_types[i]) {
	coral_diag(0, ("coral_open: %s: no support for coral_type: %d\n", 
	    src->name, cinst[id]->devinfo.type));
	errno = ENODEV;
	goto coral_open_error;
    }

    src->type = *coral_types[i];

    /* call the init function for this source's type */
    if (src->type.init(src) < 0) {
	if (!errno)
	    errno = ENODEV; /* XXX wrong errno */
	goto coral_open_error;
    }

    if ((src->dev_config.iomode.flags & CORAL_RX) && (src->dev_config.iomode.flags & CORAL_TX)) {
	coral_diag(0,
	    ("coral_open: %s: can't receive and transmit on same device\n",
	    src->name));
	errno = ENODEV;
	goto coral_open_error;
    }

    /*coral_fd_set[coral_fd(src)] = src;*/

    for (i = 0; i < src->iface_count; i++) {
	coral_iface_t *iface = cinst[src->id[i]];

	if (iface->iface_type->init)
	    iface->iface_type->init(iface);

	/* Hack to allow user to say "PPP" instead of "PPPoHDLC", etc.
	 * For each "PPP" interface, select the appropriate type of PPP
	 * framing according to the physical layer */
	if (iface->iface_info.datalink == CORAL_DLT_PPP) {
	    switch (iface->iface_info.physical) {
	    case CORAL_PHY_POS:
		iface->iface_info.datalink = CORAL_DLT_PPPoHDLC;
		break;
	    default:
		break;
	    }
	}

#ifdef HAVE_BPF_FILTER
	/* Each interface could have a different datalink type, so we can't
	 * just reuse a single compiled program on every interface.
	 */
	if (coral_config.bpf_filter_text) {
	    struct bpf_program *bpfprog;
	    bpfprog = malloc(sizeof(struct bpf_program));
	    if (!bpfprog) {
		coral_diag(0, ("malloc: %s\n", strerror(errno)));
		goto coral_open_error;
	    }
	    if (coral_pcap_compile(iface, bpfprog,
		coral_config.bpf_filter_text, 1, 0) < 0)
	    {
		free(bpfprog);
		errno = 0;
		goto coral_open_error;
	    }
	    if (coral_pcap_setfilter(iface, bpfprog) < 0) {
		free(bpfprog);
		errno = 0;
		goto coral_open_error;
	    }
	    iface->bpfprog = bpfprog;
	    iface->bpfprog_is_private = 1;
	}
#endif
    }

    return src->iface_count;

coral_open_error:
    error = errno;
    coral_close(src);
    errno = error;
    return -1;
}

/*
 * Mark iface as having reached EOF so we don't return any more data from it,
 * but don't deallocate anything.  iface->src may still have other ifaces that
 * we haven't finished reading yet.
 */
void coral_mark_eof(coral_iface_t *iface)
{
    int i, remaining;
    int saved_error = errno;
    /* note: if iface->src is a file not in uncompressed crl format,
     * then fd == -1
     */
    if (iface->eof) return; /* shouldn't happen */
    iface->eof = 1;
    if (!iface->src->type.is_live) {
	coral_offline_ifaces--;
	/* XXX coral_free_iface_blocks(cinst[src->id[i]]); */
	/* If there are no interfaces left, stop watching this source. */
	for (i = 0, remaining = 0; i < iface->src->iface_count; i++)
	    remaining += !cinst[iface->src->id[i]]->eof;
	if (!remaining)
	    coral_mark_eof_src(iface->src);
    } else if (iface->src->fd >= 0) {
	/* Devices only have one interface.  Stop watching this source. */
	coral_mark_eof_src(iface->src);
    }
    errno = saved_error;
}

/* Mark src as having reached EOF so we don't try to read from src->fd again,
 * but don't deallocate anything.  src may still have ifaces with blknode
 * queues that we've read but haven't given to the user yet.  Guranteed
 * to not modify errno.
 */
int coral_mark_eof_src(coral_source_t *src)
{
    int i;

    if (src->eof) return 0;
    src->eof = 1;
    if (!src->type.is_live) {
	coral_noselect_count--;
	for (i = 0; i < src->iface_count; i++) {
	    if (!cinst[src->id[i]]->have_data && !cinst[src->id[i]]->eof) {
		cinst[src->id[i]]->eof = 1;
		coral_offline_ifaces--;
	    }
	}
    } else if (src->fd >= 0) {
	if (!FD_ISSET(src->fd, &coral_select_fds)) {
	    coral_diag(0, ("internal error: src %d already clear\n", src->id));
	} else {
	    FD_CLR(src->fd, &coral_select_fds);
	    coral_select_count--;
	}
	if (!cinst[src->id[0]]->have_data && !cinst[src->id[0]]->eof) {
	    cinst[src->id[0]]->eof = 1;
	}
    }
    return 0;	/* for compatibility with coral_src_type_t.stop */
}

/* coral_close
 * find associated fd, close it, remove it from the set of select fds.
 * free all associated cinst's.  free struct.
 */
int coral_close(coral_source_t *src)
{
    int ret = 0;
    int i, id;

    validate_src(src, "coral_close", -1);
    coral_diag(4, ("coral_close %s\n", src->name));

    if (src->type.close)
	ret = src->type.close(src);

    if (src->fd >= 0) {
	close(src->fd);
    }
    coral_mark_eof_src(src);

    for (i = 0; i < src->iface_count; i++) {
	if ((id = src->id[i]) < 0) continue;

	if (cinst[id]->iface_type && cinst[id]->iface_type->close)
	    cinst[id]->iface_type->close(cinst[id]);

#ifdef HAVE_LIBPCAP
	if (cinst[id]->src->type.coral_type != CORAL_TYPE_PCAP &&
	    cinst[id]->src->type.coral_type != CORAL_TYPE_PCAPLIVE && 
	    cinst[id]->internal_pcap)
	{
# ifdef HAVE_PCAP_OPEN_DEAD
	    pcap_close(cinst[id]->internal_pcap);
# else
	    free(cinst[id]->internal_pcap);
# endif
	}
	cinst[id]->internal_pcap = NULL;

	if (cinst[id]->user_pcap) {
# ifdef HAVE_PCAP_OPEN_DEAD
	    pcap_close(cinst[id]->user_pcap);
# else
	    free(cinst[id]->user_pcap);
# endif
	}
	cinst[id]->user_pcap = NULL;

	if (cinst[id]->bpfprog && cinst[id]->bpfprog_is_private)
	    free(cinst[id]->bpfprog);
	if (cinst[id]->packetbuf.is_dynamic) {
	    free((void*)cinst[id]->packetbuf.buf);
	    cinst[id]->packetbuf.is_dynamic = 0;
	}
#endif

	if (!cinst[id]->eof) {
	    cinst[id]->eof = 1;	/* foil any future (mis)use of iface */
	    if (!cinst[id]->src->type.is_live)
		coral_offline_ifaces--;
	}

	cinst[id]->src = NULL;	/* foil any future (mis)use of iface */
	free(cinst[id]);
	cinst[id] = NULL;
    }

    if (src->desc) free(src->desc);
    if (src->encoding) free(src->encoding);
    if (src->name) free(src->name);
    if (src->filename) free(src->filename);
    csource[src->sd] = NULL;
    src->sd = -1;	/* foil any future (mis)use of this */
    free(src);
    coral_config.source_count--;

#if 0
# define print_stat(field) \
    coral_diag(0, (#field "\t%" PRId64 "\n", coral_internal_stats.field))

    print_stat(clock_order);
    print_stat(clock_double);
    print_stat(clock_sec_nsec);
    coral_internal_stats.clock_order = 0;
    coral_internal_stats.clock_double = 0;
    coral_internal_stats.clock_sec_nsec = 0;
#endif
    return ret;
}

/* generic close function for block sources */
int coral_blk_close(coral_source_t *src)
{
    int i, id;
    coral_blknode_t *nextbn;

    for (i = 0; i < src->iface_count; i++) {
	if ((id = src->id[i]) < 0) continue;
	nextbn = cinst[id]->u.blk.node.next;
	if (cinst[id]->src->type.release)
	    cinst[id]->src->type.release(cinst[id]);
	assert(!nextbn || src->type.release);
	while (nextbn) {
	    cinst[id]->u.blk.node = *nextbn;
	    free(nextbn);
	    nextbn = cinst[id]->u.blk.node.next;
	    src->type.release(cinst[id]);
	}
    }
    return 0;
}

int coral_close_all(void)
{
    int n = 0;	    /* number of closed sources */
    int sd;

    for (sd = 0; sd < CORAL_MAXOPEN; sd++) {
	if (!csource[sd]) continue;
	if (coral_close(csource[sd]) < 0)
	    return -1;
	n++;
    }
    return n;
}

void coral_format_iomode(char *buffer, const coral_io_mode_t *iomode)
{
    char *s = buffer;

    if (iomode->flags & CORAL_RX_UNKNOWN) {
	s += sprintf(s, "unspecified,");
    } else {
	if (iomode->flags & CORAL_RX_USER_ALL) {
	    s += sprintf(s, "user,");
	} else {
	    s += sprintf(s, "first=%d,", iomode->first_n);
	    if (iomode->flags & CORAL_RX_LAST)
		s += sprintf(s, "last,");
	}
	if (iomode->flags & CORAL_TX)
	    s += sprintf(s, "tx,");
	if (iomode->flags & CORAL_RX)
	    s += sprintf(s, "rx,");
	if (iomode->flags & CORAL_RX_IDLE)
	    s += sprintf(s, "idle,");
	if (iomode->flags & CORAL_RX_OAM)
	    s += sprintf(s, "ctrl,");
	if (iomode->flags & CORAL_RX_ECHO)
	    s += sprintf(s, "echo,");
    }
    *--s = '\0';
}

#ifndef cstrcmp
/* case-insensitive strcmp() */
int cstrcmp(const char *s, const char *t)
{
    register int diff = 0;
    while ((*s || *t) && !(diff = tolower(*s) - tolower(*t))) s++, t++;
    return diff;
}
#endif

const coral_protolabel_t coral_protolabel[] = {
#define defproto(layer, prefix, abbr, proto_ok, id, desc) \
    { coral_mkproto(layer, id), #abbr, desc, proto_ok },
#include "coral_proto.h"
    { 0, NULL, NULL }
};

const coral_protolabel_t *coral_protolabel_by_name(const char *name)
{
    const coral_protolabel_t *pl;

    if (!name) return NULL;

    for (pl = coral_protolabel; pl->abbr; pl++) {
	if (cstrcmp(pl->abbr, name) == 0 || cstrcmp(pl->desc, name) == 0)
	    return pl;
    }

    return NULL;
}

coral_protocol_t coral_proto_id(const char *name)
{
    const coral_protolabel_t *pl;

    if (!name) return CORAL_PROTO_UNKNOWN;

    for (pl = coral_protolabel; pl->abbr; pl++) {
	if (cstrcmp(pl->abbr, name) == 0 || cstrcmp(pl->desc, name) == 0)
	    return pl->id;
    }

    return CORAL_PROTO_UNKNOWN;
}

const char *coral_proto_desc(coral_protocol_t protocol)
{
    const coral_protolabel_t *pl;

    for (pl = coral_protolabel; pl->abbr; pl++) {
	if (pl->id == protocol)
	    return pl->desc;
    }

    return NULL;
}

const char *coral_proto_abbr(coral_protocol_t protocol)
{
    const coral_protolabel_t *pl;

    for (pl = coral_protolabel; pl->abbr; pl++) {
	if (pl->id == protocol)
	    return pl->abbr;
    }

    return NULL;
}

const char *coral_proto_str(coral_protocol_t protocol)
{
    static char scratch[32];
    const char *result;

    result = coral_proto_abbr(protocol);
    if (result) return result;
    sprintf(scratch, "unknown %d:%d",
	coral_proto_layer(protocol), coral_proto_num(protocol));
    return scratch;
}

const char *coral_file_version(const coral_source_t *src)
{
    static char buffer[32];
    char *p = buffer;
    u_int version = CORAL_VERSION;

    if (src) {
	validate_src(src, "coral_file_version", NULL);
	version = src->version;
    }

    if (CORAL_VERSION_IS_NEW(version)) {
        p += sprintf(p, "%d ", CORAL_VERSION_VALUE(version));
    } else {
        switch(version) {
        case CORAL_VERSION_OLD_ANY:   p += sprintf(p, "0 ");       break;
        case CORAL_VERSION_OLD_NLANR: p += sprintf(p, "0.nlanr "); break;
        case CORAL_VERSION_OLD_MCI:   p += sprintf(p, "0.mci ");   break;
        }
    }
    sprintf(p, "(%#x)", version);
    return buffer;
}

void coral_dump(const coral_source_t *src)
{
    int i;
    coral_iface_t *iface;
    coral_iface_info_t *iinfo;
    char iomode_buffer[64];
    long tzoff;

    coral_printf("%s:\n", src->name);
    coral_printf("  type:     %s\n", src->type.name);
    if (src->type.coral_type == CORAL_TYPE_FILE) {
	coral_printf("  version:  %s\n", coral_file_version(src));
	if (src->desc)
	    coral_printf("  comment:  %s\n", src->desc);
	if (src->encoding)
	    coral_printf("  encoding: %s  version %u\n",
		src->encoding, src->encoding_version);
    }

    coral_format_iomode(iomode_buffer, &src->dev_config.iomode);
    coral_printf("  iomode:   %s\n", iomode_buffer);

    for (i = 0; i < src->iface_count; i++) {
        iface = cinst[src->id[i]];
	iinfo = &iface->iface_info;
	coral_printf("  interface %d\n", i);
	coral_printf("    timestamps:  %s endian, %snormalized\n",
	    iface->iface_info.time_is_le ? "little" : "big",
	    iface->time_is_normal ? "" : "not ");
	coral_printf("    hardware:  ");
	switch (iinfo->hw_type) {
	    case CORAL_TYPE_FATM:	coral_printf("fatm"); break;
	    case CORAL_TYPE_POINT:	coral_printf("point"); break;
	    case CORAL_TYPE_DAG:	coral_printf("dag"); break;
	    default: coral_printf("(unknown: %#x)", iinfo->hw_type);
	}
	coral_printf("  version %#x\n", iinfo->hw_version);
	coral_printf("    software:  ");
	switch (iinfo->sw_type) {
	    case SW_TYPE_NONE:	coral_printf("none\n"); break;
	    case SW_TYPE_MCI:	coral_printf("MCI\n"); break;
	    case SW_TYPE_NLANR:	coral_printf("NLANR\n"); break;
	    case SW_TYPE_DAG:	coral_printf("dagtools\n"); break;
	    case SW_TYPE_CORAL:
		coral_printf("Coral version %d\n",
		    iinfo->sw_version & 0x7FFFFFFFul);
		break;
	    case SW_TYPE_PCAP:
		coral_printf("pcap version %d.%d\n",
		    iinfo->sw_version >> 16, iinfo->sw_version & 0xFFFF);
		break;
	    default: coral_printf("(unknown: %#x)", iinfo->sw_type);
	}
	if (iinfo->physical != CORAL_PROTO_UNKNOWN)
	    coral_printf("    physical:  %s\n", coral_dlt_str(iinfo->physical));
	coral_printf("    data link: %s\n", coral_dlt_str(iinfo->datalink));
	if (iinfo->capture_tv.tv_sec) {
	    char scratch[32];
	    struct tm *tm;
	    time_t tv_sec = iinfo->capture_tv.tv_sec;
	    tm = gmtime(&tv_sec); /* timeval.tv_sec may not be a time_t */
	    strftime(scratch, sizeof(scratch), "%Y-%m-%d %H:%M:%S", tm);
	    coral_printf("    captured:  %s.%06ld UTC (%ld.%06ld)\n",
		scratch, iinfo->capture_tv.tv_usec,
		iinfo->capture_tv.tv_sec, iinfo->capture_tv.tv_usec);
	}

	if (iinfo->tzoff < 0) {
	    coral_printf("    tz offset: -");
	    tzoff = -iinfo->tzoff;
	} else {
	    coral_printf("    tz offset: ");
	    tzoff = iinfo->tzoff;
	}
	coral_printf("%02d:%02d", tzoff / 3600, tzoff % 3600 / 60);
	if (tzoff % 60)
	    coral_printf(":%02d", tzoff % 60);
	coral_puts("");

	coral_format_iomode(iomode_buffer, &iinfo->iomode);
	coral_printf("    iomode:    %s\n", iomode_buffer);
	if (iinfo->bandwidth)
	    coral_printf("    bandwidth: %s\n",
		coral_bandwidth_fmt(iinfo->bandwidth));
	if (iface->proto_rules.head) {
	    coral_printf("    iface proto rules:\n");
	    coral_dump_rules(iface->proto_rules.head, 4);
	}
    }
    if (src->proto_rules.head) {
	coral_printf("  src proto rules:\n");
	coral_dump_rules(src->proto_rules.head, 4);
    }
    coral_printf("\n");
}

void coral_dump_all(void)
{
    int sd;

    for (sd = 0; sd < CORAL_MAXOPEN; sd++) {
	if (!csource[sd]) continue;
	coral_dump(csource[sd]);
    }
    if (coral_proto_rules.head) {
	coral_printf("  global proto rules:\n");
	coral_dump_rules(coral_proto_rules.head, 4);
    }
}

void coral_stats(const coral_source_t *src)
{
    int i, id;

/*  coral_printf("lost %d.09d seconds\n",
	src->devinfo.lost_time.tv_sec,
	src->devinfo.lost_time.tv_nsec)); */
    coral_printf("coral source %d (%s):\n", src->sd, src->name);
    coral_printf("  fd     %d\n", src->fd);
    coral_printf("  ubase  %p\n", src->ubase);
    coral_printf("  interface       kmem  tx err  rx err    intr\n");
    for (i = 0; i < src->iface_count; i++) {
        id = src->id[i];
#ifdef HAVE_CORAL_DEVICES
	if (coral_type_is_driver(cinst[id]->devinfo.type)) {
	    if (ioctl(src->fd, CORAL_FILLINFO, &cinst[id]->devinfo) < 0) {
		coral_diag(0, ("coral_stats: ioctl FILLINFO: %s: %s\n",
		    src->name, strerror(errno)));
	    }
	}
#endif
	coral_printf("  %3d (%3d) %10p  %6d  %6d  %6d\n", i, id,
	    cinst[id]->devinfo.kmem_buffer,
	    cinst[id]->devinfo.tx_errors,
	    cinst[id]->devinfo.rx_errors,
	    cinst[id]->devinfo.interrupts);
    }
}

void coral_stats_all(void)
{
    int sd;

    for (sd = 0; sd < CORAL_MAXOPEN; sd++) {
	if (!csource[sd]) continue;
	coral_stats(csource[sd]);
    }
}

char *coral_base(coral_source_t *src)
{
    coral_iface_t *iface;
    validate_src(src, "coral_base", NULL);
    iface = cinst[src->id[0]];
    return (coral_type_is_driver(iface->devinfo.type)) ?
	src->ubase :
	0;
}

size_t coral_msize(coral_source_t *src)
{
    coral_iface_t *iface;
    validate_src(src, "coral_msize", -1);
    iface = cinst[src->id[0]];
    return (coral_type_is_driver(iface->devinfo.type)) ?
	iface->devinfo.num_blks * iface->devinfo.blk_size :
	0;
}

#if 0
char *coral_fd(coral_source_t *src)
{
    validate_src(src, "coral_fd", NULL);
    return src->type.get_fd ? src->type.get_fd() : src->fd;
}
#endif

/* expand "~" in src, copy at most len characters into dst.  If dst is NULL,
 * coral_filename allocates len characters for it.  Returns expanded buffer.
 */
char *coral_filename(const char *str, char *dst, size_t len)
{
    const char *dir = NULL, *user;

    if (!dst && !(dst = malloc(len)))
	return NULL;

    if (*str != '~') {
	strncpy(dst, str, len);
	return dst;
    }
    for (user = ++str; *str && *str != '/'; str++);
    if (str == user) {
	dir = getenv("HOME");
#ifdef HAVE_GETPWNAM
    } else if (str - user < len) {
	struct passwd *pw;
	strncpy(dst, user, str - user);
	if ((pw = getpwnam(dst)))
	    dir = pw->pw_dir;
#endif /* HAVE_GETPWNAM */
    }
    if (!dir) {
	strncpy(dst, --user, len);
    } else {
	snprintf(dst, len, "%s%s", dir, str);
    }
    return dst;
}

