/* 
 * Copyright 1999, 2000, 2001 The Regents of the University of California 
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

/*
 *
 * $Id: coral_pkt.c,v 1.110.2.4 2001/11/21 01:33:48 kkeys Exp $
 *
 */

#include "config.h"
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#ifdef HAVE_LIBPCAP
# include <pcap.h>
#endif

#include "libcoral.h"
#include "libcoral_priv.h"
#include "ethernet.h"

static const char RCSid[] = "$Id: coral_pkt.c,v 1.110.2.4 2001/11/21 01:33:48 kkeys Exp $";
 
static pkt_info_t pi[1];
static const struct timeval tvzero = { 0, 0 };
static const struct timespec tszero = { 0, 0 };

static void reset_stats(coral_pkt_stats_t *stats)
{
    stats->l2_recv = 0;
    stats->l2_drop = 0;
    stats->pkts_recv = 0;
    stats->pkts_drop = 0;
    stats->truncated = 0;
    stats->driver_corrupt = 0;
    stats->too_many_vpvc = 0;
    stats->buffer_overflow = 0;
    stats->aal5_trailer = 0;
    stats->ok_packet = 0;
}

static void reset_iface_stats(coral_iface_t *iface)
{
    reset_stats(&iface->pkt_stats);
}

static void accum_iface_stats(coral_iface_t *iface)
{
    if (iface->iface_type->update_pkt_stats)
	iface->iface_type->update_pkt_stats(iface);

#define accum_and_msg(field, fmt) \
    do { \
	if (iface->pkt_stats.field) \
	    coral_diag(1, ("warning: iface %d: " fmt "\n", \
		iface->id, iface->pkt_stats.field)); \
	pi->stats.field += iface->pkt_stats.field; \
    } while (0)

    pi->stats.l2_recv		+= iface->pkt_stats.l2_recv;
    pi->stats.pkts_recv		+= iface->pkt_stats.pkts_recv;
    pi->stats.ok_packet		+= iface->pkt_stats.ok_packet;
    pi->stats.truncated		+= iface->pkt_stats.truncated;

    accum_and_msg(l2_drop,	    "dropped %" PRIu64 " layer 2 PDUs");
    accum_and_msg(pkts_drop,	    "dropped %" PRIu64 " packets");
    accum_and_msg(driver_corrupt,   "%" PRIu64 " driver corruptions");
    accum_and_msg(too_many_vpvc,    "%" PRIu64 " VPVC table overflows");
    accum_and_msg(buffer_overflow,  "%" PRIu64 " buffer overflows");
    accum_and_msg(aal5_trailer,	    "%" PRIu64 " bad AAL5 trailers");
}

const coral_pkt_stats_t *coral_get_iface_stats(coral_iface_t *iface)
{
    return &iface->pkt_stats;
}

/* Apply func to each iface */
static void coral_iface_loop(void (*func)(coral_iface_t *iface))
{
    int i;

    for (i = 0; i < coral_max_iface; i++) {
	/* XXX is this really necessary? */
	if (cinst[i] && (!pi->iface || cinst[i] == pi->iface) &&
	    (!pi->src || cinst[i]->src == pi->src))
	{
	    func(cinst[i]);
	}
    }
}

volatile int coral_pkt_done;


static void begin_interval(pkt_info_t *pip,
    coral_pkt_result_t *pkt_result, coral_interval_result_t *int_result)
{
    reset_stats(&pip->stats);
    coral_iface_loop(reset_iface_stats);
    if (pip->state == PKTST_NEW) {
	/* Begin first interval */
	pip->begin = pip->itime;
    } else {
	/* Begin new interval at end of last interval */
	pip->begin = pip->end;
    }
    pip->end = pip->begin;
    timespecadd(&pip->end, &pip->interval);

    pip->state = PKTST_BEGAN_INT;

    pkt_result->timestamp = NULL;
    pkt_result->packet = NULL;
    pkt_result->header = NULL;
    pkt_result->trailer = NULL;

    if (int_result) {
	int_result->begin.tv_sec = pip->begin.tv_sec;
	int_result->begin.tv_usec = pip->begin.tv_nsec / 1000;
	int_result->end.tv_sec = 0;
	int_result->end.tv_usec = 0;
	int_result->stats = NULL;
    }
}

static void end_interval(pkt_info_t *pip,
    coral_pkt_result_t *pkt_result, coral_interval_result_t *int_result)
{
    pip->state = PKTST_ENDED_INT;

    pkt_result->timestamp = NULL;
    pkt_result->packet = NULL;
    pkt_result->header = NULL;
    pkt_result->trailer = NULL;

    coral_iface_loop(accum_iface_stats);
    if (int_result) {
	int_result->begin.tv_sec = pip->begin.tv_sec;
	int_result->begin.tv_usec = pip->begin.tv_nsec / 1000;
	int_result->end.tv_sec = pip->end.tv_sec;
	int_result->end.tv_usec = pip->end.tv_nsec / 1000;
	int_result->stats = &pip->stats;
    }

    coral_diag(8, ("interval end: %u.%09u\n",
	pip->begin.tv_sec, pip->begin.tv_nsec));
}


static void check_iomode_USERALL(const coral_iface_t *iface)
{
    const coral_io_mode_t *iomode = &iface->iface_info.iomode;
    if (!(coral_config.flags & CORAL_OPT_PARTIAL_PKT ||
	iomode->flags & CORAL_RX_USER_ALL))
    {
	coral_diag(1, ("warning: coral_read_pkts (%s): interface %d "
	    "may not include all cells of every packet\n",
	    iface->src->name, iface->id));
    }
}

static void check_src_iomode_USERALL(const coral_source_t *src)
{
    int j;

    for (j = 0; j < CORAL_MAXOPEN; j++) {
	if (src->id[j] >= 0)
	    check_iomode_USERALL(cinst[src->id[j]]);
    }
}

/* Consume the timestamp and packet for iface, and try to update the timestamp
 * and packet using data already stored on iface.  Returns 1 if it was able
 * to update the TIMESTAMP, 0 if not.
 */
static inline int coral_consume_pkt(coral_iface_t *iface)
{
    if (iface->pkt_result.packet) {
	iface->pkt_result.timestamp = NULL;
	iface->pkt_result.packet = NULL;
	iface->pkt_result.header = NULL;
	iface->pkt_result.trailer = NULL;
	if (iface->iface_type->consume_pkt)
	    iface->iface_type->consume_pkt(iface);
    }
    return coral_consume(iface, CORAL_WANT_PKT);
}

/* coral_read_pkt()
 * This is necessarily hairy because it must keep state between calls and
 * pick up where it left off.
 */
coral_iface_t *coral_read_pkt(
    coral_pkt_result_t *pkt_result, coral_interval_result_t *int_result)
{
    coral_iface_t *iface;
    int i, missing, live_missing;
    fd_set rfds;    /* fd's that are readable */
    fd_set mfds;    /* fd's that are missing data */
    fd_set bfds;    /* fd's that have buffered data */
    int iface_count;
    /* TODO?: shouldn't time sort if no OPT_SORT_TIME, duration, or intervals */

    if (coral_config.max_pkts && coral_config.pkts >= coral_config.max_pkts)
	goto done; /* EOF */
start:	/* control returns here after filtering a packet or duration expired */
    iface_count = coral_select_count + coral_offline_ifaces;
    if (!iface_count)
	goto done; /* EOF */
    missing = 0;
    live_missing = 0;

    if (pi->state == PKTST_CONSUME) {	/* most common case */
	iface = NULL;
	if ((missing = !coral_consume_pkt(pi->last_iface))) {
	    FD_ZERO(&mfds);
	    if ((live_missing = pi->last_iface->src->type.is_live))
		FD_SET(pi->last_iface->src->fd, &mfds);
	}
    } else if (pi->state == PKTST_NEW) {
	iface = NULL;
	missing = pi->src ? pi->src->iface_count : pi->iface ? 1 : iface_count;
	live_missing = coral_select_count;
	coral_copy_fd_set(coral_select_fd_max, &mfds, &coral_select_fds);
    } else if (pi->state == PKTST_BEGAN_INT) {
	iface = pi->last_iface;
	missing = 0;
	live_missing = 0;
    } else if (pi->state == PKTST_ENDED_INT) {
	begin_interval(pi, pkt_result, int_result);
	return pi->last_iface;
    } else /* PKTST_DONE */ {
	goto done;
    }

reload:
    while (pi->state <= PKTST_CONSUME) {
	if (coral_pkt_done) goto done;
	/* Read until every iface has data or non-buffered timeout */
	while (missing) {
	    FD_ZERO(&bfds);
	    /* check live interfaces for buffered data */
	    if (live_missing) {
		for (i = 0; i < coral_max_iface; i++) {
		    if (!cinst[i] || !FD_ISSET(cinst[i]->src->fd, &mfds))
			continue;
		    if (cinst[i]->src->type.has_buf &&
			cinst[i]->src->type.has_buf(cinst[i]->src))
		    {
			FD_SET(cinst[i]->src->fd, &bfds);
			FD_CLR(cinst[i]->src->fd, &mfds);
			live_missing--;
		    }
		}
	    }
	    if (live_missing) {
		int n;
		struct timeval zero = { 0, 0 }, one = { 1, 0 };
		struct timeval tv, *tvp;
		/* A pcaplive iface will only select as readable after its
		 * in-kernel bpf buffer fills, but we can read even if the
		 * buffer isn't full; if there's _anything_ in the buffer we'll
		 * get it, otherwise it will timeout.  So we try to read it
		 * at least once per second, for near-human-realtime results.
		 */

		/* If we already have some readable ifaces, don't block. */
		if (missing > live_missing)
		    tvp = &zero;
		else
		    tvp = calculate_timeout(&one, &tv);

		coral_diag(24, ("selecting on %d live ifaces\n", live_missing));
		coral_copy_fd_set(coral_select_fd_max, &rfds, &mfds);
		n = select(coral_select_fd_max+1, &rfds, NULL, NULL, tvp);
		if (n < 0) {
		    coral_diag(1, ("select: %s\n", strerror(errno)));
		    return NULL;
		}
		coral_diag(24, ("select returned %d\n", n));
		if (realtime_duration_ended(NULL) &&
		    !(n || missing > live_missing))
		{
		    /* Nothing WAS readable, but coral drivers are now stopped,
		     * so they will return their last blks immediately. */
		    goto start;
		}
		coral_or_fd_set(coral_select_fd_max+1, &rfds, &bfds);
	    } else {
		coral_copy_fd_set(coral_select_fd_max+1, &rfds, &bfds);
	    }
	    missing = coral_freshen_iface(pi->src, pi->iface, CORAL_WANT_PKT,
		&rfds, &mfds, &live_missing);
	    if (missing < 0)
		return NULL;
	} /* while (missing) */
	/*missing = 0; FD_ZERO(&mfds);*/

	/* For each iface without a packet, repeatedly prep_pkt and consume
	 * until we get a packet or run out of data.  If we run out of
	 * data, we must return to the top of the outer loop to get more
	 * (after gathering all ifaces that don't have packets).
	 */
	for (i = 0; i < coral_max_iface; i++) {
	    if (!cinst[i] || cinst[i]->eof || cinst[i]->pkt_result.packet)
		continue;
	    while (1) {
		assert(cinst[i]->have_data || !cinst[i]->src->type.is_buffered);
		if (!cinst[i]->have_data) {
		    missing++;
		    break;
		}
		if (cinst[i]->iface_type->prep_pkt(cinst[i]) < 0) {
		    coral_mark_eof(cinst[i]);
		    if (!errno) errno = -1;
		    return NULL;
		}
		if (cinst[i]->pkt_result.packet)
		    break;
		if (!coral_consume(cinst[i], CORAL_WANT_PKT)) {
		    missing++;
		    if (cinst[i]->src->type.is_live) {
			live_missing++;
			FD_SET(cinst[i]->src->fd, &mfds);
		    }
		    /* If an unbuffered source has no data, it really has no
		     * data; but a buffered source may just be buffering its
		     * data, so we must read the buffer to be sure we're not
		     * missing an early packet. */
		    if (cinst[i]->src->type.is_buffered)
			goto reload;
		    break;
		}
	    }
	}

	if (!coral_select_count && !coral_offline_ifaces)
	    goto done; /* EOF */

	iface = !pi->iface ? coral_find_earliest_iface(pi->src) :
	    pi->iface->have_data ? pi->iface : NULL;

	if (iface) {
	    coral_config.pkts++;
	    pi->itime = iface->ts;
	    pi->last_iface = iface;
	    /* data timestamp duration should've been tested in read_min */
	    assert(iface->duration_end.tv_sec < 0 ||
		timespeccmp(&pi->itime, &iface->duration_end, <=));
	    break;
	}
    }

    if (pi->state == PKTST_NEW) {
	/* Begin first interval.  Even if caller didn't ask for intervals, we
	 * must still init the stats; we just don't return an int_result. */
	begin_interval(pi, pkt_result, int_result);
	if (timespeccmp(&pi->interval, &tszero, >)) {
	    return iface;
	} else {
	    pi->end.tv_sec = INT_MAX;
	    pi->end.tv_nsec = 999999999;
	}

    } else if (timespeccmp(&pi->itime, &pi->end, >=)) {
	/* End interval */
	coral_diag(8, ("interval end at pkt %" PRIu64 "\n",
	    coral_config.pkts-1));
	end_interval(pi, pkt_result, int_result);
	return iface;
    }

    pi->state = PKTST_CONSUME;

    assert (iface->pkt_result.packet);

#ifdef HAVE_BPF_FILTER
    if (/* there's a filter */
	iface->bpfprog &&
	/* packet has not already passed filter */
	!iface->pkt_result.packet->passed &&
	/* packet fails filter */
	!coral_pkt_filter(iface, iface->pkt_result.packet, NULL))
    {
	/* Packet did not pass filter. */
	coral_diag(20, ("pkt %" PRIu64 " rejected by filter\n", coral_config.pkts-1));
	/* Go to func start as if we'd returned and been called again. */
	goto start;
    }
#endif
    *pkt_result = iface->pkt_result;

    if ((iface->iface_info.datalink == CORAL_DLT_ETHER ||
	iface->iface_info.datalink == CORAL_DLT_IEEE802) &&
	pkt_result->subiface == 0)
	/* subiface==0 is normally redundant, but is needed by crl_guess */
    {
	struct ether_header *ether_hdr;
	uint16_t ethertype;

	ether_hdr = (struct ether_header*)pkt_result->packet->buf;
	ethertype = ntohs(ether_hdr->ether_type);

	if (ethertype == ETHERTYPE_VLAN) {
	    /* IEEE 802.1Q VLAN */
	    uint16_t tci; /* tag control information */
	    tci = ntohs(*(uint16_t*)(pkt_result->packet->buf +
		sizeof(*ether_hdr)));
	    pkt_result->subiface = tci & 0x0FFF;    /* vlan id */
	} else {
	    pkt_result->subiface = 0;
	}
    }

    return iface;
	
done:
    if (pi->state < PKTST_DONE && pi->last_iface) {
	/* End last interval.  If caller didn't ask for intervals, we still
	 * calculate the stats (which may print a warning), treating the entire
	 * run as one interval; we just never return an int_result. */
	coral_diag(8, ("input end at pkt %" PRIu64 "\n", coral_config.pkts-1));
	end_interval(pi, pkt_result, int_result);
	pi->state = PKTST_DONE;	    /* next time: end */
	if (timespeccmp(&pi->interval, &tszero, >)) {
	    int_result->end.tv_sec = pi->itime.tv_sec;
	    int_result->end.tv_usec = pi->itime.tv_nsec / 1000;
	    return pi->last_iface;
	}
    }

    pkt_result->timestamp = NULL;
    pkt_result->packet = NULL;
    pkt_result->header = NULL;
    pkt_result->trailer = NULL;

    if (timespeccmp(&pi->interval, &tszero, >)) {
	/* int_result->begin = NULL; */
	/* int_result->end = NULL; */
	int_result->stats = NULL;
    }

    errno = 0;
    return NULL;	/* EOF */
}

int coral_read_pkt_init(coral_source_t *src, coral_iface_t *iface,
    struct timeval *interval)
{
    int sd, i;

    coral_config.pkts = 0;

    if (iface) {
	validate_iface(iface, "coral_read_pkt_init", -1);
	check_iomode_USERALL(iface);
	pi->src = iface->src;
	pi->iface = iface;
	/* close all other ifaces on this source */
	for (i = 0; i < iface->src->iface_count; i++) {
	    if (iface->src->id[i] == iface->id) continue;
	    coral_mark_eof(cinst[iface->src->id[i]]);
	}

    } else if (src) {
	validate_src(src, "coral_read_pkt_init", -1);
	check_src_iomode_USERALL(src);
	pi->src = src;
	pi->iface = NULL;

    } else {
	pi->src = NULL;
	pi->iface = NULL;
	for (sd = 0; sd < coral_max_src; sd++) {
	    if ((src = csource[sd])) {
		validate_src(src, "coral_read_pkt_init", -1);
		check_src_iomode_USERALL(src);
	    }
	}
    }

    pi->last_iface = NULL;

    pi->pkthandler = NULL;
    pi->preinthandler = NULL;
    pi->postinthandler = NULL;

    normalize_if_needed(pi->src, pi->iface);

    if (interval) {
	pi->interval.tv_sec = interval->tv_sec;
	pi->interval.tv_nsec = interval->tv_usec * 1000;
    } else {
	pi->interval = tszero;
    }
    pi->parameter = NULL;

    reset_stats(&pi->stats);
    pi->begin.tv_sec = 0;
    pi->begin.tv_nsec = 0;
    pi->end.tv_sec = 0;
    pi->end.tv_nsec = 0;
    pi->cell_count = -1;
    pi->i = -1;
    pi->state = PKTST_NEW;
    coral_pkt_done = 0;

    return 0;
}

int coral_read_pkts(coral_source_t *src, coral_iface_t *iface,
			  coral_pkt_handler pkthandler, 
			  coral_pre_interval_handler preinthandler,
			  coral_post_interval_handler postinthandler,
			  struct timeval *interval,
			  void *param)
{
    coral_pkt_result_t pkt_result;
    coral_interval_result_t int_result;
    int result = 0;
    struct timeval start_time, end_time;

    coral_read_pkt_init(src, iface, interval);
    pi->pkthandler = pkthandler;
    if (interval && timercmp(interval, &tvzero, >)) {
	pi->preinthandler = preinthandler;
	pi->postinthandler = postinthandler;
	pi->parameter = param;
    }

    gettimeofday(&start_time, NULL);

    while (!coral_pkt_done) {
	iface = coral_read_pkt(&pkt_result, &int_result);
	if (!iface) {
	    if (errno) result = -1;
	    break;
	} else if (pkt_result.packet) {
	    pkthandler(iface, pkt_result.timestamp, param,
		pkt_result.packet, pkt_result.header, pkt_result.trailer);
	} else if (!int_result.stats) {
	    if (preinthandler)
		preinthandler(iface, &int_result.begin, param);
	} else {
	    if (postinthandler)
		postinthandler(iface, &int_result.begin, &int_result.end,
		    param, int_result.stats);
	}
    }

    gettimeofday(&end_time, NULL);
#if 0 /* XXX */
    coral_diag(20, ("ATM reassembly time: %f\n",
	(end_time.tv_sec + end_time.tv_usec / 1000000.0) -
	(start_time.tv_sec + start_time.tv_usec / 1000000.0)));
#endif

    return result;
}

#if 0
int coral_pkt_select(int nfds, fd_set *rfds, fd_set *wfds, fd_set *efds,
    struct timeval *timeout)
{
    int i, count, total = 0, have_pkt = 0;
    coral_source_t *src;
    fd_set rwant, wwant, ewant;
    fd_set rsel, wsel, esel;
    fd_set mfds;

    coral_copy_fd_set(nfds, &rwant, rfds);
    coral_copy_fd_set(nfds, &wwant, wfds);
    coral_copy_fd_set(nfds, &ewant, efds);

    coral_copy_fd_set(coral_select_fd_max+1, &mfds, &coral_select_fds);
    coral_or_fd_set(coral_select_fd_max+1, &rwant, &mfds);

    FD_ZERO(rfds);
    FD_ZERO(wfds);
    FD_ZERO(efds);

    if (coral_pkt_poll()) {
	FD_CLR(coral_pkt_fd, &rwant);
	FD_SET(coral_pkt_fd, rfds);
	have_pkt = 1;
    }

    if (total > 0)
	timeout = &tvzero;

    do {
	coral_copy_fd_set(nfds, &rsel, &rwant);
	coral_copy_fd_set(nfds, &wsel, &wwant);
	coral_copy_fd_set(nfds, &esel, &ewant);

	if (!have_pkt) {
	    for (i = 0; i < coral_max_src; i++) {
		if (!csource[i]) continue;
		if (/* XXX csource[i] does not have a pkt */) {
		    FD_SET(&rsel, csource[i]->fd);
		    if (csource[i]->fd >= nfds)
			nfds = csource[i]->fd + 1;
		}
	    }
	}

	count = select(&rsel, &wsel, &esel, timeout);
	if (count < 0) /* error */
	    return count;

	if (count > 0) {
	    for (i = 0; i < nfds; i++) {
		if (FD_ISSET(i, &rsel)) {
		    src = coral_fd_table[i];
		    if (!src || coral_pkt_poll(src)) {
			total++;
			FD_CLR(i, &rwant);
			FD_SET(i, rfds);
		    }
		}
		if (FD_ISSET(i, &wsel)) {
		    total++;
		    FD_CLR(i, &wwant);
		    FD_SET(i, wfds);
		}
		if (FD_ISSET(i, &esel)) {
		    total++;
		    FD_CLR(i, &ewant);
		    FD_SET(i, efds);
		}
	    }
	}
    } while (!have_pkt && !total)

    return total + have_pkt;
}
#endif

#ifdef HAVE_BPF_FILTER
int coral_pcap_compile(coral_iface_t *iface, struct bpf_program *fp,
    char *str, int optimize, uint32_t netmask)
{
    char *buf = NULL;
    int result;

    if (!iface->internal_pcap)
	iface->internal_pcap = coral_iface_to_alloced_pcapp(iface, 0);
    if (!iface->internal_pcap) {
        coral_diag(0, ("%s\n", strerror(errno)));
        return -1;
    }

    if (iface->src->dev_config.vlan_hack) {
	buf = malloc(strlen(str) + strlen("vlan and ()") + 1);
	if (!buf) return -1;
	sprintf(buf, "vlan and (%s)", str);
	str = buf;
    }

    result = pcap_compile(iface->internal_pcap, fp, str, optimize, netmask);
    if (result < 0) {
        coral_diag(0, ("pcap expression on interface %d: %s\n",
	    iface->id, pcap_geterr(iface->internal_pcap)));
    }
    if (buf) free(buf);
    return result;
}

int coral_pcap_setfilter(coral_iface_t *iface, struct bpf_program *fp)
{
    if (!iface->internal_pcap)
	iface->internal_pcap = coral_iface_to_alloced_pcapp(iface, 0);
    if (!iface->internal_pcap) {
        coral_diag(0, ("%s\n", strerror(errno)));
        return -1;
    }

    if (iface->bpfprog && iface->bpfprog_is_private)
	free(iface->bpfprog);
    iface->bpfprog = fp;
    if (iface->src->type.coral_type == CORAL_TYPE_PCAP ||
        iface->src->type.coral_type == CORAL_TYPE_PCAPLIVE)
    {
	if (pcap_setfilter(iface->internal_pcap, fp) < 0) {
	    coral_diag(0, ("pcap_setfilter on interface %d: %s\n",
		iface->id, pcap_geterr(iface->internal_pcap)));
	    return -1;
	}
    }
    return 0;
}
#endif

