/* 
 * Copyright 1999, 2000, 2001 The Regents of the University of California 
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

/*
 *
 * $Id: coral_proto.c,v 1.25.2.2 2001/11/21 01:33:49 kkeys Exp $
 *
 */

#include "config.h"
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#ifdef HAVE_INET6
#include <netinet/ip6.h>
#endif
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <arpa/inet.h>

#include "libcoral.h"
#include "libcoral_priv.h"
#include "ethernet.h"
#include "ip_icmp.h"	/* XXX may not be portable */
#include "coral_lane.h"

/* PPP protocol numbers.  These are not portably defined in any standard
 * header, so we define them ourselves according to the RFCs.
 */
#define PPP_IP		0x0021
#define PPP_BRIDGED_LAN	0x0031  /* used at AIX */
#define PPP_IPv6	0x0057


static const char RCSid[] = "$Id: coral_proto.c,v 1.25.2.2 2001/11/21 01:33:49 kkeys Exp $";

/* stuff that may be missing from some systems' headers */
#if (IPPROTO_IPIP - 0 != 4)
# undef IPPROTO_IPIP
# define IPPROTO_IPIP 4
#endif
#if (IPPROTO_IPV6 - 0 != 41)
# undef IPPROTO_IPV6
# define IPPROTO_IPV6 41
#endif

#ifndef IP_OFFMASK
# define IP_OFFMASK 0x1fff               /* mask for fragmenting bits */
#endif
/* */

#define NLPID_SNAP	0x80
#define NLPID_IPv6	0x8E
#define NLPID_IP	0xCC
#define NLPID_PPP	0xCF

#define llc_test(llc, val)	X_llc_test(llc, val)
#define X_llc_test(llc, dsap, ssap, cntl) \
    ((llc)->llc_dsap == (dsap) && (llc)->llc_ssap == (ssap) && \
     (llc)->llc_cntl == (cntl))
#define LLC_HAS_SNAP	0xAA, 0xAA, 0x03
#define LLC_ROUTED_ISO	0xFE, 0xFE, 0x03
#define LLC_IEEE8021D	0x42, 0x42, 0x03  /* IEEE 802.1D bridge spanning tree */

#define snap_org_test(org, val)	X_snap_org_test(org, val)
#define X_snap_org_test(org, a, b, c) \
    (((uint8_t*)org)[0] == (a) && ((uint8_t*)org)[1] == (b) && \
	((uint8_t*)org)[2] == (c))
#define OUI_NONE	0x00, 0x00, 0x00
#define OUI_CISCO	0x00, 0x00, 0x0C
#define OUI_ATM_FORUM	0x00, 0xA0, 0x3E

#define LANE_MAX_LEN	(1536-1)


static inline int check_length(const coral_pkt_buffer_t *src,
    int stated_len, int actual_len)
{
    if ((src)->totlen >= 0) {
	/* Total length of src is known, do sanity check. */
	if (stated_len > actual_len) {
	    coral_diag(8, ("payload: stated_len (%d) > actual_len (%d)\n",
		stated_len, actual_len));
	    return 0;
	}
    }
    return 1;
}


#ifdef HAVE_VSNPRINTF
/* like snprintf(), execpt *buf and *len are modified. */
int crl_snprintf(char **buf, int *len, const char *fmt, ...)
{
    va_list ap;
    int n;

    va_start(ap, fmt);
    n = vsnprintf(*buf, *len, fmt, ap);
    va_end(ap);
    if (n < 0) {
	*len = 0;
    } else {
	*buf += n;
	*len -= n;
    }
    return n;
}

#else
int crl_snprintf(char **buf, int *len, const char *fmt, ...)
{
    if (*len > 0)
	**buf = '\0';
    return -1;
}
#endif

static coral_protocol_t ethertype_to_coral_proto(uint16_t ethertype)
{
    switch (ethertype) {
	case ETHERTYPE_IP:	    return CORAL_NETPROTO_IPv4;
	case ETHERTYPE_IPv6:	    return CORAL_NETPROTO_IPv6;
	case ETHERTYPE_ARP:	    return CORAL_NETPROTO_ARP;
	case ETHERTYPE_IPX:	    return CORAL_NETPROTO_IPX;
	case ETHERTYPE_APPLETALK:   return CORAL_NETPROTO_APPLETALK;
	case ETHERTYPE_AARP:	    return CORAL_NETPROTO_AARP;
	case ETHERTYPE_IPX_E:	    return CORAL_NETPROTO_IPX_E;
	case ETHERTYPE_PPPOES:	    return CORAL_DLT_PPPoES;
	case ETHERTYPE_PPPOED:	    return CORAL_DLT_PPPoED;
	default:		    return CORAL_PROTO_UNKNOWN;
    }
}

/* format only, no payload */
static int coral_get_atm_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    union atm_hdr h = *(union atm_hdr*)src->buf;
    h.ui = ntohl(h.ui);
    crl_snprintf(&dumpbuf, &buflen, "%d:%d",
	get_vpvc_vp(h.h.vpvc), get_vpvc_vc(h.h.vpvc));
    dst->buf = NULL;
    dst->caplen = 0;
    dst->totlen = 0;
    return 0;
}

/* format only, no payload */
static int coral_get_aal5_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    aal5_trailer_t *t = (aal5_trailer_t*)src->buf;
    crl_snprintf(&dumpbuf, &buflen, "len=%d", ntohs(t->length));
    dst->buf = NULL;
    dst->caplen = 0;
    dst->totlen = 0;
    return 0;
}

/* IEEE 802.2 (also RFC 1483/2684 LLC Encapsulation, RFC 1042) */
static int coral_get_llc_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    coral_llcsnap_t *llc;
    int skip = 3;

    if (src->caplen < skip)
	return CORAL_ELENGTH;
    llc = (coral_llcsnap_t *)src->buf;

    if (dumpbuf) {
	crl_snprintf(&dumpbuf, &buflen, "LLC: %02x-%02x-%02x",
	    llc->llc_dsap, llc->llc_ssap, llc->llc_cntl);
    }

    if (llc_test(llc, LLC_HAS_SNAP)) {
	uint16_t ethertype;
	skip += 5; /* size of SNAP */

	ethertype = ntohs(llc->snap_type);
	if (dumpbuf) {
	    crl_snprintf(&dumpbuf, &buflen,
		" SNAP: OUI=%02x-%02x-%02x type=%04x",
		llc->snap_org[0], llc->snap_org[1], llc->snap_org[2],
		ethertype);
	}

	if (snap_org_test(llc->snap_org, OUI_NONE)) {
	    dst->protocol = ethertype_to_coral_proto(ethertype);
	} else {
	    dst->protocol = CORAL_PROTO_UNKNOWN;
	}

    } else if (llc_test(llc, LLC_ROUTED_ISO)) {
	uint8_t nlpid = *(src->buf + skip);
	skip += 1; /* size of NLPID */

	switch (nlpid) {
	case NLPID_PPP:		/* RFC 2364: PPP over AAL5 */
	    dst->protocol = CORAL_DLT_PPP; break;
	case NLPID_IP:		/* shouldn't be used with ATM_RFC1483 */
	    dst->protocol = CORAL_NETPROTO_IPv4; break;
	case NLPID_IPv6:	/* shouldn't be used with ATM_RFC1483 */
	    dst->protocol = CORAL_NETPROTO_IPv6; break;
	default:
	    dst->protocol = CORAL_PROTO_UNKNOWN;
	}

    } else if (llc_test(llc, LLC_IEEE8021D)) {
	dst->protocol = CORAL_DLT_IEEE8021D;

    } else {
	dst->protocol = CORAL_PROTO_UNKNOWN;
    }

    dst->buf = src->buf + skip;
    dst->caplen = src->caplen - skip;
    dst->totlen = src->totlen - skip;
    return 0;
}

/* Cisco encapsulation over HDLC */
static int coral_get_chdlc_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    struct chdlc_header {
	uint8_t id;		/* 0f=unicast, 8f=broadcast */
	uint8_t control;	/* always 0 */
	uint16_t ethertype;	/* Ethernet type of the SDU */
    } *head;
    int skip = sizeof(struct chdlc_header);
    uint16_t ethertype;

    head = (struct chdlc_header *)src->buf;
    ethertype = ntohs(head->ethertype);

    if (dumpbuf) {
	if      (head->id == 0x0f) crl_snprintf(&dumpbuf, &buflen, "unicast");
	else if (head->id == 0x8f) crl_snprintf(&dumpbuf, &buflen, "broadcast");
	else crl_snprintf(&dumpbuf, &buflen, "id=%02x", head->id);
	crl_snprintf(&dumpbuf, &buflen, ", type=%04x", ethertype);
    }

    dst->buf = src->buf + skip;
    dst->caplen = src->caplen - skip;
    dst->totlen = src->totlen - skip;
    dst->protocol = ethertype_to_coral_proto(ethertype);
    return 0;
}

/* PPP Bridged LAN (RFC 1638, section 4.2) */
static int coral_get_bridgedlan_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    const uint8_t *ubuf;
    int skip;

    struct blan_header {
#ifdef WORDS_BIGENDIAN
	unsigned fcs_present:1;
	unsigned lanid_present:1;
	unsigned zero_filled:1;
	unsigned zero:1;
	unsigned pads:4;
#else
	unsigned pads:4;
	unsigned zero:1;
	unsigned zero_filled:1;
	unsigned lanid_present:1;
	unsigned fcs_present:1;
#endif
	uint8_t mac_type;
    } *blan;

    blan = (struct blan_header*)src->buf;
    ubuf = (uint8_t*)src->buf;

    if (dumpbuf)
	crl_snprintf(&dumpbuf, &buflen, "mac type %02x", blan->mac_type);
    if (blan->lanid_present) {
	if (dumpbuf)
	    crl_snprintf(&dumpbuf, &buflen, ", lan id %02x%02x%02x%02x",
		ubuf[2], ubuf[3], ubuf[4], ubuf[5]);
	skip = 6;
    } else {
	skip = 2;
    }

    dst->buf = src->buf + skip;
    dst->caplen = src->caplen - skip;
    dst->totlen = src->totlen - skip;
    dst->protocol = CORAL_PROTO_UNKNOWN;

    switch (blan->mac_type) {
    case 1:	/* IEEE 802.3/Ethernet with canonical addresses */
	dst->protocol = CORAL_DLT_IEEE802;  return 0;
    case 2:	/* IEEE 802.4 with canonical addresses */
    case 3:	/* IEEE 802.5 with noncanonical addresses */
    case 4:	/* FDDI with noncanonical addresses */
    case 10:	/* IEEE 802.5 with canonical addresses */
    case 11:	/* FDDI with canonical addresses */
	return 0;
    default:	/* illegal value */
	return CORAL_ESYNTAX;
    }
}

/* Point-to-Point Protocol (RFC 1661) */
static int coral_get_ppp_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    uint16_t pppproto;
    int skip;

    if (src->caplen < 2) {
	return CORAL_ELENGTH;
    } else if (src->buf[0] & 1) {
	/* first byte is odd; proto is 1 byte (Protocol-Field-Compression). */
	pppproto = *(uint8_t*)src->buf;
	skip = 1;
	if (dumpbuf)
	    crl_snprintf(&dumpbuf, &buflen, "proto %02" PRIx16, pppproto);
    } else if (src->buf[1] & 1) {
	/* first byte is even, second is odd; protocol is 2 bytes. */
	pppproto = ntohs(*(uint16_t*)src->buf);
	skip = 2;
	if (dumpbuf)
	    crl_snprintf(&dumpbuf, &buflen, "proto %04" PRIx16, pppproto);
    } else {
	/* illegal */
	return CORAL_ESYNTAX;
    }

    switch (pppproto) {
	case PPP_IP:
	    dst->protocol = CORAL_NETPROTO_IPv4;
	    break;
	case PPP_IPv6:
	    dst->protocol = CORAL_NETPROTO_IPv6;
	    break;
	case PPP_BRIDGED_LAN:
	    dst->protocol = CORAL_DLT_BRIDGED_LAN;
	    break;
	default:
	    dst->protocol = CORAL_PROTO_UNKNOWN;
	    break;
    }

    dst->buf = src->buf + skip;
    dst->caplen = src->caplen - skip;
    dst->totlen = src->totlen - skip;

    return 0;
}

/* PPP in HDLC-like framing (RFC 1662) */
static int coral_get_pppohdlc_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    coral_pkt_buffer_t tmp;
    uint8_t addr, ctrl;

    addr = src->buf[0];
    ctrl = src->buf[1];
    tmp.protocol = CORAL_DLT_PPP;

    if (addr == 0xff && ctrl == 0x03) {
	/* normal frame */
	if (dumpbuf)
	    crl_snprintf(&dumpbuf, &buflen, "addr %02x, ctrl %02x, ",
		addr, ctrl);
	tmp.buf = src->buf + 2;
	tmp.caplen = src->caplen - 2;
	tmp.totlen = src->totlen - 2;
    } else {
	/* Address-and-Control-Field-Compression */
	tmp.buf = src->buf;
	tmp.caplen = src->caplen;
	tmp.totlen = src->totlen;
    }
    return coral_get_ppp_payload(&tmp, dst, dumpbuf, buflen);
}

/* PPP over Ethernet (RFC 2516) */
static int coral_get_pppoe_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    struct pppoe_header {
#ifdef WORDS_BIGENDIAN
	unsigned ver:4;
	unsigned type:4;
#else
	unsigned type:4;
	unsigned ver:4;
#endif
	uint8_t code;
	uint16_t session_id;
	uint16_t length;
    } *pppoe;
    int skip = sizeof(*pppoe);
    uint16_t len;

    pppoe = (struct pppoe_header*)src->buf;
    if (dumpbuf)
	crl_snprintf(&dumpbuf, &buflen, "ver=%01x type=%01x code=%02x ",
	    pppoe->ver, pppoe->type, pppoe->code);
    if (pppoe->ver != 1 || pppoe->type != 1 || pppoe->code != 0)
	return CORAL_ESYNTAX;
    len = ntohs(pppoe->length);
    if (dumpbuf)
	crl_snprintf(&dumpbuf, &buflen, "session_id=%04x len=%d",
	    ntohs(pppoe->session_id), len);
    dst->buf = src->buf + skip;
    dst->caplen = src->caplen - skip;
    dst->totlen = src->totlen - skip;
    dst->protocol = CORAL_PROTO_UNKNOWN;
    if (dst->caplen > len)
	dst->caplen = len;
    if (dst->totlen > len)
	dst->totlen = len;
    return 0;
}

static int coral_get_pppoe_sess_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    int result;
    if ((result = coral_get_pppoe_payload(src, dst, dumpbuf, buflen)) < 0)
	return result;
    dst->protocol = CORAL_DLT_PPP;
    return 0;
}

static int coral_get_pppoe_disc_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    return coral_get_pppoe_payload(src, dst, dumpbuf, buflen);
}

/* for systems without inet_aton() (inet_addr() is insufficient) */
/* caller needs to include <sys/types.h> and <sys/socket.h> */
int coral_inet_pton(int af, const char *src, void *dst)
{
    unsigned long octet;
    int i = 0;
    char *end;

    switch (af) {
    case AF_INET:
	while (1) {
	    if (!isdigit(*src)) return 0;
	    if ((octet = strtoul(src, &end, 10)) > 255) return 0;
#ifdef WORDS_BIGENDIAN
	    ((unsigned char*)dst)[i] = octet;
#else
	    ((unsigned char*)dst)[3-i] = octet;
#endif
	    if (++i == 4) break;
	    if (*end != '.') return 0;
	    src = end+1;
	}
	return (!*end);

#ifdef HAVE_INET6 /* Assume any system with HAVE_INET6 will have inet_pton() */
    case AF_INET6:
	return inet_pton(af, src, dst);
#endif

    default:
	return 0;
    }
}

/* ATM Forum NSAPA format */
static inline int format_nsap_addr(char **buf, int *len,
    const unsigned char *addr)
{
    return crl_snprintf(buf, len,
	"%02X.%02X%02X.%02X.%02X%02X%02X."
	"%02X%02X.%02X%02X.%02X%02X.%02X%02X.%02X%02X.%02X%02X.%02X",
	addr[0], addr[1], addr[2], addr[3],
	addr[4], addr[5], addr[6], addr[7],
	addr[8], addr[9], addr[10], addr[11],
	addr[12], addr[13], addr[14], addr[15],
	addr[16], addr[17], addr[18], addr[19],
	addr[20], addr[21], addr[22], addr[23],
	addr[24], addr[25], addr[26], addr[27],
	addr[28], addr[29], addr[30], addr[31],
	addr[32], addr[33], addr[34], addr[35],
	addr[36], addr[37], addr[38], addr[39]);
}

/* Ethernet */
static inline int format_ether_addr(char **buf, int *len,
    const unsigned char *ether_addr)
{
    /* ftp://ftp.isi.edu/in-notes/iana/assignments/ethernet-numbers */
    return crl_snprintf(buf, len, "%02X-%02X-%02X-%02X-%02X-%02X",
	ether_addr[0], ether_addr[1], ether_addr[2],
	ether_addr[3], ether_addr[4], ether_addr[5]);
}

/* Ethernet */
static inline int format_ether_addrs(char **buf, int *len,
    struct ether_header *ether_hdr)
{
    if (crl_snprintf(buf, len, "dst=") < 0) return -1;
    if (format_ether_addr(buf, len, ether_hdr->ether_dhost) < 0) return -1;
    if (crl_snprintf(buf, len, " src=") < 0) return -1;
    if (format_ether_addr(buf, len, ether_hdr->ether_shost) < 0) return -1;
    if (crl_snprintf(buf, len, " ") < 0) return -1;
    return 0;
}

/* Ethernet */
static int coral_get_ether_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    struct ether_header *ether_hdr;
    uint16_t ethertype;
    int skip = sizeof(struct ether_header);
    uint32_t vlan_id = 0;
    uint8_t cfi = 0, pri = 0;

    ether_hdr = (struct ether_header*)src->buf;
    ethertype = ntohs(ether_hdr->ether_type);
    if (dumpbuf) {
	format_ether_addrs(&dumpbuf, &buflen, ether_hdr);
    }

    if (ethertype == ETHERTYPE_VLAN) {
	/* IEEE 802.1Q VLAN */
	uint16_t tci; /* tag control information */

	tci = ntohs(((uint16_t*)(src->buf + skip))[0]);
	ethertype = ntohs(((uint16_t*)(src->buf + skip))[1]);
	skip += 4;
	cfi = (tci >> 12) & 0x0001;
	pri = (tci >> 13) & 0x0003;
	vlan_id = tci & 0x0FFF;
	if (dumpbuf) {
	    crl_snprintf(&dumpbuf, &buflen, "VLAN %u ", vlan_id);
	    if (pri)
		crl_snprintf(&dumpbuf, &buflen, "pri=%u ", pri);
	    if (cfi)
		crl_snprintf(&dumpbuf, &buflen, "cfi=%u ", cfi);
	}
    }

    if (dumpbuf)
	crl_snprintf(&dumpbuf, &buflen, "type=%04x", ethertype);

    dst->protocol = cfi ? CORAL_PROTO_UNKNOWN :
	ethertype_to_coral_proto(ethertype);
    dst->buf = src->buf + skip;
    dst->caplen = src->caplen - skip;
    dst->totlen = src->totlen - skip;
    return 0;
}

/* IEEE 802.3 */
static int coral_get_ieee8023_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    struct ether_header *ether_hdr;
    uint16_t len;
    coral_pkt_buffer_t tmp;

    ether_hdr = (struct ether_header*)src->buf;
    len = ntohs(ether_hdr->ether_type);
    if (len == 0) {
	/* This shouldn't happen in true 802.3, but can happen in LANE */
	if (src->totlen < 0)
	    return CORAL_ELENGTH;
	tmp.totlen = src->totlen - sizeof(*ether_hdr);
    } else {
	tmp.totlen = len;
	check_length(src, len, tmp.totlen);
    }
    if (dumpbuf) {
	format_ether_addrs(&dumpbuf, &buflen, ether_hdr);
	crl_snprintf(&dumpbuf, &buflen, "len=%u", len);
    }

    tmp.buf = src->buf + sizeof(*ether_hdr);
    tmp.caplen = src->caplen - sizeof(*ether_hdr);
    if (tmp.caplen > len)   /* caplen may have included padding */
	tmp.caplen = len;
    if (dumpbuf)
	crl_snprintf(&dumpbuf, &buflen, ", ");
    return coral_get_llc_payload(&tmp, dst, dumpbuf, buflen);
}

/* IEEE 802.3 or DIX Ethernet */
static int coral_get_ieee8023_or_ether_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    struct ether_header *ether_hdr;
    uint16_t typelen;

    ether_hdr = (struct ether_header*)src->buf;
    typelen = ntohs(ether_hdr->ether_type);
    if (typelen > ETHERMTU) {
	return coral_get_ether_payload(src, dst, dumpbuf, buflen);
    } else {
	return coral_get_ieee8023_payload(src, dst, dumpbuf, buflen);
    }
}

/* af-lane-0084.000 */
static int coral_get_lane_ieee8023_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    lane_ieee8023_t *lane;
    uint16_t typelen;
    coral_pkt_buffer_t tmp;

    if (src->caplen < sizeof(lane_ieee8023_t))
	return CORAL_ELENGTH;
    lane = (lane_ieee8023_t *)src->buf;
    typelen = ntohs(lane->typelen);
    tmp.buf = src->buf + sizeof(lane->le_header);
    tmp.caplen = src->caplen - sizeof(lane->le_header);
    tmp.totlen = src->totlen - sizeof(lane->le_header);

    if (dumpbuf)
	crl_snprintf(&dumpbuf, &buflen, "lecid=%04d ", lane->le_header);

    if (typelen > LANE_MAX_LEN) {
	return coral_get_ether_payload(&tmp, dst, dumpbuf, buflen);
    } else {
	return coral_get_ieee8023_payload(&tmp, dst, dumpbuf, buflen);
    }
}

/* af-lane-0084.000 */
static int coral_get_lane_llc_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    lane_llc_t *lane;
    int skip = sizeof(lane_llc_t);
    coral_pkt_buffer_t tmp;

    if (src->caplen < skip)
	return CORAL_ELENGTH;
    lane = (lane_llc_t *)src->buf;
    if (!llc_test(&lane->llc, LLC_HAS_SNAP) ||
	!snap_org_test(lane->llc.snap_org, OUI_ATM_FORUM))
    {
	return CORAL_ESYNTAX;
    }
    tmp.buf = src->buf + skip;
    tmp.caplen = src->caplen - skip;
    tmp.totlen = src->totlen - skip;

    switch (ntohs(lane->llc.snap_type)) {
	case LANETYPE_IEEE8023:
	    return coral_get_lane_ieee8023_payload(&tmp, dst, dumpbuf, buflen);
	case LANETYPE_IEEE8025:
#if 0	    /* not implemented */
	    return coral_get_lane_ieee8025_payload(&tmp, dst, dumpbuf, buflen);
#else
	    *dst = tmp;
	    dst->protocol = CORAL_DLT_LANE_IEEE8025;
	    return 0;
#endif
	default:
	    return CORAL_ESYNTAX;
    }
}

/* DEC Gigaswitch encapsulation?
 * All I know about this is that it seems to have a 16 byte header, a 4 byte
 * trailer, and (at least in the limited samples I've seen) always contains
 * ATM_RFC1483.
 */
static int coral_get_gigx_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    int skip = 16;
    coral_pkt_buffer_t tmp;

    if (src->caplen < skip)
	return CORAL_ELENGTH;
    tmp.buf = src->buf + skip;
    tmp.totlen = src->totlen - skip - 4;
    tmp.caplen = src->caplen - skip;
    if (tmp.totlen >= 0 && tmp.caplen > tmp.totlen)
	tmp.caplen = tmp.totlen;

    return coral_get_llc_payload(&tmp, dst, dumpbuf, buflen);
}

#define fits(ptr, field, len) \
    ((size_t)&((ptr)->field) - (size_t)(ptr) + sizeof((ptr)->field) <= (len))


static void coral_fmt_pro_addr(char **dumpbuf, int *buflen, uint16_t pro,
    const void *addr, uint8_t len)
{
    switch (pro) {
    case ETHERTYPE_IP:
      {
	struct in_addr in_addr;
	memcpy(&in_addr, addr, sizeof(in_addr));
	crl_snprintf(dumpbuf, buflen, "%s", inet_ntoa(in_addr));
	break;
      }
#ifdef HAVE_INET6
    case ETHERTYPE_IPv6:
      {
	struct in6_addr in6_addr;
	char addrbuf[INET6_ADDRSTRLEN+1];
	memcpy(&in6_addr, addr, sizeof(in6_addr));
	crl_snprintf(dumpbuf, buflen, "%s",
	    inet_ntop(AF_INET6, (char*)&in6_addr, addrbuf, INET6_ADDRSTRLEN));
	break;
      }
#endif
    default:
      {
        int i;
	crl_snprintf(dumpbuf, buflen, "%04x ", pro);
	for (i = 0; i < len; i++)
	    crl_snprintf(dumpbuf, buflen, "%02x", ((unsigned char*)addr)[i]);
	break;
      }
    }
}

/* http://www.iana.org/assignments/arp-parameters */
#define ARPHRD_ETHER	 1	/* RFC 826 */
#define ARPHRD_IEEE802	 6
#define ARPHRD_FRELAY	15
/*#define ARPHRD_ATM	16*/
#define ARPHRD_ATM	19	/* RFC 2225 */
/*#define ARPHRD_ATM	21*/

#define ATMARP_LEN_MASK 0x3F
#define ATMARP_TYPE_MASK 0x40

/* RFC 826 */
static int coral_get_arp_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    uint16_t op = 0, hrd = 0, pro = 0;
    int i;
    const unsigned char *p;
    coral_protocol_t coral_proto;
    char proto_name_buf[5];
    const char *proto_name;

    dst->protocol = CORAL_PROTO_UNKNOWN;
    dst->caplen = 0;
    dst->totlen = 0;
    dst->buf = NULL;

    if (!dumpbuf) return 0;

    if (src->caplen >= 2)
	hrd = ntohs(((uint16_t*)src->buf)[0]);

    if (src->caplen >= 4) {
	pro = ntohs(((uint16_t*)src->buf)[1]);
	coral_proto = ethertype_to_coral_proto(pro);
    } else {
	coral_proto = CORAL_PROTO_UNKNOWN;
    }
    if (coral_proto == CORAL_PROTO_UNKNOWN) {
	proto_name = proto_name_buf;
	sprintf(proto_name_buf, "%04x", pro);
    } else {
	proto_name = coral_proto_abbr(coral_proto);
    }

    if (src->caplen >= 8)
	op = ntohs(((uint16_t*)src->buf)[3]);

    /* http://www.iana.org/assignments/arp-parameters */
    switch (op) {
    case 1: crl_snprintf(&dumpbuf, &buflen, "Request ");   break; /* RFC 826 */
    case 2: crl_snprintf(&dumpbuf, &buflen, "Reply ");	   break; /* RFC 826 */
    case 8: crl_snprintf(&dumpbuf, &buflen, "InRequest "); break; /* RFC 1293 */
    case 9: crl_snprintf(&dumpbuf, &buflen, "InReply ");   break; /* RFC 1293 */
    case 10: crl_snprintf(&dumpbuf, &buflen, "NAK ");	   break; /* RFC 2225 */
    default: crl_snprintf(&dumpbuf, &buflen, "op=%d ", op); break;
    }

    switch (hrd) {
    case ARPHRD_ETHER: /* RFC 826, 1293 */
    case ARPHRD_IEEE802:
      {
	struct etherarp {
	    uint16_t hrd; /* hardware addr space */
	    uint16_t pro; /* protocol addr space */
	    uint8_t hln; /* hardware addr len */
	    uint8_t pln; /* protocol addr len */
	    uint16_t op; /* opcode */
	    /* source ethernet addr */
	    /* source protocol addr */
	    /* target ethernet addr */
	    /* target protocol addr */
	} *arp;

	arp = (struct etherarp*)src->buf;
	if (!dumpbuf) return 0;
	crl_snprintf(&dumpbuf, &buflen, "ETHER/%s ", proto_name);

	p = (const unsigned char *)src->buf + sizeof(struct etherarp);

	/* source hardware address */
	if ((const char *)p + arp->hln > src->buf + src->caplen) break;
	crl_snprintf(&dumpbuf, &buflen, "from ");
	format_ether_addr(&dumpbuf, &buflen, p);
	p += arp->hln;
	crl_snprintf(&dumpbuf, &buflen, " ");
	/* source protocol address */
	if ((const char *)p + arp->pln > src->buf + src->caplen) break;
	coral_fmt_pro_addr(&dumpbuf, &buflen, pro, p, arp->pln);
	p += arp->pln;


	/* target hardware address */
	if ((const char *)p + arp->hln > src->buf + src->caplen) break;
	crl_snprintf(&dumpbuf, &buflen, " to ");
	format_ether_addr(&dumpbuf, &buflen, p);
	p += arp->hln;
	crl_snprintf(&dumpbuf, &buflen, " ");
	/* target protocol address */
	if ((const char *)p + arp->pln > src->buf + src->caplen) break;
	coral_fmt_pro_addr(&dumpbuf, &buflen, pro, p, arp->pln);
	p += arp->pln;

	break;
      }
    
    case ARPHRD_ATM: /* RFC 2225 */
      {
	struct atmarp {
	    uint16_t hrd; /* hardware addr space */
	    uint16_t pro; /* protocol addr space */
	    uint8_t shtl; /* source ATM addr len */
	    uint8_t sstl; /* source ATM subaddr len */
	    uint16_t op; /* opcode */
	    uint8_t spln; /* source protocol addr len */
	    uint8_t thtl; /* target ATM addr len */
	    uint8_t tstl; /* target ATM subaddr len */
	    uint8_t tpln; /* target protocol addr len */
	    /* source ATM addr */
	    /* source ATM subaddr */
	    /* source protocol addr */
	    /* target ATM addr */
	    /* target ATM subaddr */
	    /* target protocol addr */
	} *arp;

	arp = (struct atmarp*)src->buf;
	if (!dumpbuf) return 0;
	crl_snprintf(&dumpbuf, &buflen, "ATM/%s ", proto_name);

#if 0 /* debugging */
	crl_snprintf(&dumpbuf, &buflen, "shtl=%d ", arp->shtl);
	crl_snprintf(&dumpbuf, &buflen, "sstl=%d ", arp->sstl);
	crl_snprintf(&dumpbuf, &buflen, "spln=%d ", arp->spln);
	crl_snprintf(&dumpbuf, &buflen, "thtl=%d ", arp->thtl);
	crl_snprintf(&dumpbuf, &buflen, "tstl=%d ", arp->tstl);
	crl_snprintf(&dumpbuf, &buflen, "tpln=%d ", arp->tpln);
#endif

	p = (const unsigned char *)src->buf + sizeof(struct atmarp);

	/* source hardware address */
	if ((const char *)p + arp->shtl > src->buf + src->caplen) break;
	crl_snprintf(&dumpbuf, &buflen, "from ");
	if (arp->shtl & ATMARP_LEN_MASK) {
	    if (!(arp->shtl & ATMARP_TYPE_MASK)) {
		format_nsap_addr(&dumpbuf, &buflen, p);
		p += arp->shtl & ATMARP_LEN_MASK;
	    } else {
		for (i = 0; i < (arp->shtl & ATMARP_LEN_MASK); i++, p++)
		    crl_snprintf(&dumpbuf, &buflen, "%02x", *p);
	    }
	    /* source hardware subaddress */
	    if ((const char *)p + arp->sstl > src->buf + src->caplen) break;
	    if (arp->sstl & ATMARP_LEN_MASK) {
		crl_snprintf(&dumpbuf, &buflen, "[");
		for (i = 0; i < (arp->sstl & ATMARP_LEN_MASK); i++, p++)
		    crl_snprintf(&dumpbuf, &buflen, "%02x", *p);
		crl_snprintf(&dumpbuf, &buflen, "]");
	    }
	    crl_snprintf(&dumpbuf, &buflen, " ");
	}
	/* source protocol address */
	if ((const char *)p + arp->spln > src->buf + src->caplen) break;
	if (arp->spln) {
	    coral_fmt_pro_addr(&dumpbuf, &buflen, pro, p, arp->spln);
	    p += arp->spln;
	    crl_snprintf(&dumpbuf, &buflen, " ");
	}

	/* target hardware address */
	if ((const char *)p + arp->thtl > src->buf + src->caplen) break;
	crl_snprintf(&dumpbuf, &buflen, "to ");
	if (arp->thtl & ATMARP_LEN_MASK) {
	    if (!(arp->thtl & ATMARP_TYPE_MASK)) {
		format_nsap_addr(&dumpbuf, &buflen, p);
		p += arp->thtl & ATMARP_LEN_MASK;
	    } else {
		for (i = 0; i < (arp->thtl & ATMARP_LEN_MASK); i++, p++)
		    crl_snprintf(&dumpbuf, &buflen, "%02x", *p);
	    }
	/* target hardware subaddress */
	    if ((const char *)p + arp->tstl > src->buf + src->caplen) break;
	    if (arp->tstl & ATMARP_LEN_MASK) {
		crl_snprintf(&dumpbuf, &buflen, "[");
		for (i = 0; i < (arp->tstl & ATMARP_LEN_MASK); i++, p++)
		    crl_snprintf(&dumpbuf, &buflen, "%02x", *p);
		crl_snprintf(&dumpbuf, &buflen, "]");
	    }
	    crl_snprintf(&dumpbuf, &buflen, " ");
	}
	/* target protocol address */
	if ((const char *)p + arp->tpln > src->buf + src->caplen) break;
	if (arp->tpln) {
	    coral_fmt_pro_addr(&dumpbuf, &buflen, pro, p, arp->tpln);
	    p += arp->tpln;
	}
	break;
      }

    default:
	crl_snprintf(&dumpbuf, &buflen, "hardware type %04x", hrd);
	break;
    }

    return 0;
}

/* RFC 791 */
static int coral_get_ip4_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    struct ip *ip;
    int i, hdr_len;
    uint16_t ip_len, ip_off;
    uint8_t *opt, len;

    if (src->caplen < 1)
	return CORAL_ELENGTH;
    ip = (struct ip *)src->buf;

    if (!fits(ip, ip_len, src->caplen))
	return CORAL_ELENGTH;
    if (ip->ip_v == 4) {
	if (ip->ip_hl < 5)	/* impossible */
	    return CORAL_ESYNTAX;
	hdr_len = 4 * ip->ip_hl;
	ip_len = ntohs(ip->ip_len);

	dst->caplen = src->caplen - hdr_len;
	dst->buf = src->buf + hdr_len;
	dst->totlen = (src->totlen >= 0 && ip_len > src->totlen) ?
	    src->totlen - hdr_len : ip_len - hdr_len;
	if (dst->totlen >= 0 && dst->caplen > dst->totlen)
	    dst->caplen = dst->totlen;	/* caplen included padding */
	dst->protocol = CORAL_PROTO_UNKNOWN;

	if (fits(ip, ip_off, src->caplen)) {
	    ip_off = ntohs(ip->ip_off);
	    if (!(ip_off & IP_OFFMASK)) {
		switch (ip->ip_p) {
		case IPPROTO_IPIP:  dst->protocol = CORAL_NETPROTO_IPv4; break;
		case IPPROTO_IPV6:  dst->protocol = CORAL_NETPROTO_IPv6; break;
		default:	    dst->protocol = coral_mkproto(4, ip->ip_p);
		}
	    }
	}
	if (dumpbuf) {
	    crl_snprintf(&dumpbuf, &buflen, "hl=%uB", hdr_len);
	    if (!fits(ip, ip_tos, src->caplen)) goto ip_print_trunc;
	    if (ip->ip_tos)
		crl_snprintf(&dumpbuf, &buflen, " tos=%#02x", ip->ip_tos);
	    if (!fits(ip, ip_len, src->caplen)) goto ip_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, " len=%u", ntohs(ip->ip_len));
	    if (!fits(ip, ip_id, src->caplen)) goto ip_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, " id=%u", ntohs(ip->ip_id));
	    if (!fits(ip, ip_off, src->caplen)) goto ip_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, "%s%s",
		ip_off & IP_DF ? " DF" : "", ip_off & IP_MF ? " MF" : "");
	    if ((ip_off & IP_MF) || (ip_off & IP_OFFMASK))
		crl_snprintf(&dumpbuf, &buflen, " off=%uB",
		    (ip_off & IP_OFFMASK) * 8);
	    if (!fits(ip, ip_ttl, src->caplen)) goto ip_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, " ttl=%u", ip->ip_ttl);
	    if (!fits(ip, ip_p, src->caplen)) goto ip_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, " p=%u", ip->ip_p);
	    if (coral_config.flags & CORAL_OPT_IP_CKSUM) {
		if (!fits(ip, ip_sum, src->caplen)) goto ip_print_trunc;
		crl_snprintf(&dumpbuf, &buflen, " sum=%04x", ntohs(ip->ip_sum));
		if (hdr_len <= src->caplen) {
		    unsigned long cksum = coral_in_cksum_add(0, ip, hdr_len);
		    crl_snprintf(&dumpbuf, &buflen, "(%s)",
			(coral_in_cksum_result(cksum) == 0) ? "pass" : "fail");
		}
	    }
	    if (!fits(ip, ip_src, src->caplen)) goto ip_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, " %s", inet_ntoa(ip->ip_src));
	    if (!fits(ip, ip_dst, src->caplen)) goto ip_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, " > %s", inet_ntoa(ip->ip_dst));

	    if (hdr_len > src->caplen) hdr_len = src->caplen;
	    hdr_len -= sizeof(*ip);
	    for (opt = (uint8_t*)src->buf + sizeof(*ip); hdr_len >= 2; ) {
		if (*opt == IPOPT_EOL)
		    break;
		if (*opt == IPOPT_NOP) {
		    opt++; hdr_len--; continue;
		}
		if ((len = opt[1]) < 2) { /* error */
		    crl_snprintf(&dumpbuf, &buflen, " ?");
		    break;
		}
#if 0 /* debugging */
		crl_snprintf(&dumpbuf, &buflen, " ");
		for (i = 0; i < len && i < hdr_len; i++)
		    crl_snprintf(&dumpbuf, &buflen, "%02x", opt[i]);
#endif
		switch (*opt) {
		case IPOPT_RR:
		case IPOPT_LSRR:
		case IPOPT_SSRR:
		    crl_snprintf(&dumpbuf, &buflen, "; %sRR",
			(*opt == IPOPT_LSRR) ? "LS" :
			(*opt == IPOPT_SSRR) ? "SS" :
			""
		    );
		    if (*opt != IPOPT_RR)
			crl_snprintf(&dumpbuf, &buflen, " seen");
		    for (i = 3; i + 4 <= opt[2] && i + 4 <= hdr_len; i += 4) {
			crl_snprintf(&dumpbuf, &buflen, " %d.%d.%d.%d",
			    opt[i], opt[i+1], opt[i+2], opt[i+3]);
		    }
		    if (*opt != IPOPT_RR) {
			crl_snprintf(&dumpbuf, &buflen, " next");
			for (i = (opt[2] - 3) / 4 * 4 + 3;
			    i + 4 <= len && i + 4 <= hdr_len; i += 4)
			{
			    crl_snprintf(&dumpbuf, &buflen, " %d.%d.%d.%d",
				opt[i], opt[i+1], opt[i+2], opt[i+3]);
			}
		    }
		    break;
		default:
		    crl_snprintf(&dumpbuf, &buflen, "; OPT %u%s", opt[0],
			len > 2 ? "=0x" : "");
		    for (i = 2; i < len && i < hdr_len; i++)
			crl_snprintf(&dumpbuf, &buflen, "%02x", opt[i]);
		    break;
		}
		opt += len;
		hdr_len -= len;
	    }

	    goto ip_print_done;

	    ip_print_trunc:
	    crl_snprintf(&dumpbuf, &buflen, " (truncated at %d bytes)",
		src->caplen);
	}
	ip_print_done:

	return 0;

    } else {
	if (dumpbuf)
	    crl_snprintf(&dumpbuf, &buflen, "v%u", ip->ip_v);
	return CORAL_ESYNTAX;
    }
}

#ifdef HAVE_INET6
/* RFC 2460 */
static int coral_get_ip6_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    struct ip6_hdr *ip6;
    int hdr_len;
    uint16_t plen;
    char addrbuf[INET6_ADDRSTRLEN+1];

    if (src->caplen < 1)
	return CORAL_ELENGTH;
    ip6 = (struct ip6_hdr *)src->buf;

    if ((ip6->ip6_vfc >> 4) == 6) {
	plen = ntohs(ip6->ip6_plen);

	if (dumpbuf) {
	    if (!fits(ip6, ip6_flow, src->caplen)) goto ip6_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, "tclass=%#02x ",
		(ntohl(ip6->ip6_flow) & 0x0ff00000) >> 20);
	    crl_snprintf(&dumpbuf, &buflen, "label=%#05x ",
		ntohl(ip6->ip6_flow) & 0x000fffff);
	    if (!fits(ip6, ip6_plen, src->caplen)) goto ip6_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, "plen=%u ", plen);
	    if (!fits(ip6, ip6_nxt, src->caplen)) goto ip6_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, "nxt=%u ", ip6->ip6_nxt);
	    if (!fits(ip6, ip6_hops, src->caplen)) goto ip6_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, "hops=%u ", ip6->ip6_hops);
	    if (!fits(ip6, ip6_src, src->caplen)) goto ip6_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, " %s",
		inet_ntop(AF_INET6, (char*)&ip6->ip6_src, addrbuf,
		    INET6_ADDRSTRLEN));
	    if (!fits(ip6, ip6_dst, src->caplen)) goto ip6_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, " > %s",
		inet_ntop(AF_INET6, (char*)&ip6->ip6_dst, addrbuf,
		    INET6_ADDRSTRLEN));

	    goto ip6_print_done;

	    ip6_print_trunc:
	    crl_snprintf(&dumpbuf, &buflen, " (truncated at %d bytes)",
		src->caplen);
	}
	ip6_print_done:

	if (!fits(ip6, ip6_nxt, src->caplen)) {
	    dst->protocol = CORAL_PROTO_UNKNOWN;
	} else {
	    switch (ip6->ip6_nxt) {
	    case IPPROTO_IPIP:	dst->protocol = CORAL_NETPROTO_IPv4; break;
	    case IPPROTO_IPV6:	dst->protocol = CORAL_NETPROTO_IPv6; break;
	    default:		dst->protocol = coral_mkproto(4, ip6->ip6_nxt);
	    }
	}

	hdr_len = sizeof(struct ip6_hdr);

	dst->caplen = src->caplen - hdr_len;
	dst->buf = src->buf + hdr_len;
	dst->totlen = src->totlen >= 0 && hdr_len + plen > src->totlen ?
	    src->totlen - hdr_len : plen;
	if (dst->totlen >= 0 && dst->caplen > dst->totlen)
	    dst->caplen = dst->totlen;	/* caplen included padding */

	return 0;

    } else {
	if (dumpbuf)
	    crl_snprintf(&dumpbuf, &buflen, "v%u", ip6->ip6_vfc >> 4);
	return CORAL_ESYNTAX;
    }
}

/* RFC 2460, 2402, 2406 */
static int coral_get_ipexthdr_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    int hdr_len, len, i, j;
    const uint8_t *p;
    uint8_t nxt;
    uint16_t offset;

    if (src->caplen < 2)
	return CORAL_ELENGTH;
    p = (uint8_t*)src->buf;

    if (src->protocol == CORAL_IPPROTO_ESP) {	/* encap security payload */
	dst->protocol = CORAL_PROTO_UNKNOWN;
	dst->buf = src->buf + 8;
	dst->caplen = src->caplen - 8;
	dst->totlen = src->totlen - 8;
	if (dumpbuf) {
	    crl_snprintf(&dumpbuf, &buflen, " spi=%08x",
		ntohl(*(uint32_t*)(p+0)));
	    crl_snprintf(&dumpbuf, &buflen, " seq=%08x",
		ntohl(*(uint32_t*)(p+4)));
	}

    } else {
	/* All other headers start with 8-bit next_hdr and 8-bit ext_hdr_len */
	nxt = p[0];
	hdr_len = (p[1] + 1) * 8;
	len = hdr_len <= src->caplen ? hdr_len : src->caplen;
	dst->protocol = coral_mkproto(4, nxt);

	switch (src->protocol) {
	case CORAL_IPPROTO_HOPOPTS:	/* hop-by-hop options */
	    if (dumpbuf) {
		for (i = 2; i < len; i += (p[i] == 0) ? 1 : 2 + p[i+1]) {
		    switch(p[i]) {
		    case 0:
			crl_snprintf(&dumpbuf, &buflen, "pad1");
			break;
		    case 1:
			crl_snprintf(&dumpbuf, &buflen, "padN %d", p[i+1]);
			break;
		    default:
			crl_snprintf(&dumpbuf, &buflen, "opt %d: ", p[i]);
			for (j = i+2; j < i + 2 + p[i+1]; j++)
			    crl_snprintf(&dumpbuf, &buflen, "%02x ", p[j]);
		    }
		    crl_snprintf(&dumpbuf, &buflen, "; ");
		}
	    }
	    break;
	case CORAL_IPPROTO_DSTOPTS:	/* destination options */
	    if (dumpbuf) {
		for (i = 2; i < len; i += (p[i] == 0) ? 1 : 2 + p[i+1]) {
		    switch(p[i]) {
		    case 0:
			crl_snprintf(&dumpbuf, &buflen, "pad1");
			break;
		    case 1:
			crl_snprintf(&dumpbuf, &buflen, "padN %d", p[i+1]);
			break;
		    default:
			crl_snprintf(&dumpbuf, &buflen, "opt %d: ", p[i]);
			for (j = i+2; j < i + 2 + p[i+1]; j++)
			    crl_snprintf(&dumpbuf, &buflen, "%02x ", p[j]);
		    }
		    crl_snprintf(&dumpbuf, &buflen, "; ");
		}
	    }
	    break;
	case CORAL_IPPROTO_ROUTING:	/* routing header */
	    if (dumpbuf) {
		if (len < 3) goto print_done;
		crl_snprintf(&dumpbuf, &buflen, "type=%02x ", p[2]);
		if (len < 4) goto print_done;
		crl_snprintf(&dumpbuf, &buflen, "left=%02x ", p[3]);
		for (i = 4; i < len; i++)
		    crl_snprintf(&dumpbuf, &buflen, "%02x ", p[i]);
	    }
	    break;
	case CORAL_IPPROTO_FRAGMENT:	/* fragment header */
	    if (len < 4 || (offset = ntohs(*(uint16_t*)(p+2)))) /* !first */
		dst->protocol = CORAL_PROTO_UNKNOWN;
	    hdr_len = 8;
	    if (dumpbuf) {
		if (len < 4) goto print_done;
		crl_snprintf(&dumpbuf, &buflen, "off=%d ", offset >> 3);
		if (offset & 1) crl_snprintf(&dumpbuf, &buflen, " M");
		if (len < 8) goto print_done;
		crl_snprintf(&dumpbuf, &buflen, "id=%08x ",
		    ntohl(*(uint32_t*)(p+4)));
	    }
	    break;
	case CORAL_IPPROTO_AH:		/* authentication header */
	    if (dumpbuf) {
		if (len < 8) goto print_done;
		crl_snprintf(&dumpbuf, &buflen, "spi=%08x ",
		    ntohl(*(uint32_t*)(p+4)));
		if (len < 12) goto print_done;
		crl_snprintf(&dumpbuf, &buflen, "seq=%08x ",
		    ntohl(*(uint32_t*)(p+8)));
		crl_snprintf(&dumpbuf, &buflen, "data=");
		for (i = 12; i < len; i++)
		    crl_snprintf(&dumpbuf, &buflen, "%02x ", p[i]);
	    }
	    break;
	default:			/* unrecognized extension header */
	    return CORAL_ESYNTAX;
	}
	print_done:
	if (dumpbuf && len < hdr_len) {
	    crl_snprintf(&dumpbuf, &buflen, " (truncated at %d bytes)",
		src->caplen);
	}

	dst->buf = src->buf + hdr_len;
	dst->caplen = src->caplen - hdr_len;
	dst->totlen = src->totlen - hdr_len;
    }

    return 0;
}
#endif /* HAVE_INET6 */

/* RFC 793 */
static int coral_get_tcp_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    struct tcphdr *tcp;
    int hdr_len, i;
    const uint8_t *opt;
    uint8_t len;

    tcp = (struct tcphdr*)src->buf;

    hdr_len = (src->caplen < 13) ? sizeof(*tcp) : (tcp->th_off * 4);
    if (hdr_len < sizeof(*tcp))
	return CORAL_ELENGTH;

    dst->totlen = src->totlen - hdr_len;
    dst->caplen = src->caplen - hdr_len;
    dst->buf = src->buf + hdr_len;
    dst->protocol = CORAL_PROTO_UNKNOWN;

    if (dumpbuf) {
	if (!fits(tcp, th_sport, src->caplen)) goto tcp_print_trunc;
	crl_snprintf(&dumpbuf, &buflen, "src=%u ", ntohs(tcp->th_sport));
	if (!fits(tcp, th_sport, src->caplen)) goto tcp_print_trunc;
	crl_snprintf(&dumpbuf, &buflen, "dst=%u ", ntohs(tcp->th_dport));
	if (!fits(tcp, th_seq, src->caplen)) goto tcp_print_trunc;
	crl_snprintf(&dumpbuf, &buflen, "seq=%lu ", ntohl(tcp->th_seq));
	if (src->caplen < 14) goto tcp_print_trunc;
	crl_snprintf(&dumpbuf, &buflen, "off=%luB ", tcp->th_off * 4);
	crl_snprintf(&dumpbuf, &buflen, "%s%s%s%s",
	    tcp->th_flags & TH_FIN ? "FIN," : "",
	    tcp->th_flags & TH_SYN ? "SYN," : "",
	    tcp->th_flags & TH_RST ? "RST," : "",
	    tcp->th_flags & TH_PUSH ? "PSH," : "");
	*--dumpbuf = '\0';
	buflen++;
	if (tcp->th_flags & TH_ACK) {
	    crl_snprintf(&dumpbuf, &buflen, " ack=%lu", ntohl(tcp->th_ack));
	    if (!fits(tcp, th_win, src->caplen)) goto tcp_print_trunc;
	    crl_snprintf(&dumpbuf, &buflen, " win=%u", ntohs(tcp->th_win));
	}
	if (tcp->th_flags & TH_URG) {
	    if (!fits(tcp, th_urp, src->caplen))
		crl_snprintf(&dumpbuf, &buflen, " URG");
	    else
		crl_snprintf(&dumpbuf, &buflen, " urg=%u", ntohs(tcp->th_urp));
	}

	if (hdr_len > src->caplen) hdr_len = src->caplen;
	hdr_len -= sizeof(*tcp);
	for (opt = (uint8_t*)src->buf + sizeof(*tcp); hdr_len >= 2; ) {
	    if (*opt == TCPOPT_EOL)
		break;
	    if (*opt == TCPOPT_NOP) {
		opt++; hdr_len--; continue;
	    }
	    if ((len = opt[1]) < 2) { /* error */
		crl_snprintf(&dumpbuf, &buflen, " ?");
		break;
	    }
	    switch (*opt) {
	    case TCPOPT_MAXSEG: /* MSS */
		if (opt[1] == 4) {
		    crl_snprintf(&dumpbuf, &buflen, " MSS=%u",
			ntohs(*(uint16_t*)(opt+2)));
		    break;
		}
		/* fall through */
	    default:
		crl_snprintf(&dumpbuf, &buflen, " OPT %u%s", opt[0],
		    opt[1] > 2 ? "=0x" : "");
		for (i = 2; i < opt[1] && i < hdr_len; i++)
		    crl_snprintf(&dumpbuf, &buflen, "%02x", opt[i]);
		break;
	    }
	    opt += len;
	    hdr_len -= len;
	}

	goto tcp_print_done;

	tcp_print_trunc:
	crl_snprintf(&dumpbuf, &buflen, " (truncated at %d bytes)",
	    src->caplen);
    }
    tcp_print_done:

    return 0;
}

/* RFC 768 */
static int coral_get_udp_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    struct udphdr *udp;
    int skip = sizeof(struct udphdr);
    uint16_t sport, dport;

    if (src->caplen < skip)
	return CORAL_ELENGTH;
    dst->buf = src->buf + skip;
    dst->caplen = src->caplen - skip;
    dst->totlen = src->totlen - skip;

    udp = (struct udphdr*)src->buf;
    sport = ntohs(udp->uh_sport);
    dport = ntohs(udp->uh_dport);

    switch(dport) {
    case 161:	dst->protocol = CORAL_PROTO_SNMP;	break;
    default:	dst->protocol = CORAL_PROTO_UNKNOWN;	break;
    }

    if (dumpbuf) {
	crl_snprintf(&dumpbuf, &buflen, "src=%u dst=%u", sport, dport);
	if (!fits(udp, uh_ulen, src->caplen)) goto udp_print_trunc;
	crl_snprintf(&dumpbuf, &buflen, " len=%u", ntohs(udp->uh_ulen));
	goto udp_print_done;

	udp_print_trunc:
	crl_snprintf(&dumpbuf, &buflen, " (truncated at %d bytes)",
	    src->caplen);
    }
    udp_print_done:

    return 0;
}

/* RFC 792, 1122 */
static int coral_get_icmp4_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    struct icmp *icmp;
    const char *atype = NULL, *acode = NULL;
    int is_error = 0;

    dst->buf = src->buf + 8;
    dst->caplen = src->caplen - 8;
    dst->totlen = src->totlen - 8;
    dst->protocol = CORAL_PROTO_UNKNOWN;

    icmp = (struct icmp*)src->buf;
    switch (icmp->icmp_type) {
    case ICMP_UNREACH:
	is_error = 1;
	atype = "UNREACH";
	switch (icmp->icmp_code) {
	case ICMP_UNREACH_NET:		    acode = "NET"; break;
	case ICMP_UNREACH_HOST:		    acode = "HOST"; break;
	case ICMP_UNREACH_PROTOCOL:	    acode = "PROTOCOL"; break;
	case ICMP_UNREACH_PORT:		    acode = "PORT"; break;
	case ICMP_UNREACH_NEEDFRAG:	    acode = "NEEDFRAG"; break;
	case ICMP_UNREACH_SRCFAIL:	    acode = "SRCFAIL"; break;
	case ICMP_UNREACH_NET_UNKNOWN:	    acode = "NET_UNKNOWN"; break;
	case ICMP_UNREACH_HOST_UNKNOWN:	    acode = "HOST_UNKNOWN"; break;
	case ICMP_UNREACH_ISOLATED:	    acode = "ISOLATED"; break;
	case ICMP_UNREACH_NET_PROHIB:	    acode = "NET_PROHIB"; break;
	case ICMP_UNREACH_HOST_PROHIB:	    acode = "HOST_PROHIB"; break;
	case ICMP_UNREACH_TOSNET:	    acode = "TOSNET"; break;
	case ICMP_UNREACH_TOSHOST:	    acode = "TOSHOST"; break;
	case ICMP_UNREACH_FILTER_PROHIB:    acode = "FILTER_PROHIB"; break;
	case ICMP_UNREACH_HOST_PRECEDENCE:  acode = "HOST_PRECEDENCE"; break;
	case ICMP_UNREACH_PRECEDENCE_CUTOFF:acode = "PRECEDENCE_CUTOFF"; break;
	}
	break;
    case ICMP_REDIRECT:
	is_error = 1;
	atype = "REDIRECT";
	switch (icmp->icmp_code) {
	case ICMP_REDIRECT_NET:		    acode = "NET";		break;
	case ICMP_REDIRECT_HOST:	    acode = "HOST";		break;
	case ICMP_REDIRECT_TOSNET:	    acode = "TOSNET";		break;
	case ICMP_REDIRECT_TOSHOST:	    acode = "TOSHOST";		break;
	}
	break;
    case ICMP_TIMXCEED:
	is_error = 1;
	atype = "TIMXCEED";
	switch (icmp->icmp_code) {
	case ICMP_TIMXCEED_INTRANS:	    acode = "INTRANS";		break;
	case ICMP_TIMXCEED_REASS:	    acode = "REASS";		break;
	}
	break;
    case ICMP_PARAMPROB:
	is_error = 1;
	atype = "PARAMPROB";
	switch (icmp->icmp_code) {
	case ICMP_PARAMPROB_OPTABSENT:	    acode = "OPTABSENT";	break;
	case ICMP_PARAMPROB_LENGTH:	    acode = "LENGTH";		break;
	}
	break;
    case ICMP_SOURCEQUENCH:
	is_error = 1;
	atype = "SOURCEQUENCH";
	break;
    case ICMP_PHOTURIS:
	atype = "PHOTURIS";
	switch (icmp->icmp_code) {
	case ICMP_PHOTURIS_BAD_SPI:	    acode = "BAD_SPI";		break;
	case ICMP_PHOTURIS_AUTHENT_FAIL:    acode = "AUTHENT_FAIL";	break;
	case ICMP_PHOTURIS_DECOMP_FAIL:	    acode = "DECOMP_FAIL";	break;
	case ICMP_PHOTURIS_DECRYPT_FAIL:    acode = "DECRYPTION_FAIL";	break;
	case ICMP_PHOTURIS_NEED_AUTHENT:    acode = "NEED_AUTHENT";	break;
	case ICMP_PHOTURIS_NEED_AUTHORIZ:   acode = "NEED_AUTHORIZ";	break;

	}
	break;
    case ICMP_ECHO:		atype = "ECHO";			break;
    case ICMP_ECHOREPLY:	atype = "ECHOREPLY";		break;
    case ICMP_ROUTERADVERT:	atype = "ROUTERADVERT";		break;
    case ICMP_ROUTERSOLICIT:	atype = "ROUTERSOLICIT";	break;
    case ICMP_TSTAMP:		atype = "TSTAMP";		break;
    case ICMP_TSTAMPREPLY:	atype = "TSTAMPREPLY";		break;
    case ICMP_IREQ:		atype = "IREQ";			break;
    case ICMP_IREQREPLY:	atype = "IREQREPLY";		break;
    case ICMP_MASKREQ:		atype = "MASKREQ";		break;
    case ICMP_MASKREPLY:	atype = "MASKREPLY";		break;
    case ICMP_TRACEROUTE:	atype = "TRACEROUTE";		break;
    case ICMP_DGRAM_CONV_ERR:	atype = "DGRAM_CONV_ERR";	break;
    case ICMP_MOBILE_HOST_REDIR: atype = "MOBILE_HOST_REDIR";	break;
    case ICMP_WHERE_ARE_YOU:	atype = "WHERE_ARE_YOU";	break;
    case ICMP_I_AM_HERE:	atype = "I_AM_HERE";		break;
    case ICMP_MOBILE_REG_REQ:	atype = "MOBILE_REG_REQ";	break;
    case ICMP_MOBILE_REG_REP:	atype = "MOBILE_REG_REP";	break;
    case ICMP_SKIP:		atype = "SKIP";			break;
    }

    if (is_error) {
	dst->protocol = CORAL_NETPROTO_IPv4;
    }

    if (dumpbuf) {
	if (atype)
	    crl_snprintf(&dumpbuf, &buflen, "%s", atype);
	else
	    crl_snprintf(&dumpbuf, &buflen, "type=%d", icmp->icmp_type);
	if (acode)
	    crl_snprintf(&dumpbuf, &buflen, "_%s", acode);
	else if (icmp->icmp_code)
	    crl_snprintf(&dumpbuf, &buflen, " code=%d", icmp->icmp_code);
	crl_snprintf(&dumpbuf, &buflen, " cksum=%04x", ntohs(icmp->icmp_cksum));

	if (!fits(icmp, icmp_hun, src->caplen)) goto icmp_print_trunc;
	if (icmp->icmp_type == ICMP_UNREACH &&
	    icmp->icmp_code == ICMP_UNREACH_NEEDFRAG)
	{
	    crl_snprintf(&dumpbuf, &buflen, " nextmtu=%04x",
		ntohs(icmp->icmp_hun.ih_pmtu.ipm_nextmtu));
	} else if (icmp->icmp_type == ICMP_REDIRECT) {
	    crl_snprintf(&dumpbuf, &buflen, " gateway=%s",
		inet_ntoa(icmp->icmp_hun.ih_gwaddr));
	} else if (icmp->icmp_type == ICMP_PARAMPROB) {
	    crl_snprintf(&dumpbuf, &buflen, " offset=%04x",
		ntohs(icmp->icmp_hun.ih_pptr));
	} else if (!is_error) {
	    crl_snprintf(&dumpbuf, &buflen, " id=%04x seq=%04x",
		ntohs(icmp->icmp_hun.ih_idseq.icd_id),
		ntohs(icmp->icmp_hun.ih_idseq.icd_seq));
	}
	goto icmp_print_done;

	icmp_print_trunc:
	crl_snprintf(&dumpbuf, &buflen, " (truncated at %d bytes)",
	    src->caplen);
    }
    icmp_print_done:

    return 0;
}

/* ATM Forum af-ilmi-0065.000.pdf; RFC 1157; ITU-T X.680, X.690 */
/* supports only SNMPv1 */
static int coral_get_snmp_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen,
    const char *community)
{
    const char *p = src->buf;
    uint32_t seqlen, comlen;
    int i;

#define ASN1_TAG_INT		 2
#define ASN1_TAG_OCTSTR		 4
#define ASN1_TAG_SEQ		16
#define ASN1_TAG_HIGH		0x1F
#define ASN1_TAG(x)		((x) & 0x1F)
#define ASN1_PRIMITIVE		0x00
#define ASN1_CONSTRUCTED	0x20
#define ASN1_CLASS_UNIVERSAL	0x00
#define ASN1_CLASS_APPLICATION	0x40
#define ASN1_CLASS_CONTEXT	0x80
#define ASN1_CLASS_PRIVATE	0xC0

    /* SEQUENCE */
    if (p+1 > src->buf + src->caplen)
	return CORAL_ELENGTH;
    if (*p != (ASN1_CLASS_UNIVERSAL | ASN1_CONSTRUCTED | ASN1_TAG_SEQ))
	return CORAL_ESYNTAX;
    p++;
    if (p+1 > src->buf + src->caplen)
	return CORAL_ELENGTH;
    if (!(*p & 0x80)) { /* short form */
	seqlen = *(p++);
    } else if (*p & 0x7F) { /* long form */
	i = *p & 0x7F;
	p++;
	if (p+i > src->buf + src->caplen)
	    return CORAL_ELENGTH;
	for (seqlen = 0; i; i--, p++) {
	    seqlen = (seqlen << 8) + *(unsigned char*)p;
	}
    } else { /* indefinite form */
	return CORAL_ESYNTAX;
    }
    if (dumpbuf)
	crl_snprintf(&dumpbuf, &buflen, " len=%" PRIu32, seqlen);

    /* version INTEGER */
    if (p+3 > src->buf + src->caplen)
	return CORAL_ELENGTH;
    if (*p != (ASN1_CLASS_UNIVERSAL | ASN1_PRIMITIVE | ASN1_TAG_INT))
	return CORAL_ESYNTAX;
    p++;
    if (*p != 1) return CORAL_ESYNTAX;
    p++;
    if (*p != 0) return CORAL_ESYNTAX;
    if (dumpbuf)
	crl_snprintf(&dumpbuf, &buflen, " SNMPv1");
    p++;

    /* community OCTET STRING */
    if (p+2 > src->buf + src->caplen)
	return CORAL_ELENGTH;
    if (*p != (ASN1_CLASS_UNIVERSAL | ASN1_PRIMITIVE | ASN1_TAG_OCTSTR))
	return CORAL_ESYNTAX;
    p++;
    comlen = *(unsigned char*)p;
    p++;
    if (p+comlen > src->buf + src->caplen)
	return CORAL_ELENGTH;
    if (community) {
	if (comlen != strlen(community)) return CORAL_ESYNTAX;
	if (strncmp(p, community, comlen) != 0) return CORAL_ESYNTAX;
    } else {
	for (i = 0; i < comlen; i++)
	    if (!isprint(p[i]))
		return CORAL_ESYNTAX;
    }

#define litstrcmp(var, len, lit) \
    (len == sizeof(lit) - 1 && strncmp(var, lit, len) == 0)

    if (litstrcmp(p, comlen, "ILMI"))
	dst->protocol = CORAL_PROTO_ILMI_PDU;
    else
	dst->protocol = CORAL_PROTO_SNMP_PDU;

    if (dumpbuf)
	crl_snprintf(&dumpbuf, &buflen, " community=\"%.*s\"", comlen, p);

    dst->buf = p + comlen;
    dst->totlen = seqlen - 5 - comlen;
    dst->caplen = src->caplen - (dst->buf - src->buf);
    if (dst->caplen > dst->totlen)
	dst->caplen = dst->totlen;
    return 0;
}

int coral_fmt_get_payload(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, char *dumpbuf, int buflen)
{
    if (dumpbuf && coral_proto_str(src->protocol))
	crl_snprintf(&dumpbuf, &buflen, "%s: ", coral_proto_str(src->protocol));

    switch (src->protocol) {
    case CORAL_PHY_ATM:
	return coral_get_atm_payload(src, dst, dumpbuf, buflen);

    case CORAL_DLT_ATM_AAL5:
	return coral_get_aal5_payload(src, dst, dumpbuf, buflen);

    case CORAL_DLT_ATM_RFC1483:
	return coral_get_llc_payload(src, dst, dumpbuf, buflen);

    case CORAL_DLT_CHDLC:
	return coral_get_chdlc_payload(src, dst, dumpbuf, buflen);

    case CORAL_DLT_ETHER:
    case CORAL_DLT_IEEE802:
	return coral_get_ieee8023_or_ether_payload(src, dst, dumpbuf, buflen);

    case CORAL_DLT_LANE_IEEE8023:
	return coral_get_lane_ieee8023_payload(src, dst, dumpbuf, buflen);

    case CORAL_DLT_LANE_LLC:
	return coral_get_lane_llc_payload(src, dst, dumpbuf, buflen);

    case CORAL_DLT_GIGX:
	return coral_get_gigx_payload(src, dst, dumpbuf, buflen);

    case CORAL_DLT_BRIDGED_LAN:
	return coral_get_bridgedlan_payload(src, dst, dumpbuf, buflen);

    /* Note: coral_open() changes user-configured CORAL_DLT_PPP to
     * CORAL_DLT_PPPo* based on physical layer
     */
    case CORAL_DLT_PPPoHDLC:
	return coral_get_pppohdlc_payload(src, dst, dumpbuf, buflen);

    case CORAL_DLT_PPPoES:
	return coral_get_pppoe_sess_payload(src, dst, dumpbuf, buflen);

    case CORAL_DLT_PPPoED:
	return coral_get_pppoe_disc_payload(src, dst, dumpbuf, buflen);

    case CORAL_DLT_PPP:
	return coral_get_ppp_payload(src, dst, dumpbuf, buflen);

    case CORAL_DLT_ILMI: /* ATM Forum af-ilmi-0065.000.pdf (SNMPv1 only) */
	return coral_get_snmp_payload(src, dst, dumpbuf, buflen, "ILMI");

    case CORAL_NETPROTO_ARP:
	return coral_get_arp_payload(src, dst, dumpbuf, buflen);

    case CORAL_NETPROTO_IPv4:
	return coral_get_ip4_payload(src, dst, dumpbuf, buflen);

#ifdef HAVE_INET6
    case CORAL_NETPROTO_IPv6:
	return coral_get_ip6_payload(src, dst, dumpbuf, buflen);

    case CORAL_IPPROTO_HOPOPTS:
    case CORAL_IPPROTO_ROUTING:
    case CORAL_IPPROTO_FRAGMENT:
    case CORAL_IPPROTO_ESP:
    case CORAL_IPPROTO_AH:
    case CORAL_IPPROTO_DSTOPTS:
	return coral_get_ipexthdr_payload(src, dst, dumpbuf, buflen);
#endif

#if 0
    case CORAL_NETPROTO_APPLETALK:
    case CORAL_NETPROTO_AARP:
    /* see:
      http://developer.apple.com/techpubs/mac/NetworkingOT/NetworkingWOT-58.html
      http://www.cisco.com/univercd/cc/td/doc/cisintwk/ito_doc/applet.htm
     */
#endif

    case CORAL_IPPROTO_TCP:
	return coral_get_tcp_payload(src, dst, dumpbuf, buflen);

    case CORAL_IPPROTO_UDP:
	return coral_get_udp_payload(src, dst, dumpbuf, buflen);

    case CORAL_IPPROTO_ICMP:
	return coral_get_icmp4_payload(src, dst, dumpbuf, buflen);

#if 0
#ifdef HAVE_INET6
    case CORAL_IPPROTO_ICMP6:
	return coral_get_icmp6_payload(src, dst, dumpbuf, buflen);
#endif
#endif

    case CORAL_PROTO_SNMP:
	return coral_get_snmp_payload(src, dst, dumpbuf, buflen, NULL);

    default:
	return CORAL_ENOPROTO;
    }
}

int coral_get_payload(const coral_pkt_buffer_t *src, coral_pkt_buffer_t *dst)
{
    return coral_fmt_get_payload(src, dst, NULL, 0);
}

/* deprecated */
const char *coral_get_net_pkt(const coral_pkt_buffer_t *buffer, int *protocol)
{
    coral_pkt_buffer_t dst;

    if (coral_get_payload(buffer, &dst) < 0)
	return NULL;
    *protocol = dst.protocol;
    return dst.buf;
}

int coral_get_payload_by_proto(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, coral_protocol_t proto)
{
    int x = 0, result;
    coral_pkt_buffer_t buf[2], *ldst;
    const coral_pkt_buffer_t *lsrc;

    lsrc = src;
    ldst = &buf[x];
    while (lsrc->protocol != proto) {
	if ((result = coral_get_payload(lsrc, ldst)) < 0)
	    return result;
	/* swap buffers */
	x = !x;
	ldst = &buf[x];
	lsrc = &buf[!x];
    }

    *dst = *lsrc;
    return 0;
}

int coral_get_payload_by_layer(const coral_pkt_buffer_t *src,
    coral_pkt_buffer_t *dst, int layer)
{
    int x = 0, result;
    coral_pkt_buffer_t buf[2], *ldst;
    const coral_pkt_buffer_t *lsrc;

    lsrc = src;
    ldst = &buf[x];
    while (coral_proto_layer(lsrc->protocol) < layer) {
	if ((result = coral_get_payload(lsrc, ldst)) < 0)
	    return result;
	/* swap buffers */
	x = !x;
	ldst = &buf[x];
	lsrc = &buf[!x];
    }

    *dst = *lsrc;
    return 0;
}

void coral_fmprint_pkt(FILE *file, coral_iface_t *iface,
    const coral_timestamp_t *timestamp, int minlayer, int maxlayer,
    coral_pkt_buffer_t *packet,
    coral_pkt_buffer_t *header, coral_pkt_buffer_t *trailer)
{
    struct timespec ts;
    int x = 0;
    coral_pkt_buffer_t buf[2], *src, *dst;
    char dumpbuf[256];

    dst = &buf[x];

    if (iface) {
	fprintf(file, "interface: %d\n", coral_interface_get_number(iface));
	if (timestamp) {
	    coral_clock_timespec_inline(iface, timestamp, &ts);
	    fprintf(file, "time:  %ld.%09ld\n", ts.tv_sec, ts.tv_nsec);
	}
    }

    /* Print physical name, unless there's a physical header to print */
    if (iface && iface->iface_info.physical != CORAL_PROTO_UNKNOWN &&
	coral_proto_layer(iface->iface_info.physical) >= minlayer &&
	coral_proto_layer(iface->iface_info.physical) <= maxlayer &&
	!(header && header->protocol == iface->iface_info.physical))
    {
	fprintf(file, "proto: %s\n",
	    coral_proto_str(iface->iface_info.physical));
    }

    /* Print header */
    if (header && header->protocol != CORAL_PROTO_UNKNOWN &&
	coral_proto_layer(header->protocol) >= minlayer &&
	coral_proto_layer(header->protocol) <= maxlayer)
    {
	coral_fmt_get_payload(header, dst, dumpbuf, sizeof(dumpbuf));
	fprintf(file, "proto: %s\n", dumpbuf);
    }

    /* Print trailer */
    if (trailer && trailer->protocol != CORAL_PROTO_UNKNOWN &&
	coral_proto_layer(trailer->protocol) >= minlayer &&
	coral_proto_layer(trailer->protocol) <= maxlayer)
    {
	coral_fmt_get_payload(trailer, dst, dumpbuf, sizeof(dumpbuf));
	fprintf(file, "proto: %s\n", dumpbuf);
    }

    src = packet;

    while (src->protocol != CORAL_PROTO_UNKNOWN && src->caplen > 0 &&
	coral_proto_layer(src->protocol) <= maxlayer)
    {
	if (coral_fmt_get_payload(src, dst,
	    coral_proto_layer(src->protocol) >= minlayer ? dumpbuf : NULL,
	    sizeof(dumpbuf)) < 0)
	{
	    dst = src;
	    break;
	}
	if (coral_proto_layer(src->protocol) >= minlayer)
	    fprintf(file, "proto: %s\n", dumpbuf);
	x = !x;
	src = dst;
	dst = &buf[x];
    }

    if (src->protocol != CORAL_PROTO_UNKNOWN || src->caplen || src->totlen) {
	fprintf(file, "data:  ");
	if (src->protocol != CORAL_PROTO_UNKNOWN)
	    fprintf(file, "%s, ", coral_proto_str(src->protocol));
	fprintf(file, "%d bytes", src->caplen);
	if (src->totlen != src->caplen) {
	    if (src->totlen >= 0)
		fprintf(file, " (of %d)", src->totlen);
	    else
		fprintf(file, " (of ?)");
	}
	fprintf(file, "\n");
	coral_fprint_data(file, 7, (u_char*)src->buf, src->caplen);
    }
    fprintf(file, "\n");
}

void coral_fprint_pkt(FILE *file, coral_iface_t *iface,
    const coral_timestamp_t *timestamp, int maxlayer,
    coral_pkt_buffer_t *packet,
    coral_pkt_buffer_t *header, coral_pkt_buffer_t *trailer)
{
    coral_fmprint_pkt(stdout, iface, timestamp,
	1, maxlayer, packet, header, trailer);
}

void coral_mprint_pkt(coral_iface_t *iface, const coral_timestamp_t *timestamp,
    void *layerp, coral_pkt_buffer_t *packet,
    coral_pkt_buffer_t *header, coral_pkt_buffer_t *trailer)
{
    coral_fmprint_pkt(stdout, iface, timestamp,
	((int*)layerp)[0], ((int*)layerp)[1], packet, header, trailer);
}

void coral_print_pkt(coral_iface_t *iface, const coral_timestamp_t *timestamp,
    void *layerp, coral_pkt_buffer_t *packet,
    coral_pkt_buffer_t *header, coral_pkt_buffer_t *trailer)
{
    coral_fmprint_pkt(stdout, iface, timestamp,
	1, ((int*)layerp)[0], packet, header, trailer);
}

/* assumes buf len is at least CORAL_FMT_SUBIF_LEN */
int coral_fmt_subif(char *buf, const coral_iface_t *iface, uint32_t subif)
{
    if (iface->iface_info.physical == CORAL_PHY_ATM) {
	return sprintf(buf, "%d:%d", get_vpvc_vp(subif), get_vpvc_vc(subif));
    } else if (iface->iface_info.datalink == CORAL_DLT_ETHER) {
	return sprintf(buf, "%d", subif);
    }
    buf[0] = '\0';
    return 0;
}

/* assumes buf len is at least CORAL_FMT_IF_SUBIF_LEN */
int coral_fmt_if_subif(char *buf, const coral_iface_t *iface, uint32_t subif)
{
    int len, subif_len;
    len = sprintf(buf, "%d", iface->id);
    buf[len++] = '[';
    subif_len = coral_fmt_subif(buf + len, iface, subif);
    if (!subif_len) {
	len--;
    } else {
	len += subif_len;
	buf[len++] = ']';
    }
    buf[len] = '\0';
    return len;
}
