/* 
 * Copyright 1999, 2000, 2001 The Regents of the University of California 
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

#include "config.h"

static const char RCSid[] = "$Id: coral_to_pcap.c,v 1.11.4.2 2001/11/21 01:33:49 kkeys Exp $";

#ifdef HAVE_LIBPCAP
#if defined(HAVE_PCAP_INT_H) || defined(HAVE_PCAP_OPEN_DEAD)

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>

#include <pcap.h>
#ifndef HAVE_PCAP_OPEN_DEAD
# include <pcap-int.h>
#endif

#include "libcoral.h"
#include "libcoral_priv.h"

static int coral_proto_to_pcap(coral_protocol_t proto)
{
    switch(proto) {
	case CORAL_DLT_ATM_AAL5:	/* XXX fall through */
	case CORAL_DLT_ATM_RFC1483:	return DLT_ATM_RFC1483;
	case CORAL_DLT_NULL:		return DLT_NULL;
	case CORAL_DLT_ETHER:		return DLT_EN10MB;
	case CORAL_DLT_IEEE802:		return DLT_IEEE802;
	case CORAL_DLT_PPP:		return DLT_PPP;
#ifdef DLT_CHDLC
	case CORAL_DLT_CHDLC:		return DLT_CHDLC;
#endif
	case CORAL_NETPROTO_IPv4:	return DLT_RAW;
	case CORAL_DLT_LANE_IEEE8023:   /* fall through */
	case CORAL_DLT_LANE_IEEE8025:   /* fall through */
	case CORAL_DLT_LANE_LLC:	/* fall through */
	case CORAL_DLT_GIGX:		/* fall through */
	default:			return 0;
    }
}

pcap_t *coral_iface_to_alloced_pcapp(const coral_iface_t *iface, int raw)
{
    int dlt;
    int snaplen;

    if (iface->iface_info.iomode.flags & CORAL_RX_USER_ALL) {
	snaplen = MAX_PACKET_SIZE;
    } else {
	snaplen = iface->iface_info.iomode.first_n;
    }

    if (raw) {
	dlt = DLT_RAW;
    } else {
	dlt = coral_proto_to_pcap(iface->iface_info.datalink);
	if (!dlt) {
	    coral_diag(1, ("Warning: don't know pcap equivalent of %s "
	       "protocol on interface %d, will try to convert to raw IP.\n",
	       coral_proto_str(iface->iface_info.datalink), iface->id));
	    dlt = DLT_RAW;
	}
    }

#ifdef HAVE_PCAP_OPEN_DEAD /* prefered method */
    return pcap_open_dead(dlt, snaplen);
#else /* old method */
    {
	pcap_t *pcap;
	if (!(pcap = malloc(sizeof(pcap_t)))) return NULL;
	memset(pcap, 0, sizeof(*pcap));
	pcap->fd = -1;
	pcap->tzoff = iface->iface_info.tzoff;
	pcap->linktype = dlt;
	pcap->snapshot = snaplen;
	return pcap;
    }
#endif
}

pcap_t *coral_iface_to_pcapp(coral_iface_t *iface)
{
    if (!iface->user_pcap)
	iface->user_pcap = coral_iface_to_alloced_pcapp(iface, 0);
    return iface->user_pcap;
}

pcap_t *coral_iface_to_pcapp_raw(coral_iface_t *iface)
{
    if (!iface->user_pcap)
	iface->user_pcap = coral_iface_to_alloced_pcapp(iface, 1);
    return iface->user_pcap;
}

int coral_pkt_to_pcap(coral_iface_t *iface,
    const coral_timestamp_t *timestamp,
    const coral_pkt_buffer_t *pkt, pcap_t *pcap, struct pcap_pkthdr *hdr)
{
    coral_pkt_buffer_t ippkt;
    int offset = 0;

    if (pcap_datalink(pcap) != DLT_RAW) {
	int first_time = !iface->pcap_proto;
	if (!iface->pcap_proto)
	    iface->pcap_proto = coral_proto_to_pcap(iface->iface_info.datalink);
	if (iface->pcap_proto != pcap_datalink(pcap)) {
	    if (first_time)
		coral_diag(0, ("Can not convert %s (%d) packets from interface"
		    " %d to pcap interface with incompatible protocol (%d).\n",
		    coral_proto_str(iface->iface_info.datalink),
		    iface->pcap_proto, iface->id, pcap_datalink(pcap)));
	    return -1;
	}

    } else if (iface->iface_info.datalink != CORAL_NETPROTO_IPv4) {
	if (coral_get_payload_by_proto(pkt, &ippkt, CORAL_NETPROTO_IPv4) < 0)
	    return -1;
	offset = ippkt.buf - pkt->buf;
	pkt = &ippkt;
    }

    if (timestamp)
	CORAL_TIMESTAMP_TO_TIMEVAL_NORM(iface, timestamp, &hdr->ts, 1);
    else
	hdr->ts.tv_sec = hdr->ts.tv_usec = 0;
    hdr->caplen = pkt->caplen;
    hdr->len = pkt->totlen;
    return offset;
}

int coral_write_pkt_to_pcap(coral_iface_t *iface,
    const coral_timestamp_t *timestamp,
    const coral_pkt_buffer_t *pkt, pcap_t *pcap, pcap_dumper_t *pcap_dumper)
{
    int offset;
    struct pcap_pkthdr pphr;

    offset = coral_pkt_to_pcap(iface, timestamp, pkt, pcap, &pphr);
    if (offset < 0) return offset;
    pcap_dump((unsigned char*)pcap_dumper, &pphr,
	(unsigned char*)pkt->buf + offset);
    return 0;
}

int coral_pkt_filter(coral_iface_t *iface, const coral_pkt_buffer_t *pkt,
    int *errp)
{
    struct pcap_pkthdr pcap_hdr;
    int offset;

    offset = coral_pkt_to_pcap(iface, NULL, pkt, iface->internal_pcap,
	&pcap_hdr);
    if (offset < 0) return 0;

    return crl_bpf_filter(iface->bpfprog->bf_insns, (uint8_t*)pkt->buf + offset,
	pcap_hdr.len, pcap_hdr.caplen, errp);
}

#endif /* defined(HAVE_PCAP_INT_H) || defined(HAVE_PCAP_OPEN_DEAD) */
#endif /* HAVE_LIBPCAP */
