/*
 * Copyright (c) 1999, 2000 The University of Waikato, Hamilton, New Zealand.
 * All rights reserved.
 *
 * This code is derived from software contributed to CAIDA
 * by the University of Waikato WAND research group.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      Waikato, Hamilton, NZ, and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE UNIVERSITY OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <time.h>

#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <errno.h>
#ifdef HAVE_SYS_IOCCOM_H
#include <sys/ioccom.h>
#endif
#include <sys/ioctl.h>
#include <sys/mman.h>

#include "drivers/coral_ioctl.h"
#include "libcoral.h"
#include "libcoral_priv.h"

#include "drivers/dag_coral.h"

# if 0
#include "dag3.h"
#include "dag3io.h"
#include "dag3lib.h"
# endif

#include "dag.h"

static const char RCSid[] = "$Id: coral_type_dag.c,v 1.8.2.3 2001/11/21 01:33:49 kkeys Exp $";

static int
coral_dag_init(coral_source_t *src)
{
    int id = src->id[0];
    int dagspeed;
    uint32_t bw;
    coral_protocol_t phy;

    if (ioctl(src->fd, DAG_IOGETSPEED, &dagspeed) < 0) {
	coral_diag(0, ("coral_open: %s: ioctl DAG_IOGETSPEED: %s\n",
	    src->name, strerror(errno)));
	phy = CORAL_PROTO_UNKNOWN;
	bw = 0;
    } else {
	switch (dagspeed) {
        case OC3c_ATM:   phy = CORAL_PHY_ATM;  bw = KBPS_OC3c;   break;
        case OC12c_ATM:  phy = CORAL_PHY_ATM;  bw = KBPS_OC12c;  break;
        case OC48c_ATM:  phy = CORAL_PHY_ATM;  bw = KBPS_OC48c;  break;
        case OC3c_POS:   phy = CORAL_PHY_POS;  bw = KBPS_OC3c;   break;
        case OC12c_POS:  phy = CORAL_PHY_POS;  bw = KBPS_OC12c;  break;
        case OC48c_POS:  phy = CORAL_PHY_POS;  bw = KBPS_OC48c;  break;
        case DS3:
        case E3:
        case ETH_10:
        case ETH_100:
        case ETH_1000:
	default:         phy = CORAL_PROTO_UNKNOWN; bw = 0; break;
	}
	coral_diag(5, ("coral_dag_init: %s is configured for %s %s\n",
	    src->name, coral_bandwidth_fmt(bw), coral_proto_str(phy)));
    }

    if (src->dev_config.physical == CORAL_PROTO_UNKNOWN) {
	src->dev_config.physical = phy;
    } else if (src->dev_config.physical != phy) {
	coral_diag(0, ("coral: %s: illegal physical type: %s\n",
	    src->name, coral_proto_str(src->dev_config.physical)));
	return -1;
    }

    if (src->dev_config.bandwidth <= 0) {
	src->dev_config.bandwidth = bw;
    } else if (src->dev_config.bandwidth != bw) {
	coral_diag(0, ("coral: %s: unsupported bandwidth: %s\n",
	    src->filename, coral_bandwidth_fmt(src->dev_config.bandwidth)));
	return -1;
    }

    if (coral_dag_common_init(src) < 0)
	return -1;

    cinst[id]->iface_info.fw_type       = 0x19671001; /* XXX ? */

    /* DAG driver looks for first_n==0, not USER_ALL; give it what it wants. */
    if (src->dev_config.iomode.flags & CORAL_RX_USER_ALL)
	src->dev_config.iomode.first_n = 0;
    if (ioctl(src->fd, CORAL_INIT, &src->dev_config.iomode) < 0) {
	char iomode_buffer[64];
	coral_format_iomode(iomode_buffer, &src->dev_config.iomode);
	coral_diag(0, ("coral (%s): CORAL_INIT (iomode %s): %s\n",
		src->filename, iomode_buffer, strerror(errno)));
	return -1;
    }

    return 0;
}

static coral_iface_t *
coral_dag_nextblk(coral_source_t *src, coral_blk_info_t **binfop,
    coral_atm_cell_t **cellp, int endonly)
{
    int idx;

    if (ioctl(src->fd, CORAL_NEXTBLK, &idx) < 0) {
	coral_diag(2, ("coral: dag_nextblk NEXTBLK: %s\n", strerror(errno)));
	return NULL;
    }

    /* Note: A bug in dag driver 2.2.1d and earlier did not initialize
     * devinfo.num_blks early enough, causing coral to core here on the
     * first run after loading the drivers.  Subsequent runs are ok.
     */
    assert(src->ubase);
    *binfop = &src->binfo_buf;
    if (ioctl(src->fd, CORAL_FILL_BLKHEADER, *binfop) < 0) {
	coral_diag(2, ("coral: dag_nextblk FILL_BLKHEADER: %s\n",
	    strerror(errno)));
	return NULL;
    }

    if (idx < 0) {
	/* EOF on this fd */
	errno = 0;
	return NULL;
    }

    *cellp = (coral_atm_cell_t *)(src->ubase + DAG_BLK_SIZE * idx);
    coral_diag(19, ("coral: dag_nextblk: %d\n", idx));

    return cinst[src->id[0]];
}

const coral_src_type_t coral_src_type_dag = {
    CORAL_TYPE_DAG,
    "dag",
    1 /* is_live */,
    1 /* is_buffered */,
    1 /* is_coral_dev */,
    coral_dag_init,
    coral_blk_read_min,
    coral_dag_nextblk,
    NULL /* release */,
    NULL /* has_buf */,
    coral_coraldev_stop,
    NULL /* coral_dag_close */
};
