/* 
 * Copyright 2000, 2001 The Regents of the University of California 
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>

#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#include "libcoral.h"
#include "libcoral_priv.h"

static const char RCSid[] = "$Id: coral_type_dagfile.c,v 1.8.2.2 2001/11/21 01:33:49 kkeys Exp $";

int coral_dag_common_init(coral_source_t *src)
{
    int id = src->id[0];
    int max_first_n;

    cinst[id]->iface_info.hw_version    = 0;	/* default */
    cinst[id]->iface_info.fw_version    = 0;	/* default */

    /* DAG supports ATM and POS.  If physical was not set, infer from datalink.
     * Do NOT just default to ATM; we want to force user to specify a protocol,
     * especially for dagfiles.
     */
    if (src->dev_config.physical == CORAL_PROTO_UNKNOWN) {
	switch (src->dev_config.datalink) {
	    case CORAL_DLT_ATM_RFC1483:
	    case CORAL_DLT_LANE_IEEE8023:
	    case CORAL_DLT_LANE_IEEE8025:
	    case CORAL_DLT_LANE_LLC:
	    case CORAL_DLT_GIGX:
		src->dev_config.physical = CORAL_PHY_ATM;
		break;
	    case CORAL_DLT_CHDLC:
	    case CORAL_DLT_PPP:
		src->dev_config.physical = CORAL_PHY_POS;
		break;
	    default:
		src->dev_config.physical = CORAL_PROTO_UNKNOWN;
		break;
	}
    }

    switch (src->dev_config.physical) {
	case CORAL_PHY_ATM:
	    cinst[id]->iface_type = &coral_iface_type_dag_atm;
	    max_first_n = ATM_PAYLOAD_SIZE * 15;
	    coral_init_atm_iface(cinst[id]);
	    break;
	case CORAL_PHY_POS:
	    cinst[id]->iface_type = &coral_iface_type_dag_pos;
	    max_first_n = 48;
	    break;
	default:
	    coral_diag(0, ("coral: %s: illegal physical type: %s\n",
		src->name, coral_proto_str(src->dev_config.physical)));
	    goto fail;
    }

    /* If this is a device and any src->dev_config fields are still unset,
     * use defaults for DAG and physical type.
     */
    if (coral_type_is_driver(src->type.coral_type)) {
	coral_set_dev_config_defaults(&src->dev_config,
	    &cinst[id]->iface_type->default_config);
    }

    if (src->dev_config.iomode.flags & ~CORAL_DAG_MODES) {
	char iomode_buffer[64];
	coral_format_iomode(iomode_buffer, &src->dev_config.iomode);
	coral_diag(0, ("%s: unsupported mode: %s\n",
	    src->filename, iomode_buffer));
	goto fail;
    }

    if (src->dev_config.iomode.first_n > max_first_n) {
	coral_diag(1, ("warning: %s: %s can capture only first %d bytes\n",
	    src->filename, cinst[id]->iface_type->name, max_first_n));
	src->dev_config.iomode.first_n = max_first_n;
    }

    if (src->dev_config.bandwidth > 0) {
	switch (src->dev_config.bandwidth) {
	case KBPS_OC3c:  break;
	case KBPS_OC12c: break;
	default:
	    coral_diag(0, ("%s: unsupported bandwidth %s\n",
		src->filename, coral_bandwidth_fmt(src->dev_config.bandwidth)));
	    goto fail;
	}
    }

    cinst[id]->iface_info.hw_type		= CORAL_TYPE_DAG;
    cinst[id]->iface_info.fw_type		= 0;
    cinst[id]->iface_info.iomode		= src->dev_config.iomode;
    cinst[id]->iface_info.capture_time		= 0;
    cinst[id]->iface_info.bandwidth		= src->dev_config.bandwidth;
    cinst[id]->iface_info.time_is_le		= 1;
    cinst[id]->iface_info.datalink		= src->dev_config.datalink;
    cinst[id]->iface_info.physical		= src->dev_config.physical;
    cinst[id]->iface_info.tzoff			= 0;
    cinst[id]->iface_info.capture_tv.tv_sec	= 0;
    cinst[id]->iface_info.capture_tv.tv_usec	= 0;

    return 0;
fail:
    return -1;
}

#include "drivers/dag_coral.h"
#define DAG_RECORD_SIZE (64)

typedef struct dag_blk_and_info {
    char data[DAG_BLK_SIZE];
    coral_blk_info_t binfo;
} dag_blk_and_info_t;

static int coral_dagfile_init(coral_source_t *src)
{
    if (src->fd >= 0) {
	/* already open */
    } else if (strcmp(src->filename, "-") == 0) {
	src->fd = STDIN_FILENO;
	cinst[src->id[0]]->devinfo.type = CORAL_TYPE_FILE; /* XXX ? */
    } else {
	if ((src->fd = open(src->filename, O_RDONLY, 0)) < 0) {
	    coral_diag(0, ("coral_open: %s: open: %s\n",
		src->name, strerror(errno)));
	    return -1;
	}
	coral_diag(4, ("coral_open: %s: fd %d\n", src->name, src->fd));
    }

    if (coral_dag_common_init(src) < 0)
	return -1;
    cinst[src->id[0]]->iface_info.sw_type = SW_TYPE_DAG;
    if (coral_file_common_init(src, sizeof(dag_blk_and_info_t)) < 0)
	return -1;
    return 0;
}

static coral_iface_t *
coral_dagfile_nextblk(coral_source_t *src, coral_blk_info_t **binfop,
    coral_atm_cell_t **cellp, int endonly)
{
    ssize_t len, need;

    /* If time-sorted reading api stole ubase, we must alloc a new one. */
    if (!src->ubase) {
        src->ubase = (char *)malloc(sizeof(dag_blk_and_info_t));
        if (!src->ubase) {
            coral_diag(0, ("coral_nextblk: malloc: %s\n", strerror(errno)));
            goto coral_dagfile_nextblk_error;
        }
    }

    endonly = endonly && src->can_seek;
retry:
    need = endonly ? DAG_RECORD_SIZE : DAG_BLK_SIZE;
    len = coral_readall(src, src->ubase, need);
    if (!len) {
	goto coral_dagfile_nextblk_eof;
    } else if (len < 0) {
	coral_diag(0, ("coral: %s: %s\n", src->filename, strerror(errno)));
	goto coral_dagfile_nextblk_error;
    }
    if (endonly) {
	off_t old, new;
	need = DAG_BLK_SIZE - 2 * DAG_RECORD_SIZE;
	old = lseek(src->fd, 0, SEEK_CUR);
	new = lseek(src->fd, need, SEEK_CUR);
	if (new == (off_t)-1) {
            coral_diag(0, ("coral_nextblk: seek: %s\n", strerror(errno)));
            goto coral_dagfile_nextblk_error;
	}
	if (new - old != need)
	    goto reset;
	len = coral_readall(src, src->ubase + DAG_BLK_SIZE - DAG_RECORD_SIZE,
	    DAG_RECORD_SIZE);
	if (len < 0) {
	    coral_diag(0, ("coral: %s: %s\n", src->filename, strerror(errno)));
	    goto coral_dagfile_nextblk_error;
	} else if (len != DAG_RECORD_SIZE) {
	    goto reset;
	}
	len = DAG_BLK_SIZE;
	goto ok;
reset:
	endonly = 0;
	lseek(src->fd, old - DAG_RECORD_SIZE, SEEK_SET);
	goto retry;
    }
ok:
    if (len % DAG_RECORD_SIZE) {
	coral_diag(1, ("warning: %s: discarding %d bytes outside record.\n",
	    src->filename, (len % DAG_RECORD_SIZE)));
    }
    *binfop = &((dag_blk_and_info_t*)src->ubase)->binfo;
    (*binfop)->interface = htonl(0);
    (*binfop)->blk_size = htonl(DAG_BLK_SIZE);
    (*binfop)->cell_count = htonl(len / DAG_RECORD_SIZE);
    (*binfop)->tbegin.tv_sec = (*binfop)->tbegin.tv_nsec = htonl(0);
    (*binfop)->tend.tv_sec = (*binfop)->tend.tv_nsec = htonl(0);

    *cellp = (coral_atm_cell_t*)src->ubase;

    errno = 0;
    return cinst[src->id[0]];

coral_dagfile_nextblk_eof:
    errno = 0;
coral_dagfile_nextblk_error:
    *binfop = NULL;
    *cellp = NULL;
    return NULL;
}

const coral_src_type_t coral_src_type_dagfile = {
    CORAL_TYPE_DAGFILE,
    "dagfile",
    0 /* is_live */,
    0 /* is_buffered */,
    0 /* is_coral_dev */,
    coral_dagfile_init,
    coral_file_read_min,
    coral_dagfile_nextblk,
    coral_file_release,
    NULL /* has_buf */,
    coral_mark_eof_src /* stop */,
    coral_file_close
};
