/* 
 * Copyright 1999, 2000, 2001 The Regents of the University of California 
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#include "libcoral.h"
#include "libcoral_priv.h"

#include <pcap.h>
#ifdef HAVE_PCAP_INT_H
# include <pcap-int.h>
#endif

static const char RCSid[] = "$Id: coral_type_pcap.c,v 1.22.2.4 2001/11/21 01:33:50 kkeys Exp $";

static int coral_pcap_init2(coral_source_t *src);


static int coral_pcap_init(coral_source_t *src)
{
    pcap_t *pfile;
    char ebuf[PCAP_ERRBUF_SIZE];

    src->version = 0;
    src->iface_count = 1;

    if (src->fd >= 0) {
	coral_diag(0, ("coral_open: %s: %s", src->filename,
	    "pcap files can not be opened from file descriptors"));
	return -1;
    }
    src->file = pfile = pcap_open_offline(src->filename, ebuf);
    if (!pfile) {
	coral_diag(0, ("coral_open (%s): %s\n", src->name, ebuf));
	return -1;
    }
    coral_diag(4, ("coral_open (%s): pcap\n", src->name));

    cinst[src->id[0]]->iface_info.iomode.flags = CORAL_RX;
    cinst[src->id[0]]->iface_info.iomode.first_n = pcap_snapshot(pfile);

    if (coral_pcap_init2(src) < 0) return -1;
    coral_file_check_iomode(src);
    return 0;
}

static int coral_pcaplive_init(coral_source_t *src)
{
    pcap_t *pfile;
    char ebuf[PCAP_ERRBUF_SIZE];

    src->version = 0;
    src->iface_count = 1;

    /* If any src->dev_config fields are still unset, use device defaults. */
    coral_set_dev_config_defaults(&src->dev_config,
	&coral_iface_type_pcap.default_config);

    if (src->dev_config.iomode.flags & CORAL_RX_USER_ALL) {
	src->dev_config.iomode.first_n = MAX_PACKET_SIZE;
    }

    src->fd = -1;
    /* Open with a timeout so pcap_dispatch does not block indefinitely.
     * We need it to not block so we can read other interfaces, and handle
     * any flags set by interrupt handlers (pcap_dispatch does not return
     * when interrupted).
     */
    src->file = pfile = pcap_open_live(src->filename,
	src->dev_config.iomode.first_n, 1, 1, ebuf);
    if (!pfile) {
	coral_diag(0, ("coral_open (%s): %s\n", src->name, ebuf));
	return -1;
    }
    coral_diag(4, ("coral_open (%s): pcap-opened with snaplen %d\n",
	src->name, src->dev_config.iomode.first_n));
    src->fd = pcap_fileno(pfile);

    cinst[src->id[0]]->iface_info.iomode = src->dev_config.iomode;

    if (coral_pcap_init2(src) < 0) return -1;
    coral_file_check_iomode(src);
    return 0;
}

static int coral_pcap_init2(coral_source_t *src)
{
    coral_iface_t *iface;
    pcap_t *pfile = src->file;
    coral_proto32_t *dlt;

    iface = cinst[src->id[0]];
    iface->internal_pcap		= src->file;
    iface->user_pcap			= NULL;
    iface->iface_info.hw_type           = 0;
    iface->iface_info.hw_version        = 0;
    iface->iface_info.fw_type           = 0;
    iface->iface_info.fw_version        = 0;
    iface->iface_info.sw_type           = SW_TYPE_PCAP;
    iface->iface_info.sw_version        = (pcap_major_version(pfile) << 16) |
					 (pcap_minor_version(pfile) & 0xFFFF);
    iface->iface_info.capture_time      = 0;
    iface->iface_info.bandwidth         = 0;

    /* libpcap always returns data in host order */
    iface->iface_info.time_is_le        = !WORDS_BIGENDIAN;

    iface->iface_type			= &coral_iface_type_pcap;
#ifdef HAVE_PCAP_INT_H
    iface->iface_info.tzoff		= pfile->tzoff;
#else
    iface->iface_info.tzoff		= 0;
#endif
    iface->iface_info.capture_tv.tv_sec = 0;
    iface->iface_info.capture_tv.tv_usec= 0;

    dlt = &iface->iface_info.datalink;
    switch (pcap_datalink(pfile)) {
	case DLT_NULL:		*dlt = CORAL_DLT_NULL;		break;
	case DLT_EN10MB:	*dlt = CORAL_DLT_ETHER;		break;
	case DLT_IEEE802:	*dlt = CORAL_DLT_IEEE802;	break;
	case DLT_PPP:		*dlt = CORAL_DLT_PPP;		break;
	case DLT_ATM_RFC1483:	*dlt = CORAL_DLT_ATM_RFC1483;	break;
#ifdef DLT_RAW
	case DLT_RAW:		*dlt = CORAL_NETPROTO_IPv4;	break;
#endif
#ifdef DLT_CHDLC
	case DLT_CHDLC:		*dlt = CORAL_DLT_CHDLC;		break;
#endif
	default: {
	    *dlt = CORAL_PROTO_UNKNOWN;
	    coral_diag(1,
		("coral_open (%s): unknown/unsupported data link type %d\n",
		src->name, pcap_datalink(pfile)));
	    break;
	}
    }

    return 0;
}

/* from libpcap-0.4/pcap.c */
struct singleton {
    struct pcap_pkthdr *hdr;
    const u_char *pkt;
};

static void
pcap_oneshot(u_char *userData, const struct pcap_pkthdr *h, const u_char *pkt)
{
    struct singleton *sp = (struct singleton *)userData;
    *sp->hdr = *h;
    sp->pkt = pkt;
}
/* end pcap.c */


/* Read the next packet and store it and its timestamp on the iface struct.
 * Since this can only be called from the packet API, we also update the
 * iface->pkt_result, eliminating the need for a pcap prep_pkt().
 */
static coral_iface_t *coral_pcap_read_min(coral_iface_t *iface)
{
    int result;
    struct singleton s;

    s.hdr = &iface->u.pcap.hdr;
    assert(!iface->eof);

    result = pcap_dispatch(iface->src->file, 1, pcap_oneshot, (u_char*)&s);
    /* pcap_dispatch returns:
     *              file live
     * read packet:    1    1
     * no data:        x    0
     * eof:            0    x
     * error:         -1   -1
     */
    if (coral_verbosity >= 24) {
	coral_printf("pcap_dispatch returned %d\n", result);
	if (result < 0)
	    coral_printf("    errno=%d (%s)\n", errno, strerror(errno));
    }
    if (result < 0) {
	coral_diag(0, ("%s: %s\n",
	    iface->src->name, pcap_geterr(iface->src->file)));
	coral_mark_eof(iface);
	if (!errno) errno = -1;
	return NULL;
    }
    if (result == 0) {
	if (iface->src->type.coral_type == CORAL_TYPE_PCAPLIVE) {
	    errno = EAGAIN;
	} else {
	    errno = 0;
	    coral_mark_eof(iface);
	}
	return NULL;
    }

    iface->packetbuf.buf = (char*)s.pkt;
    iface->packetbuf.caplen = iface->u.pcap.hdr.caplen;
    iface->packetbuf.totlen = iface->u.pcap.hdr.len;
    iface->packetbuf.is_dynamic = 0;
    iface->packetbuf.protocol = iface->iface_info.datalink;
    iface->packetbuf.passed = 1;	/* libpcap already applied bpf filter */

    iface->pkt_result.packet = &iface->packetbuf;
    iface->pkt_result.header = NULL;
    iface->pkt_result.trailer = NULL;
    iface->pkt_result.subiface = 0;
    iface->pkt_result.timestamp = (coral_timestamp_t*)&iface->u.pcap.hdr.ts;

    coral_clock_timespec_inline(iface, iface->pkt_result.timestamp, &iface->ts);

    if (!iface->iface_info.capture_tv.tv_sec)
	coral_set_iface_capture_tv(iface, &iface->ts, NULL);
    if (timestamp_duration_ended(iface, &iface->ts)) {
	return NULL;
    }

    iface->have_data = 1;
    return iface;
}

/* clean up stuff we malloc'd or opened in coral_file_open */
static int coral_pcap_close(coral_source_t *src)
{
    if (src->ubase)
	free(src->ubase);
    if (src->file)
	pcap_close(src->file);
    return 0;
}

#if 0
static int coral_pcap_fd(coral_source_t *src)
{
    return ((struct pcap*)src->file)->fd;
}
#endif

static int coral_pcaplive_has_buf(coral_source_t *src)
{
#ifdef HAVE_PCAP_INT_H
    /* Internal knowledge of libpcap:  p->cc contains the number
     * of bytes already read into p->buffer.
     */
    struct pcap *p = (struct pcap*)src->file;
    return (p->cc > 0);
#else
    /* Always return true.  This will cause a 1ms busy wait. */
    return 1;
#endif
}

static int coral_pcaplive_stop(coral_source_t *src)
{
    if (src->file) {
	coral_pcap_update_pkt_stats(cinst[src->id[0]]); /* get stats first! */
	pcap_close(src->file);
    }
    src->file = NULL;
    coral_mark_eof(cinst[src->id[0]]);
    return 0;
}

const coral_src_type_t coral_src_type_pcap = {
    CORAL_TYPE_PCAP,
    "pcap",
    0 /* is_live */,
    0 /* is_buffered */,
    0 /* is_coral_dev */,
    coral_pcap_init,
    coral_pcap_read_min,
    NULL,   /* nextblk */
    NULL,   /* release */
    NULL,   /* has_buf */
    coral_mark_eof_src,   /* stop */
    coral_pcap_close
};

const coral_src_type_t coral_src_type_pcaplive = {
    CORAL_TYPE_PCAPLIVE,
    "pcaplive",
    1 /* is_live */,
    0 /* is_buffered */,	/* pcap_dispatch returns before buffer fills */
    0 /* is_coral_dev */,
    coral_pcaplive_init,
    coral_pcap_read_min,
    NULL,   /* nextblk */
    NULL,   /* release */
    coral_pcaplive_has_buf,
    coral_pcaplive_stop,
    coral_pcap_close
};

