/* 
 * Copyright 2000, 2001 The Regents of the University of California 
 * All Rights Reserved 
 * 
 * Permission to use, copy, modify and distribute any part of this
 * CoralReef software package for educational, research and non-profit
 * purposes, without fee, and without a written agreement is hereby
 * granted, provided that the above copyright notice, this paragraph
 * and the following paragraphs appear in all copies.
 * 
 * Those desiring to incorporate this into commercial products or use
 * for commercial purposes should contact the Technology Transfer
 * Office, University of California, San Diego, 9500 Gilman Drive, La
 * Jolla, CA 92093-0910, Ph: (858) 534-5815, FAX: (858) 534-7345.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY
 * PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL
 * DAMAGES, INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE SOFTWARE PROVIDED HEREIN IS ON AN "AS IS" BASIS, AND THE
 * UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE,
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS. THE UNIVERSITY
 * OF CALIFORNIA MAKES NO REPRESENTATIONS AND EXTENDS NO WARRANTIES
 * OF ANY KIND, EITHER IMPLIED OR EXPRESS, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A
 * PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE
 * ANY PATENT, TRADEMARK OR OTHER RIGHTS.
 * 
 * The CoralReef software package is developed by the CoralReef
 * development team at the University of California, San Diego under
 * the Cooperative Association for Internet Data Analysis (CAIDA)
 * Program. Support for this effort is provided by the CAIDA grant
 * NCR-9711092, DARPA NGI Contract N66001-98-2-8922, DARPA NMS Grant
 * N66001-01-1-8909, and by CAIDA members.
 * 
 * Report bugs and suggestions to coral-bugs@caida.org.
 */

/*
 * This TSH format handler does not use a block interface (it does not have a
 * nextblk function nor use a binfo structure), but it still uses some of the
 * low-level block code (the per-interface block queue, coral_block_expired()).
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>

#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <errno.h>

#include "libcoral.h"
#include "libcoral_priv.h"

static const char RCSid[] = "$Id: coral_type_tsh.c,v 1.18.2.1 2001/11/21 01:33:50 kkeys Exp $";


#define TSH_RECORD_SIZE (44)
#define TSH_PAYLOAD_SIZE (36)

/* 44B doesn't divide evenly into 1MB, but this number is close to 1MB and is
 * also a multiple of 8196, so it will line up on disk block boundaries. */
#define TSH_BLK_SIZE (TSH_RECORD_SIZE * 22528)

static int coral_tsh_init(coral_source_t *src)
{
    int i;

    if (src->fd >= 0) {
	/* already open */
    } else if (strcmp(src->filename, "-") == 0) {
	src->fd = STDIN_FILENO;
    } else {
	if ((src->fd = open(src->filename, O_RDONLY, 0)) < 0) {
	    coral_diag(0, ("coral_open: %s: open: %s\n",
		src->name, strerror(errno)));
	    return -1;
	}
	coral_diag(4, ("coral_open: %s: fd %d\n", src->name, src->fd));
    }

    if (src->dev_config.iomode.flags == CORAL_RX_UNKNOWN) {
	src->dev_config.iomode.flags = CORAL_RX;
	src->dev_config.iomode.first_n = TSH_PAYLOAD_SIZE;
    }

    if (src->dev_config.iomode.flags & ~CORAL_RX) {
	char iomode_buffer[64];
	coral_format_iomode(iomode_buffer, &src->dev_config.iomode);
	coral_diag(0, ("%s: unsupported mode: %s\n",
	    src->filename, iomode_buffer));
	return -1;
    }

    if (src->dev_config.iomode.first_n != TSH_PAYLOAD_SIZE) {
	coral_diag(1, ("warning: %s: %s length is fixed at %d.\n",
	    src->filename, coral_iface_type_tsh.name, TSH_PAYLOAD_SIZE));
	src->dev_config.iomode.first_n = TSH_PAYLOAD_SIZE;
    }

    if (coral_file_common_init(src, TSH_BLK_SIZE) < 0)
	return -1;

    /* TSH format always has exactly 2 interfaces */
    if (coral_new_instance(src, 1) < 0)
	return -1;
    src->iface_count = 2;

    for (i = 0; i < src->iface_count; i++) {
	coral_iface_t *iface = cinst[src->id[i]];
	iface->devinfo.type = CORAL_TYPE_TSH;
	iface->iface_type = &coral_iface_type_tsh;
	iface->iface_info.hw_type		= CORAL_TYPE_NONE;
	iface->iface_info.hw_version		= 0;
	iface->iface_info.fw_type		= 0;
	iface->iface_info.fw_version		= 0;
	iface->iface_info.sw_type		= SW_TYPE_NLANR;
	iface->iface_info.iomode.flags		= 0;
	iface->iface_info.iomode.first_n	= TSH_PAYLOAD_SIZE;
	iface->iface_info.capture_time		= 0;
	iface->iface_info.bandwidth		= 0;
	iface->iface_info.time_is_le		= 0;
	iface->iface_info.datalink		= CORAL_NETPROTO_IPv4;
	iface->iface_info.physical		= CORAL_PROTO_UNKNOWN;
	iface->iface_info.tzoff			= 0;
	iface->iface_info.capture_tv.tv_sec	= 0;
	iface->iface_info.capture_tv.tv_usec	= 0;
    }

    /* Get time from the first record, and if it's normalized, use it as the
     * capture time for both interfaces.  (If timestamps are normalized, it
     * would be wrong to infer capture time from filename, since iface clocks
     * may not be synchronized with clock used to create filename time.)
     */
    {
	ssize_t len;
	const coral_timestamp_t *tstamp;
	struct timespec tspec;
	long max_rel_time =      3600; /* Startup can't take > 1 hour. */
	long min_abs_time = 883612800; /* 1998-01-01, before HWB started taking
					* traces with seconds in filename */

	len = coral_readall(src, src->ubase, TSH_BLK_SIZE);
	if (!len) {
	    return 0;
	} else if (len < 0) {
	    coral_diag(0, ("coral: %s: %s\n", src->filename, strerror(errno)));
	    return -1;
	}
	src->preread = len;
	tstamp = coral_cell_time_req(cinst[src->id[0]], src->ubase);

	/* Don't use coral_clock_timespec_*().  We don't want this time to
	 * get cached, since time_is_normal and capture_tv may change. */
	coral_iface_type_tsh.clock_read_sec_nsec(cinst[src->id[0]], tstamp,
	    &tspec.tv_sec, &tspec.tv_nsec, 0);

	if (tspec.tv_sec > min_abs_time) {
	    cinst[src->id[0]]->time_is_normal = 1;
	    cinst[src->id[1]]->time_is_normal = 1;
	    coral_set_iface_capture_tv(cinst[src->id[0]], &tspec, NULL);
	    coral_set_iface_capture_tv(cinst[src->id[1]], &tspec, NULL);
	} else if (tspec.tv_sec > max_rel_time) {
	    coral_diag(0, ("coral: %s: timestamp out of bounds (%ld)\n",
		src->filename, tspec.tv_sec));
	    return -1;
	}
    }

    /* If data timestamps aren't normalized, try to infer capture time
     * from filename.
     */
    if (!cinst[src->id[0]]->time_is_normal)
	if (!coral_infer_capture_time(src))
	    coral_diag(1, ("%s: could not find capture time; "
		"timestamps will not be normalized.\n", src->name));

    return 0;
}

/* Read next block if there is one, and store it on the first iface.
 * Or return NULL and errno==EAGAIN.  Nonblocking.  Checks for duration.
 */
static coral_iface_t *tsh_queue_nextblk(coral_source_t *src)
{
    coral_iface_t *iface[2];
    ssize_t len;
    uint32_t cell_count;

    if (src->eof) return NULL;
    iface[0] = cinst[src->id[0]];
    iface[1] = cinst[src->id[1]];

    /* Time-sorted reading api can not steal ubase. */
    assert(src->ubase);

    if (src->preread < TSH_BLK_SIZE) {
	len = coral_readall(src, src->ubase + src->preread, TSH_BLK_SIZE);
	if (len < 0) {
	    coral_diag(0, ("coral: %s: %s\n", src->filename, strerror(errno)));
	} else if (!(len += src->preread)) {
	    errno = 0; /* EOF */
	}
	if (len <= 0) {
	    cinst[src->id[0]]->have_data = 0;
	    cinst[src->id[1]]->have_data = 0;
	    if (errno == EAGAIN) return NULL;
	    if (!iface[0]->eof) coral_mark_eof(iface[0]);
	    if (!iface[1]->eof) coral_mark_eof(iface[1]);
	    if (errno)
		coral_diag(0, ("%s: %s\n", src->filename, strerror(errno)));
	    return NULL; /* with errno */
	}

    } else {
	len = src->preread;
    }
    src->preread = 0;

    if (len % TSH_RECORD_SIZE) {
	coral_diag(1, ("warning: %s: discarding %d bytes outside record.\n",
	    src->filename, (len % TSH_RECORD_SIZE)));
    }

    cell_count = len / TSH_RECORD_SIZE;

    coral_diag(19, ("libcoral: read buffer of %d records from ifaces %d, %d\n",
	cell_count, iface[0]->id, iface[1]->id));

    /* if a timestamp > iface->duration_end, truncate the block
     * and mark the source as EOF'd */
    if (iface[0]->duration_end.tv_sec >= 0) {
	cell_count = coral_block_expired(iface[0], src->ubase, cell_count, 0);
	iface[1]->expired = iface[0]->expired;
	if (!cell_count) {
	    if (!iface[0]->eof) coral_mark_eof(iface[0]);
	    if (!iface[1]->eof) coral_mark_eof(iface[1]);
	    errno = 0;
	    return NULL;
	}
    }

    iface[0]->u.blk.node.data = src->ubase;
    iface[0]->u.blk.node.cell_count = cell_count;
    iface[0]->u.blk.node.index = 0;
    iface[0]->has_read_blocks = 1;

    return iface[0];
}

/* Reads until it gets a non-empty block, EOF, EAGAIN, or error.  */
static coral_iface_t *coral_tsh_read_min(coral_iface_t *target_iface)
{
    coral_iface_t *iface[2];
    int id;

    iface[0] = cinst[target_iface->src->id[0]];
    iface[1] = cinst[target_iface->src->id[1]];

    if (!iface[0]->u.blk.node.data ||
        ++iface[0]->u.blk.node.index >= iface[0]->u.blk.node.cell_count)
    {
	/* We've reached the end of the current block. */
	if (iface[0]->u.blk.node.data) {
	    coral_diag(10, ("end of block on ifaces %d, %d\n",
		iface[0]->id, iface[1]->id));
	    if (iface[0]->expired) {
		coral_mark_eof_src(target_iface->src);
		errno = 0;
		return NULL;
	    }
	}
	if (!tsh_queue_nextblk(target_iface->src))
	    return NULL /* with errno */;
    }
    coral_clock_timespec_inline(iface[0],
	coral_cell_time_req(iface[0], coral_iccell(iface[0])),
	&iface[0]->ts);
    iface[1]->ts = iface[0]->ts;

    /* Everything was stored on iface[0].  We now figure out which iface the
     * next cell really belongs to, and artificially raise the timestamp
     * on the _other_ iface. */
    id = ((char*)coral_iccell(iface[0]))[4] - 1;
    iface[!id]->ts.tv_nsec = 1000000000; /* purposely illegal */
    iface[0]->have_data = 1;
    iface[1]->have_data = 1;
    return iface[id];
}

const coral_src_type_t coral_src_type_tsh = {
    CORAL_TYPE_TSH,
    "tsh",
    0 /* is_live */,
    0 /* is_buffered */,
    0 /* is_coral_dev */,
    coral_tsh_init,
    coral_tsh_read_min,
    NULL /* nextblk */,
    NULL /* release */,
    NULL /* has_buf */,
    coral_mark_eof_src /* stop */,
    coral_file_close
};
