#include "config.h"
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <libcoral.h>
#include <libcoral_priv.h>
#include <crl_byteorder.h>

#define HEADER_SIZE \
    (sizeof(coral_file_header_t) + sizeof(coral_iface_info_t) + 511) / 512 * 512

coral_blk_info_t binfo = {
    crl_htonl(0),	/* interface */
    crl_htonl(2048),	/* blk_size (intentionally weird) */
    crl_htonl(5),	/* cell_count */
    crl_htonl(1),	/* cells_lost */
    crl_htonl(2),	/* unknown_vpi_vci */
    { crl_htonl(940000000), crl_htonl(0) },
    { crl_htonl(940000002), crl_htonl(300000000) },
    ""
};

static struct timeval start = { 940000000, 0 };
static uint32_t dagcell[16];
static uint32_t tshcell[11];

static void write_cfh(int fd, const char *desc, int iface_count, int version,
    int type, int bw, int le, int size,
    coral_protocol_t phy, coral_protocol_t datalink)
{
    int i;
    coral_iface_info_t iinfo;

    coral_file_header_t cfh = {
	crl_htonl(CORAL_MAGIC),
	crl_htonl(HEADER_SIZE),
	crl_htonl(CORAL_VERSION_9),
	crl_htonl(CORAL_PROTO_ATM_AAL5),
	crl_htonl(0) /* site_id */,
	{ crl_htonl(0), crl_htonl(0), crl_htonl(0), crl_htonl(0) } /* cksum */,
	"cidr", crl_htonl(10), /* encoding */
	crl_htons(1) /* iface_info_count */,
	crl_htons(0) /* proto_rule_count */,
	"Bogus comment, should never be seen."
    };

    strcpy(cfh.desc, desc);
    cfh.version = crl_htonl(version);
    cfh.iface_info_count = htons(iface_count);
    write(fd, &cfh, sizeof(cfh));

    iinfo.hw_type = crl_htonl(type);
    iinfo.hw_version = crl_htonl(0);
    iinfo.fw_type = crl_htonl(0);
    iinfo.fw_version = crl_htonl(0);
    iinfo.sw_type = crl_htonl(SW_TYPE_CORAL);
    iinfo.sw_version = crl_htonl(version);
    iinfo.iomode.flags = crl_htonl(CORAL_RX);
    iinfo.iomode.first_n = crl_htonl(48);
    iinfo.capture_time = 0;
    iinfo.tzoff = crl_htonl(-8 * 3600);
    iinfo.capture_tv.tv_sec = htonl(start.tv_sec);
    iinfo.capture_tv.tv_usec = htonl(start.tv_usec);
    iinfo.bandwidth = crl_htonl(bw);
    iinfo.time_is_le = crl_htonl(le);
    iinfo.proto_rule_count = crl_htons(0);
    iinfo.iomode.flags = crl_htonl(CORAL_RX);
    iinfo.iomode.first_n = crl_htonl(48);
    iinfo.physical = crl_htonl(phy);
    iinfo.datalink = crl_htonl(datalink);
    iinfo.pad = 0;

    for (i = 0; i < iface_count; i++)
	write(fd, &iinfo, size);

    lseek(fd, HEADER_SIZE, SEEK_SET);
}

static void write_crl_cell(int fd, int cellnum, uint32_t fw, uint32_t hw)
{
    int i;
    coral_old_atm_cell_t cell;

    cell.cu.ui = htonl(0x02a00640);		/* vpi 42, vci 100 */
    cell.t.i[0] = crl_htolel(hw);
    cell.t.i[1] = crl_htolel(fw);
    for (i = 0; i < 12; i++)
	cell.p.payload_i[i] = htonl(cellnum << 24 | i << 16 | cellnum << 8 | i);
    write(fd, &cell, sizeof(cell));
}

static void write_dag_cell(int fd, int ifid, int cellnum,
    uint32_t sec, uint32_t nsec)
{
    int i;

    dagcell[1] = crl_htolel(start.tv_sec + sec);		/* seconds */
    dagcell[0] = crl_htolel((uint32_t)((nsec + ifid) * 4.294967296));	/* fraction */

    for (i = 0; i < 12; i++)
	dagcell[i+4] = htonl(0xDA << 24 | ifid << 16 | cellnum << 8 | i);
    write(fd, &dagcell, sizeof(dagcell));
}

static void write_dag_cells(int fd, int ifid)
{
    int cellnum = 0;

    write_dag_cell(fd, ifid, cellnum++, 0, 200000000);
    write_dag_cell(fd, ifid, cellnum++, 0, 300000000);
    write_dag_cell(fd, ifid, cellnum++, 1, 200000000);
    write_dag_cell(fd, ifid, cellnum++, 1, 700000000);
    write_dag_cell(fd, ifid, cellnum++, 3, 200000000+1000);
}

static void write_dag_crl_file(const char *filename, int ifnum)
{
    int fd;

    fd = open(filename, O_RDWR | O_CREAT | O_TRUNC, 0666);
    if (fd < 0) { perror(filename); exit(1); }

    write_cfh(fd, "OC12 POS PPPoHDLC dag", 1, CORAL_VERSION_34,
	CORAL_TYPE_DAG, KBPS_OC12c, 0, CORAL_V34_IFACE_INFO_SIZE,
	CORAL_PHY_POS, CORAL_DLT_PPPoHDLC);
    binfo.interface = crl_htonl(0);
    write(fd, &binfo, sizeof(binfo));

    write_dag_cells(fd, ifnum);
    close(fd);
}

int main(int argc, char *argv[])
{
    int fd;

    /* point crl */

    fd = open("point.crl", O_RDWR | O_CREAT | O_TRUNC, 0666);
    if (fd < 0) { perror("point.crl"); exit(1); }

    write_cfh(fd, "OC12 ATM RFC1483 point", 1, CORAL_VERSION_9,
	CORAL_TYPE_POINT, KBPS_OC12c, 1, CORAL_V9_IFACE_INFO_SIZE,
	CORAL_PHY_ATM, CORAL_DLT_ATM_RFC1483);
    binfo.interface = crl_htonl(0);
    write(fd, &binfo, sizeof(binfo));

    /*		       firmware    hardware */
    write_crl_cell(fd, 0, 0x00000007, (0xA1200) << 12);
    write_crl_cell(fd, 1, 0x00000008, (0x02C80) << 12);	/* hw wrap, fw inc */
    write_crl_cell(fd, 2, 0x0000002D, (0xC6C00) << 12);
    write_crl_cell(fd, 3, 0x00000040, (0xD9900) << 12);
    write_crl_cell(fd, 4, 0x0000007A, (0x12000+40) << 12);

    close(fd);


    /* fatm crl */

    fd = open("fatm.crl", O_RDWR | O_CREAT | O_TRUNC, 0666);
    if (fd < 0) { perror("fatm.crl"); exit(1); }

    write_cfh(fd, "OC3 ATM RFC1483 fatm", 2, CORAL_VERSION_9,
	CORAL_TYPE_FATM, KBPS_OC3c, 1, CORAL_V9_IFACE_INFO_SIZE,
	CORAL_PHY_ATM, CORAL_DLT_ATM_RFC1483);
    binfo.interface = crl_htonl(1);
    write(fd, &binfo, sizeof(binfo));

    /*		       firmware    hardware */
    write_crl_cell(fd, 0, 0x0000004C, (0x4B40) << 16);
    /* libcoral should compensate for failed fw clock increment  */
    write_crl_cell(fd, 1, 0x0000004C, (0x0E90) << 16);	/* hw wrap, fw !inc */
    write_crl_cell(fd, 2, 0x000001C9, (0xC380) << 16);
    write_crl_cell(fd, 3, 0x00000288, (0x7FA0) << 16);
    write_crl_cell(fd, 4, 0x000004C4, (0xB400+25) << 16);

    close(fd);


    /* fatm tsh */

    tshcell[ 2] = crl_htonl(0x4500003c);	/* IP */
    tshcell[ 3] = crl_htonl(0x009c0000);
    tshcell[ 4] = crl_htonl(0x40065804);
    tshcell[ 5] = crl_htonl(0x01020304);
    tshcell[ 6] = crl_htonl(0x05060708);
    tshcell[ 7] = crl_htonl(0x11110050);	/* TCP */
    tshcell[ 8] = crl_htonl(0x00000000);
    tshcell[ 9] = crl_htonl(0x00000000);
    tshcell[10] = crl_htonl(0x50000000);

    fd = open("fatm.tsh", O_RDWR | O_CREAT | O_TRUNC, 0666);
    if (fd < 0) { perror("fatm.tsh"); exit(1); }

    tshcell[0] = htonl(start.tv_sec+0);		/* seconds */
    tshcell[1] = htonl((1<<24) | 200000);	/* microseconds */
    write(fd, &tshcell, sizeof(tshcell));

    tshcell[0] = htonl(start.tv_sec+0);		/* seconds */
    tshcell[1] = htonl((2<<24) | 300000);	/* microseconds */
    write(fd, &tshcell, sizeof(tshcell));

    tshcell[0] = htonl(start.tv_sec+1);		/* seconds */
    tshcell[1] = htonl((1<<24) | 200000);	/* microseconds */
    write(fd, &tshcell, sizeof(tshcell));

    tshcell[0] = htonl(start.tv_sec+1);		/* seconds */
    tshcell[1] = htonl((2<<24) | 700000);	/* microseconds */
    write(fd, &tshcell, sizeof(tshcell));

    tshcell[0] = htonl(start.tv_sec+3);		/* seconds */
    tshcell[1] = htonl((1<<24) | (200000+1));	/* microseconds */
    write(fd, &tshcell, sizeof(tshcell));

    close(fd);


    /* dag crl */

    dagcell[ 2] = crl_htonl(64);		/* slen */
    dagcell[ 3] = crl_htonl(0<<16 | 64);	/* lctr, wlen */
#if 0
    dagcell[ 4] = crl_htonl(0xff030021ul);	/* PPP over HDLC */
    dagcell[ 5] = crl_htonl(0x4500003c);	/* IP */
    dagcell[ 6] = crl_htonl(0x009c0000);
    dagcell[ 7] = crl_htonl(0x40065804);
    dagcell[ 8] = crl_htonl(0x01020304);
    dagcell[ 9] = crl_htonl(0x05060708);
    dagcell[10] = crl_htonl(0x11110050);	/* TCP */
    dagcell[11] = crl_htonl(0x00000000);
    dagcell[12] = crl_htonl(0x00000000);
    dagcell[13] = crl_htonl(0x50000000);
    dagcell[14] = crl_htonl(0x00000000);
    dagcell[15] = crl_htonl(0x11223344);	/* payload */
#endif

    write_dag_crl_file("dag0.crl", 0);
    write_dag_crl_file("dag1.crl", 1);

    /* dagtools format */

    fd = open("dag0.dag", O_RDWR | O_CREAT | O_TRUNC, 0666);
    if (fd < 0) { perror("dag0.dag"); exit(1); }
    write_dag_cells(fd, 0);
    close(fd);

    fd = open("dag1.dag", O_RDWR | O_CREAT | O_TRUNC, 0666);
    if (fd < 0) { perror("dag1.dag"); exit(1); }
    write_dag_cells(fd, 1);
    close(fd);

    return 0;
}
