#include "config.h"
#undef NDEBUG
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <math.h>
#include <fcntl.h>	/* open() */
#include <libcoral.h>
#include <crl_byteorder.h>
#include <sys/types.h>	/* for coral_inet_pton() */
#include <sys/socket.h>	/* for coral_inet_pton() */
#include <netinet/in.h>	/* for coral_inet_pton() */
#include <arpa/inet.h>	/* for coral_inet_pton() */


typedef struct {
    int pkt;
    int i;
    struct timespec last_ts;
    int interval;
    int pktnum[2];
} pkt_data_t;

static struct timeval last_begin;
static const struct timespec start = { 940000000, 0 };
static coral_pkt_result_t pkt_result;
static coral_interval_result_t int_result;
static int src_count = 0;
static int printflag = 1;
static int block_hook_count;
static int verbosity = 8;

extern struct timeval coral_duration_end;

static coral_source_t *new_source(const char *filename, int fd)
{
    coral_source_t *src;
    if (fd > 0)
	src = coral_new_fdsource(fd, filename);
    else
	src = coral_new_source(filename);
    if (src) src_count++;
    assert(coral_get_source_count() == src_count);
    return src;
}

static int close_source(coral_source_t *src)
{
    int result;
    src_count--;
    result = coral_close(src);
    assert(coral_get_source_count() == src_count);
    return result;
}

static void block_hook(const coral_iface_t *iface,
    const coral_blk_info_t *binfo)
{
    assert(iface);
    assert(binfo);
    block_hook_count++;
}

static void test_read_cell(coral_source_t *src, int duration,
    int iface_type, int bw, int iface_count)
{
    int i, j;
    int cell_count;
    coral_iface_t *iface;
    coral_blk_info_t *binfo;
    coral_atm_cell_t *cell;
    struct timespec last_ts, ts;
    uint32_t *payload;

    last_ts.tv_sec = last_ts.tv_nsec = 0;
    coral_duration_end.tv_sec = coral_duration_end.tv_usec = -1; /* hack */
    assert(coral_set_duration(duration) == 0);
    assert(coral_get_duration() == duration);
    assert(coral_source_get_number(src) >= 0);
    assert(coral_open(src) == iface_count);
    block_hook_count = 0;
    if (printflag) coral_dump(src);
    assert((iface = coral_next_interface(NULL)) != NULL);
    /* XXX should test reading before starting */
    assert(coral_start(src) == 0);

    switch (duration) {
	case 0:  cell_count = 0;  break;
	case 1:  cell_count = 2;  break;
	case 2:  cell_count = 4;  break;
	case 3:  cell_count = 4;  break;
	default: cell_count = 5;  break;
    }

    i = 0;
    for (i = 0; (iface = coral_read_cell(src, &binfo, &cell, NULL)); i++) {
	if (verbosity > 0) {
	    coral_printf("\ncell %d:\n", i);
	    coral_fprint_cell(stderr, 4, iface, cell);
	}

	assert(coral_interface_get_type(iface) == iface_type);
	/*assert(ntohl(iface->hw_version) == 0);*/
	/*assert(ntohl(iface->fw_type) == 0);*/
	/*assert(ntohl(iface->fw_version) == 0);*/
	/*assert(ntohl(iface->sw_type) == SW_TYPE_CORAL);*/
	/*assert(ntohl(iface->sw_version) == CORAL_VERSION);*/
	/*assert(ntohl(iface->iomode.flags) == CORAL_RX);*/
	/*assert(ntohl(iface->iomode.first_n) == 1);*/
	/*assert(ntohl(iface->capture_time) == 0);*/
	assert(coral_interface_get_bandwidth(iface) == bw);
	/*assert(ntohl(iface->time_is_le) == 1);*/
	assert(coral_interface_get_physical(iface) == CORAL_PHY_ATM);
	assert(coral_interface_get_datalink(iface) == CORAL_DLT_ATM_RFC1483);
	assert(coral_interface_get_capture_tv(iface)->tv_sec == start.tv_sec);
	assert(coral_interface_get_capture_tv(iface)->tv_usec == start.tv_nsec/1000);
	/*assert(ntohl(iface->tzoff) == -8 * 3600);*/

	assert(ntohl(binfo->interface) == iface_count - 1);
	assert(ntohl(binfo->blk_size) == 2048);
	assert(ntohl(binfo->cell_count) == cell_count);
	assert(ntohl(binfo->cells_lost) == 1);
	assert(ntohl(binfo->unknown_vpi_vci) == 2);
	assert(ntohl(binfo->tbegin.tv_sec) == start.tv_sec);
	assert(ntohl(binfo->tbegin.tv_nsec) == start.tv_nsec);
	assert(ntohl(binfo->tend.tv_sec) == start.tv_sec+2);
	assert(ntohl(binfo->tend.tv_nsec) == 300000000);

	CORAL_TIMESTAMP_TO_TIMESPEC(iface, coral_cell_time(iface, cell), &ts);
	assert(timespeccmp(&ts, &last_ts, >));
	last_ts = ts;

	payload = (uint32_t*)coral_cell_payload(iface, cell);
	for (j = 0; j < 12; j++)
	    assert(ntohl(payload[j]) == (i << 24 | j << 16 | i << 8 | j));
    }
    assert(block_hook_count == 1);
    assert(i == cell_count);
    assert(errno == 0);
    assert(coral_stop(src) == 0);
}


static void verify_int_begin(void)
{
    coral_iface_t *iface;

    assert((iface = coral_read_pkt(&pkt_result, &int_result)) != NULL);
    assert(pkt_result.packet == NULL);
    assert(int_result.stats == NULL);
    coral_diag(1, ("interval begin:  %ld.%06ld\n",
	int_result.begin.tv_sec, int_result.begin.tv_usec));
}

static void verify_int_end(void)
{
    coral_iface_t *iface;

    assert((iface = coral_read_pkt(&pkt_result, &int_result)) != NULL);
    assert(pkt_result.packet == NULL);
    assert(int_result.stats != NULL);
    coral_diag(1, ("interval end:    %ld.%06ld - %ld.%06ld\n",
	int_result.begin.tv_sec, int_result.begin.tv_usec,
	int_result.end.tv_sec, int_result.end.tv_usec));
    assert(timercmp(&int_result.begin, &int_result.end, <));
}

static void verify_point_pkt(pkt_data_t *pkt_data)
{
    struct timespec ts;
    struct timeval tv;
    coral_iface_t *iface;
    /* depends on global int_result set by verify_int_begin */

    assert((iface = coral_read_pkt(&pkt_result, &int_result)) != NULL);
    assert(pkt_result.packet != NULL);
    assert(pkt_data->pkt < 5);
    coral_diag(1, ("pkt %d", pkt_data->pkt));
    pkt_data->pkt++;
    CORAL_TIMESTAMP_TO_TIMESPEC(iface, pkt_result.timestamp, &ts);
    CORAL_TIMESTAMP_TO_TIMEVAL (iface, pkt_result.timestamp, &tv);
    if (printflag) {
	coral_printf("\n");
	coral_fprint_pkt(stderr, iface, pkt_result.timestamp, 7,
	    pkt_result.packet, pkt_result.header, pkt_result.trailer);
	coral_printf("\n");
    } else {
	coral_diag(1, (": %10ld.%09ld\n", ts.tv_sec, ts.tv_nsec));
    }

    assert(coral_interface_get_type(iface) == CORAL_TYPE_POINT);
    assert(coral_interface_get_bandwidth(iface) == KBPS_OC12c);
    assert(coral_interface_get_physical(iface) == CORAL_PHY_ATM);
    assert(coral_interface_get_datalink(iface) == CORAL_DLT_ATM_RFC1483);
    assert(coral_interface_get_capture_tv(iface)->tv_sec == start.tv_sec);
    assert(coral_interface_get_capture_tv(iface)->tv_usec == start.tv_nsec/1000);
    /*assert(ntohl(iface->tzoff) == -8 * 3600);*/

    assert(timespeccmp(&ts, &pkt_data->last_ts, >));
    assert(timercmp(&tv, &int_result.begin, >=));
    pkt_data->last_ts = ts;
}

static void test_point_pkt(const char *filename, int duration)
{
    coral_source_t *src;
    pkt_data_t pkt_data;
    struct timeval interval;

    interval.tv_sec = 1;
    interval.tv_usec = 0;
    coral_diag(1, ("\n####### %s: duration %d, interval %ld\n",
	filename, duration, interval.tv_sec));
    assert((src = new_source(filename, -1)) != NULL);
    pkt_data.i = 0;
    pkt_data.pkt = 0;
    pkt_data.last_ts.tv_sec = pkt_data.last_ts.tv_nsec = 0;

    coral_duration_end.tv_sec = coral_duration_end.tv_usec = -1; /* hack */
    assert(coral_set_duration(duration) == 0);
    assert(coral_get_duration() == duration);
    assert(coral_source_get_number(src) >= 0);
    assert(coral_open(src) == 1);
    block_hook_count = 0;
    if (printflag) coral_dump(src);
    /* XXX should test reading before starting */
    assert(coral_start(src) == 0);
    assert(coral_read_pkt_init(src, NULL, &interval) == 0);

    if (duration <= 0) goto eof;
    verify_int_begin();
    verify_point_pkt(&pkt_data);
    verify_point_pkt(&pkt_data);
    verify_int_end();

    if (duration <= 1) goto eof;
    verify_int_begin();
    verify_point_pkt(&pkt_data);
    verify_point_pkt(&pkt_data);
    verify_int_end();

    if (duration <= 2) goto eof;
    verify_int_begin();
    verify_int_end();

    if (duration <= 3) goto eof;
    verify_int_begin();
    verify_point_pkt(&pkt_data);
    verify_int_end();

eof:
    assert(coral_read_pkt(&pkt_result, &int_result) == NULL);
    assert(block_hook_count == 1);
    assert(errno == 0);
    assert(coral_stop(src) == 0);
    coral_diag(1, ("\n"));
    assert(close_source(src) == 0);
}


/* packet interval callback */
static void pre_int_handler(coral_iface_t *iface, const struct timeval *begin,
    void *mydata)
{
    pkt_data_t *my_pkt_data = mydata;
    assert(iface != NULL);
    assert(my_pkt_data->i == 2 * ((begin->tv_sec%start.tv_sec)/my_pkt_data->interval));
    my_pkt_data->i++;
    coral_diag(1, ("begin interval %ld.%06ld\n", begin->tv_sec, begin->tv_usec));
    assert(timercmp(&last_begin, begin, <));
    last_begin = *begin;
}

/* packet interval callback */
static void post_int_handler(coral_iface_t *iface, const struct timeval *begin,
    const struct timeval *end, void *mydata, const coral_pkt_stats_t *stats)
{
    pkt_data_t *my_pkt_data = mydata;
    assert(iface != NULL);
    assert(my_pkt_data->i == 2 * ((begin->tv_sec%start.tv_sec)/my_pkt_data->interval) + 1);
    my_pkt_data->i++;
    coral_diag(1, ("end interval %ld.%06ld - %ld.%06ld\n",
	begin->tv_sec, begin->tv_usec, end->tv_sec, end->tv_usec));
    assert(timercmp(&last_begin, begin, ==));
    assert(timercmp(begin, end, <));
}

/* fatm packet callback */
static void fatm_handler(coral_iface_t *iface,
    const coral_timestamp_t *timestamp,
    void *mydata, coral_pkt_buffer_t *packet,
    coral_pkt_buffer_t *header, coral_pkt_buffer_t *trailer)
{
    pkt_data_t *my_pkt_data = mydata;
    struct timespec ts;

    assert(iface != NULL);

    CORAL_TIMESTAMP_TO_TIMESPEC(iface, timestamp, &ts);

    /* packet */
    assert(my_pkt_data->pkt < 5);
    coral_diag(1, ("pkt %d", my_pkt_data->pkt));
    my_pkt_data->pkt++;
    if (printflag) {
	coral_printf("\n");
	coral_fprint_pkt(stderr, iface, timestamp, 7, packet, header, trailer);
	coral_printf("\n");
    } else {
	coral_diag(1, (": %10ld.%09ld\n", ts.tv_sec, ts.tv_nsec));
    }

    assert(my_pkt_data->i == 2 * ((ts.tv_sec%start.tv_sec)/my_pkt_data->interval) + 1);

    assert(coral_interface_get_capture_tv(iface)->tv_sec == start.tv_sec);
    /*assert(ntohl(iface->tzoff) == -8 * 3600);*/

    assert(timespeccmp(&ts, &my_pkt_data->last_ts, >));
    my_pkt_data->last_ts = ts;
}

static void fatm_crl_handler(coral_iface_t *iface,
    const coral_timestamp_t *timestamp,
    void *mydata, coral_pkt_buffer_t *packet,
    coral_pkt_buffer_t *header, coral_pkt_buffer_t *trailer)
{
    assert(coral_source_get_type(coral_interface_get_src(iface)) == CORAL_TYPE_FILE);
    assert(coral_interface_get_type(iface) == CORAL_TYPE_FATM);
    assert(coral_interface_get_bandwidth(iface) == KBPS_OC3c);
    assert(coral_interface_get_physical(iface) == CORAL_PHY_ATM);
    assert(coral_interface_get_datalink(iface) == CORAL_DLT_ATM_RFC1483);
    assert(coral_interface_get_capture_tv(iface)->tv_usec == start.tv_nsec/1000);
    fatm_handler(iface, timestamp, mydata, packet, header, trailer);
}

static void fatm_tsh_handler(coral_iface_t *iface,
    const coral_timestamp_t *timestamp,
    void *mydata, coral_pkt_buffer_t *packet,
    coral_pkt_buffer_t *header, coral_pkt_buffer_t *trailer)
{
    assert(coral_source_get_type(coral_interface_get_src(iface)) == CORAL_TYPE_TSH);
    assert(coral_interface_get_type(iface) == CORAL_TYPE_NONE);
    assert(coral_interface_get_bandwidth(iface) == 0); /* unknown */
    assert(coral_interface_get_physical(iface) == CORAL_PROTO_UNKNOWN);
    assert(coral_interface_get_datalink(iface) == CORAL_NETPROTO_IP);
    /* TSH format has no capture_tv, so it's inferred from 1st cell */
    assert(coral_interface_get_capture_tv(iface)->tv_usec ==
	start.tv_nsec/1000 + 200000);
    fatm_handler(iface, timestamp, mydata, packet, header, trailer);
}

static void fatm_pcap_handler(coral_iface_t *iface,
    const coral_timestamp_t *timestamp,
    void *mydata, coral_pkt_buffer_t *packet,
    coral_pkt_buffer_t *header, coral_pkt_buffer_t *trailer)
{
    assert(coral_source_get_type(coral_interface_get_src(iface)) == CORAL_TYPE_PCAP);
    assert(coral_interface_get_type(iface) == CORAL_TYPE_NONE);
    assert(coral_interface_get_bandwidth(iface) == 0); /* unknown */
    assert(coral_interface_get_physical(iface) == CORAL_PROTO_UNKNOWN);
    assert(coral_interface_get_datalink(iface) == CORAL_DLT_ATM_RFC1483);
    assert(coral_interface_get_capture_tv(iface)->tv_usec > start.tv_nsec/1000);
    fatm_handler(iface, timestamp, mydata, packet, header, trailer);
}

#ifndef min
# define min(a,b)   ((a)<(b) ? (a) : (b))
#endif
static void verify_counts(pkt_data_t *pkt_data, int duration, int multiplier)
{
    switch (duration) {
	case 0:  assert(pkt_data->pkt == 0 * multiplier);  break;
	case 1:  assert(pkt_data->pkt == 2 * multiplier);  break;
	case 2:  assert(pkt_data->pkt == 4 * multiplier);  break;
	case 3:  assert(pkt_data->pkt == 4 * multiplier);  break;
	default: assert(pkt_data->pkt == 5 * multiplier);  break;
    }
    assert(pkt_data->i == 2 * ceil(min(duration,4)/(double)pkt_data->interval));
}

static void test_fatm_pkt(int fd, const char *filename, int duration,
    int interval_sec, int iface_count, coral_pkt_handler pkt_handler)
{
    coral_source_t *src;
    struct timeval interval;
    pkt_data_t pkt_data;
    int n;

    coral_diag(1, ("\n####### %s: duration %d, interval %d\n",
	filename, duration, interval_sec));
    assert((src = new_source(filename, fd)) != NULL);
    pkt_data.i = 0;
    pkt_data.pkt = 0;
    pkt_data.last_ts.tv_sec = pkt_data.last_ts.tv_nsec = 0;
    pkt_data.interval = interval_sec;
    assert(coral_set_duration(duration) == 0);
    assert(coral_get_duration() == duration);
    assert(coral_source_get_number(src) >= 0);
    assert((n = coral_open(src)) == iface_count);
    block_hook_count = 0;
    if (printflag) coral_dump(src);
    /* XXX should test reading before starting */
    assert(coral_start(src) == 0);

    interval.tv_sec = interval_sec;
    interval.tv_usec = 0;
    last_begin.tv_sec = last_begin.tv_usec = 0;
    coral_read_pkts(src, NULL, pkt_handler, pre_int_handler, post_int_handler,
	&interval, &pkt_data);

    assert(errno == 0);
    if (coral_source_get_type(src) == CORAL_TYPE_FILE)
	assert(block_hook_count == 1);
    else
	assert(block_hook_count == 0);
    verify_counts(&pkt_data, duration, 1);

    assert(coral_stop(src) == 0);
    assert(close_source(src) == 0);
    coral_diag(1, ("\n"));
}

/* dag packet callback */
static void dag_handler(coral_iface_t *iface,
    const coral_timestamp_t *timestamp,
    void *mydata, coral_pkt_buffer_t *packet,
    coral_pkt_buffer_t *header, coral_pkt_buffer_t *trailer)
{
    pkt_data_t *my_pkt_data = mydata;
    struct timespec ts;
    const struct timeval *capture_tv;
    coral_source_t *src;
    int ifnum, i;

    assert(iface != NULL);
    CORAL_TIMESTAMP_TO_TIMESPEC(iface, timestamp, &ts);

    /* packet */
    ifnum = coral_interface_get_number(iface);
    assert(my_pkt_data->pkt < 10);
    assert(my_pkt_data->pktnum[ifnum] < 5);
    coral_diag(1, ("pkt %d", my_pkt_data->pkt));
    if (printflag) {
	coral_printf("\n");
	coral_fprint_pkt(stderr, iface, timestamp, 7, packet, header, trailer);
	coral_printf("\n");
    } else {
	coral_diag(1, (": %10ld.%09ld\n", ts.tv_sec, ts.tv_nsec));
    }
    for (i = 0; i < 12; i++) {
	assert(((uint32_t*)packet->buf)[i] ==
	    htonl(0xDA<<24 | ifnum<<16 | my_pkt_data->pktnum[ifnum] << 8 | i));
    }
    my_pkt_data->pktnum[ifnum]++;
    my_pkt_data->pkt++;

    assert(my_pkt_data->i == 2 * ((ts.tv_sec%start.tv_sec)/my_pkt_data->interval) + 1);

    assert(coral_interface_get_type(iface) == CORAL_TYPE_DAG);
    assert(coral_interface_get_bandwidth(iface) == KBPS_OC12c);
    assert(coral_interface_get_physical(iface) == CORAL_PHY_POS);
    assert(coral_interface_get_datalink(iface) == CORAL_DLT_PPPoHDLC);
    capture_tv = coral_interface_get_capture_tv(iface);
    assert(capture_tv->tv_sec == start.tv_sec);
    src = coral_interface_get_src(iface);
    if (coral_source_get_type(src) == CORAL_TYPE_DAGFILE) {
	/* dagtools format has no capture_tv, so it's inferred from 1st cell */
	assert(capture_tv->tv_usec == start.tv_nsec/1000 + 200000);
    } else if (coral_source_get_type(src) == CORAL_TYPE_FILE) {
	assert(capture_tv->tv_usec == start.tv_nsec/1000);
    }
    /*assert(ntohl(iface->tzoff) == -8 * 3600);*/

    assert(timespeccmp(&ts, &my_pkt_data->last_ts, >));
    my_pkt_data->last_ts = ts;
}

static void test_dag(int use_fdopen, const char *suffix,
    int duration, int interval_sec)
{
    pkt_data_t pkt_data;
    coral_source_t *src;
    struct timeval interval;
    char file0[32];
    char file1[32];
    int n, fd;

    /* close everything to prepare for dag read */
    assert(coral_close_all() >= 0);
    assert(coral_get_source_count() == (src_count = 0));

    sprintf(file0, "dag0.%s", suffix);
    sprintf(file1, "dag1.%s", suffix);
    coral_diag(1, ("\n####### %s %s: duration %d, interval %d\n",
	file0, file1, duration, interval_sec));

    pkt_data.i = 0;
    pkt_data.pkt = 0;
    pkt_data.last_ts.tv_sec = pkt_data.last_ts.tv_nsec = 0;
    pkt_data.interval = interval_sec;
    pkt_data.pktnum[0] = 0;
    pkt_data.pktnum[1] = 0;

    assert(coral_set_duration(duration) == 0);
    assert(coral_get_duration() == duration);

    if (use_fdopen) {
	assert((fd = open(file0, O_RDONLY)) >= 0);
	assert((src = new_source(file0+3, fd)) != NULL);
    } else {
	assert((src = new_source(file0, -1)) != NULL);
    }
    assert(coral_source_get_number(src) >= 0);
    assert((n = coral_open(src)) == 1);
    if (printflag) coral_dump(src);
#if 1
    if (use_fdopen) {
	assert((fd = open(file1, O_RDONLY)) >= 0);
	assert((src = new_source(file1+3, fd)) != NULL);
    } else {
	assert((src = new_source(file1, -1)) != NULL);
    }
    assert(coral_source_get_number(src) >= 0);
    assert((n = coral_open(src)) == 1);
    if (printflag) coral_dump(src);
#endif
    block_hook_count = 0;
    /* XXX should test reading before starting */
    assert(coral_start_all() == 0);

    interval.tv_sec = pkt_data.interval;
    interval.tv_usec = 0;
    last_begin.tv_sec = last_begin.tv_usec = 0;

    coral_read_pkts(NULL, NULL, dag_handler, pre_int_handler, post_int_handler,
	&interval, &pkt_data);

    assert(errno == 0);
    assert(block_hook_count == 2);
    verify_counts(&pkt_data, duration, 2);
    assert(coral_stop_all() == 0);
    coral_diag(1, ("\n"));
}

static void test_inet_pton(const char *src)
{
    struct in_addr myaddr, stdaddr;
    int myresult, stdresult;

    stdresult = inet_aton(src, &stdaddr);
    myresult = coral_inet_pton(AF_INET, src, &myaddr);
    assert(stdresult == myresult);
    assert(memcmp(&myaddr, &stdaddr, sizeof(struct in_addr)) == 0);
}


int main(int argc, char *argv[])
{
    coral_source_t *src;
    const char comment[] = "This is a comment.";
    int argc_a = 2;
    const char *argv_a[] = { "argv_a", "-Csource point.crl", NULL };
    int argc_b = 4;
    const char *argv_b[] = { "argv_b", "-C", "iomode user", "fatm.crl", NULL };
    int argc_c = 3;
    const char *argv_c[] = { "argv_c", "-Csource fatm.crl", "-Cduration 60",
	NULL };
    int default_options = CORAL_OPT_PARTIAL_PKT | CORAL_OPT_NORMALIZE_TIME;
    struct timeval tv1, tv2;
    int fd;

    printf("Note: some error messages are expected.  A failed test will be "
	"indicated by a failed assertion.\n\n");

    if (argc > 2) {
	fprintf(stderr, "usage: %s [verbosity]\n", argv[0]);
	exit(1);
    } else if (argc > 1) {
	verbosity = atoi(argv[1]);
    }

    test_inet_pton("1.2.3.4");
    test_inet_pton("1.2..4");
    test_inet_pton("255.255.255.255");
    test_inet_pton("255.255.255.256");
    test_inet_pton("0.0.0.0");

    /* config */

    coral_cell_block_hook = block_hook;

    assert(coral_get_source_count() == 0);
    assert(coral_next_source(NULL) == NULL);
    assert(coral_next_interface(NULL) == NULL);
    assert(coral_get_max_sources() == 16);
    assert(coral_get_options() == default_options);
    assert(coral_get_errfile() == stderr);
    assert(coral_get_verbosity() == 1);
    assert(coral_get_comment() == NULL);
    assert(coral_get_duration() == 0);
    assert(coral_get_interval(&tv1) == -1);
    assert(tv1.tv_sec == -1 && tv1.tv_usec == -1);

    assert(coral_set_max_sources(12) == 0);
    assert(coral_set_max_sources(120) == -1);
    assert(coral_get_max_sources() == 12);

    assert(coral_set_options(0, CORAL_OPT_SORT_TIME) == 0);
    assert(coral_get_options() == (default_options | CORAL_OPT_SORT_TIME));
    assert(coral_set_options(CORAL_OPT_SORT_TIME, 0) == 0);
    assert(coral_get_options() == default_options);

    assert(coral_set_errfilename("errfile") == 0);
    assert(coral_get_errfile() != NULL);
    assert(coral_get_errfile() != stderr);
    assert(coral_set_errfile(NULL) == 0);
    assert(coral_get_errfile() == stderr);
    assert(coral_set_errfile(stderr) == 0);
    assert(coral_get_errfile() == stderr);

    assert(coral_set_verbosity(verbosity) == 0);
    assert(coral_get_verbosity() == verbosity);
    printflag = verbosity > 0;

    assert(coral_set_comment(comment) == 0);
    assert(strcmp(coral_get_comment(), comment) == 0);

    assert(coral_set_duration(30) == 0);
    assert(coral_get_duration() == 30);
    tv1.tv_sec = 2;
    tv1.tv_usec = 5;
    assert(coral_set_interval(&tv1) == 0);
    assert(coral_get_interval(&tv2) == 0);
    assert(tv1.tv_sec == tv2.tv_sec && tv1.tv_usec == tv2.tv_usec);

    assert((src = new_source("no_such_file", -1)) != NULL);
    assert(coral_source_get_number(src) >= 0);
    assert(coral_next_source(NULL) == src);
    assert(coral_next_source(src) == NULL);
    assert(coral_open(src) == -1);
    assert(errno == ENOENT);
    assert(coral_get_source_count() == --src_count);


    /* command line options */

    assert(coral_config_arguments(argc_a, (char**)argv_a) >= 0);
    assert(coral_get_source_count() == ++src_count);
    assert((src = coral_next_source(NULL)));
    assert(coral_open(src) == 1);
    assert(strcmp(coral_source_get_filename(src), "point.crl") == 0);
    assert(coral_source_get_type(src) == CORAL_TYPE_FILE);

    optind = 1; opterr = 0; optarg = NULL;
    assert(coral_config_arguments(argc_b, (char**)argv_b) >= 0);
    assert(coral_get_source_count() == ++src_count);
    assert((src = coral_next_source(src)));
    assert(strcmp(coral_source_get_filename(src), "fatm.crl") == 0);
    assert(coral_open(src) == 2);
    assert(coral_source_get_type(src) == CORAL_TYPE_FILE);

    optind = 1; opterr = 0; optarg = NULL;
    assert(coral_config_arguments(argc_c, (char**)argv_c) >= 0);
    assert(coral_get_source_count() == ++src_count);
    assert((src = coral_next_source(src)));
    assert(coral_open(src) == 2);
    assert(strcmp(coral_source_get_filename(src), "fatm.crl") == 0);
    assert(coral_source_get_type(src) == CORAL_TYPE_FILE);
    assert(coral_get_duration() == 60);

    assert(!(src = coral_next_source(src)));
    coral_close_all();
    assert(coral_get_source_count() == (src_count = 0));


    /* reading point: cells */
    assert((src = new_source("point.crl", -1)) != NULL);
    assert(coral_next_source(NULL) == src);
    assert(coral_next_source(src) == NULL);
    assert(coral_next_interface(NULL) == NULL);
    test_read_cell(src, 4, CORAL_TYPE_POINT, KBPS_OC12c, 1);  /* hit EOF */
    coral_diag(1, ("\n"));

    /* reading point: cells */
    assert((src = new_source("point.crl", -1)) != NULL);
    test_read_cell(src, 2, CORAL_TYPE_POINT, KBPS_OC12c, 1);  /* hit duration */
    coral_diag(1, ("\n"));

    /* reading point: packets (single) */
    printflag = (verbosity > 0);
    test_point_pkt("point.crl", 4); /* should hit EOF */
    printflag = 0;
    test_point_pkt("point.crl", 2); /* should hit duration */

    /* reading fatm: cells */
    assert((src = new_source("fatm.crl", -1)) != NULL);
    test_read_cell(src, 4, CORAL_TYPE_FATM, KBPS_OC3c, 2);
    assert(close_source(src) == 0);
    coral_diag(1, ("\n"));

    /* reading fatm: packets (callback) */
    printflag = (verbosity > 0);
    test_fatm_pkt(-1, "fatm.crl", 4, 1, 2, fatm_crl_handler); /* EOF < duration */
    printflag = 0;
    test_fatm_pkt(-1, "fatm.crl", 2, 1, 2, fatm_crl_handler); /* dur < EOF */
    test_fatm_pkt(-1, "fatm.crl", 4, 10, 2, fatm_crl_handler); /* EOF < int */
    test_fatm_pkt(-1, "fatm.crl", 2, 10, 2, fatm_crl_handler); /* dur < int */

    /* reading pcap: packets (callback) */
    printflag = (verbosity > 0);
    test_fatm_pkt(-1, "fatm.pcap", 4, 1, 1, fatm_pcap_handler); /* EOF < dur */
    printflag = 0;
    test_fatm_pkt(-1, "fatm.pcap", 2, 1, 1, fatm_pcap_handler); /* dur < EOF */
    test_fatm_pkt(-1, "fatm.pcap", 4, 10, 1, fatm_pcap_handler); /* EOF < int */
    test_fatm_pkt(-1, "fatm.pcap", 2, 10, 1, fatm_pcap_handler); /* dur < int */

    /* fdopen pcap: not allowed */
    assert((fd = open("fatm.pcap", O_RDONLY)) >= 0);
    assert((src = coral_new_fdsource(fd, "pcap:")) != NULL);
    assert(coral_open(src) == -1);
    close(fd);

    /* reading tsh: packets (callback) */
    coral_set_iomode(0, CORAL_RX_UNKNOWN, 0, 0);
    printflag = (verbosity > 0);
    test_fatm_pkt(-1, "fatm.tsh", 4, 1, 2, fatm_tsh_handler); /* EOF < dur */
    printflag = 0;
    test_fatm_pkt(-1, "fatm.tsh", 2, 1, 2, fatm_tsh_handler); /* dur < EOF */
    test_fatm_pkt(-1, "fatm.tsh", 4, 10, 2, fatm_tsh_handler); /* EOF < int */
    test_fatm_pkt(-1, "fatm.tsh", 2, 10, 2, fatm_tsh_handler); /* dur < int */

    /* fdopen tsh: packets (callback) */
    assert((fd = open("fatm.tsh", O_RDONLY)) >= 0);
    test_fatm_pkt(fd, "tsh:", 2, 10, 2, fatm_tsh_handler); /* dur < int */
    close(fd);

    /* reading dag: packets (callback) */
    printflag = (verbosity > 0);
    test_dag(0, "crl", 4, 1); /* EOF < duration */
    printflag = 0;
    test_dag(0, "crl", 2, 1); /* duration < EOF */
    test_dag(0, "crl", 4, 10); /* EOF < interval */
    test_dag(0, "crl", 2, 10); /* duration < interval */

    /* reading dagtools format: open should fail because protocol is unknown */
    assert((src = new_source("dag0.dag", -1)) != NULL);
    assert(coral_source_get_number(src) >= 0);
    assert(coral_open(src) < 0);
    src_count--;

    coral_set_api(CORAL_API_PKT);   /* allow proto rules */

    /* reading dagtools format: packets (callback) */
    assert(coral_config_command("iomode bw=oc12c,phy=pos,proto=ppp,48") == 0);
    printflag = (verbosity > 0);
    test_dag(0, "dag", 4, 1); /* EOF < duration */
    printflag = 0;
    test_dag(0, "dag", 2, 1); /* duration < EOF */
    test_dag(0, "dag", 4, 10); /* EOF < interval */
    test_dag(0, "dag", 2, 10); /* duration < interval */

#if 1
    /* fdopen dagfile: packets (callback) */
    test_dag(1, "dag", 2, 10); /* duration < interval */
#endif

    /* done reading */

    assert(coral_close_all() >= 0);
    assert(coral_get_source_count() == (src_count = 0));

#ifdef HAVE_INET6
{
    coral_pkt_buffer_t pkt;
    int ipv6_payload[] = {
	crl_htonl(0x60000000), /* version, tclass, flow label */
	crl_htonl(0x2000002a), /* plen, nxt, hops */
	crl_htonl(0x00010203), /* src */
	crl_htonl(0x04050607),
	crl_htonl(0x08090a0b),
	crl_htonl(0x0c0d0e0f),
	crl_htonl(0xffeeddcc), /* dst */
	crl_htonl(0xbbaa9988),
	crl_htonl(0x77665544),
	crl_htonl(0x33221100),
	crl_htonl(0x06000000), /* HOPOPTS: nxt, len, pad1, pad1 */
	crl_htonl(0x01020000), /* padN */
	crl_htonl(0x04d20050), /* TCP: sport, dport */
	crl_htonl(0x12345678), /* seq */
	crl_htonl(0x87654321), /* ack */
	crl_htonl(0x50100000), /* off, flags, win */
	crl_htonl(0xabcd0000), /* cksum, urg */
	crl_htonl(0x01234567)  /* data */
    };

    pkt.buf = (char*)ipv6_payload;
    pkt.protocol = CORAL_NETPROTO_IPv6;
    pkt.caplen = sizeof(ipv6_payload);
    pkt.totlen = sizeof(ipv6_payload);
    if (verbosity > 0)
	coral_fprint_pkt(stdout, NULL, NULL, 7, &pkt, NULL, NULL);
}
#else
    printf("\nUnable to test IPv6.");
#endif

    printf("\n%s: all tests passed.\n", argv[0]);
    return 0;
}
